@@ -25,20 +25,27 @@
 #include <gdk/gdkkeysyms.h>
 #include "swfdec_gtk_widget.h"
 #include "swfdec_gtk_keys.h"
+#include "swfdec_gtk_player.h"
 
 struct _SwfdecGtkWidgetPrivate
 {
   SwfdecPlayer *	player;		/* the video we play */
 
   gboolean		renderer_set;	/* TRUE if a special renderer has been set */
-  cairo_surface_type_t	renderer;	/* the renderer that was set */
+  cairo_surface_type_t	renderer_type;	/* the renderer that was set */
+  SwfdecRenderer *	renderer;	/* renderer in use */
   gboolean		interactive;	/* TRUE if this widget propagates keyboard and mouse events */
+  GdkRegion *		invalid;	/* invalid regions we didn't yet repaint */
+  guint			invalidator;	/* GSource used for invalidating window contents */
+  GtkWidget *		fullscreen_window;/* the window we use for displaying fullscreen */
+  gboolean		fullscreen_mode;/* TRUE if we're a fullscreen widget */
 };
 
 enum {
   PROP_0,
-  PROP_PLAYER,
+  PROP_FULLSCREEN_MODE,
   PROP_INTERACTIVE,
+  PROP_PLAYER,
   PROP_RENDERER_SET,
   PROP_RENDERER
 };
@@ -103,7 +110,10 @@ swfdec_gtk_widget_button_press (GtkWidget *gtkwidget, GdkEventButton *event)
   SwfdecGtkWidgetPrivate *priv = widget->priv;
 
   if (event->type == GDK_BUTTON_PRESS && event->button <= 32 && priv->interactive && priv->player) {
-    swfdec_player_mouse_press (priv->player, event->x, event->y, event->button);
+    SwfdecPlayer *player = priv->player;
+    swfdec_player_set_allow_fullscreen (player, TRUE);
+    swfdec_player_mouse_press (player, event->x, event->y, event->button);
+    swfdec_player_set_allow_fullscreen (player, FALSE);
   }
   return FALSE;
 }
@@ -115,7 +125,10 @@ swfdec_gtk_widget_button_release (GtkWidget *gtkwidget, GdkEventButton *event)
   SwfdecGtkWidgetPrivate *priv = widget->priv;
 
   if (event->button <= 32 && priv->interactive && priv->player) {
-    swfdec_player_mouse_release (priv->player, event->x, event->y, event->button);
+    SwfdecPlayer *player = priv->player;
+    swfdec_player_set_allow_fullscreen (player, TRUE);
+    swfdec_player_mouse_release (player, event->x, event->y, event->button);
+    swfdec_player_set_allow_fullscreen (player, FALSE);
   }
   return FALSE;
 }
@@ -156,8 +169,11 @@ swfdec_gtk_widget_key_press (GtkWidget *gtkwidget, GdkEventKey *event)
   if (priv->interactive && priv->player) {
     guint keycode = swfdec_gtk_event_to_keycode (event);
     if (keycode != 0) {
-      swfdec_player_key_press (priv->player, keycode, 
+      SwfdecPlayer *player = priv->player;
+      swfdec_player_set_allow_fullscreen (player, TRUE);
+      swfdec_player_key_press (player, keycode, 
 	  gdk_keyval_to_unicode (event->keyval));
+      swfdec_player_set_allow_fullscreen (player, FALSE);
     }
     return TRUE;
   }
@@ -174,8 +190,11 @@ swfdec_gtk_widget_key_release (GtkWidget *gtkwidget, GdkEventKey *event)
   if (priv->interactive && priv->player) {
     guint keycode = swfdec_gtk_event_to_keycode (event);
     if (keycode != 0) {
-      swfdec_player_key_release (priv->player, keycode, 
+      SwfdecPlayer *player = priv->player;
+      swfdec_player_set_allow_fullscreen (player, TRUE);
+      swfdec_player_key_release (player, keycode, 
 	  gdk_keyval_to_unicode (event->keyval));
+      swfdec_player_set_allow_fullscreen (player, FALSE);
     }
     return TRUE;
   }
@@ -183,6 +202,45 @@ swfdec_gtk_widget_key_release (GtkWidget *gtkwidget, GdkEventKey *event)
   return FALSE;
 }
 
+/* NB: called for both focus in and focus out */
+static gboolean
+swfdec_gtk_widget_focus_inout (GtkWidget *gtkwidget, GdkEventFocus *focus)
+{
+  SwfdecGtkWidget *widget = SWFDEC_GTK_WIDGET (gtkwidget);
+  SwfdecGtkWidgetPrivate *priv = widget->priv;
+
+  if (priv->interactive && priv->player)
+    swfdec_player_set_focus (priv->player, focus->in);
+  return FALSE;
+}
+
+static gboolean
+swfdec_gtk_widget_focus (GtkWidget *gtkwidget, GtkDirectionType direction)
+{
+  SwfdecGtkWidget *widget = SWFDEC_GTK_WIDGET (gtkwidget);
+  SwfdecGtkWidgetPrivate *priv = widget->priv;
+
+  if (!priv->interactive || priv->player == NULL)
+    return FALSE;
+
+  return GTK_WIDGET_CLASS (swfdec_gtk_widget_parent_class)->focus (gtkwidget, direction);
+}
+
+static void
+swfdec_gtk_widget_clear_invalidations (SwfdecGtkWidget *widget)
+{
+  SwfdecGtkWidgetPrivate *priv = widget->priv;
+
+  if (priv->invalidator) {
+    g_source_remove (priv->invalidator);
+    priv->invalidator = 0;
+    gdk_region_destroy (priv->invalid);
+    priv->invalid = gdk_region_new ();
+  } else {
+    g_assert (gdk_region_empty (priv->invalid));
+  }
+}
+
 static cairo_surface_t *
 swfdec_gtk_widget_create_renderer (cairo_surface_type_t type, int width, int height)
 {
@@ -201,32 +259,39 @@ swfdec_gtk_widget_expose (GtkWidget *gtkwidget, GdkEventExpose *event)
   cairo_t *cr;
   cairo_surface_t *surface = NULL;
 
-  if (event->window != gtkwidget->window)
-    return FALSE;
-  if (priv->player == NULL)
+  if (event->window != gtkwidget->window ||
+      priv->player == NULL ||
+      priv->fullscreen_window != NULL)
     return FALSE;
 
+  /* FIXME: This might be ugly */
+  gdk_region_union (event->region, priv->invalid);
+  gdk_window_begin_paint_region (event->window, event->region);
+
   if (!priv->renderer_set ||
-      (surface = swfdec_gtk_widget_create_renderer (priv->renderer, 
+      (surface = swfdec_gtk_widget_create_renderer (priv->renderer_type, 
 	      event->area.width, event->area.height)) == NULL) {
     cr = gdk_cairo_create (gtkwidget->window);
   } else {
+    cairo_surface_set_device_offset (surface, -event->area.x, -event->area.y);
     cr = cairo_create (surface);
-    cairo_translate (cr, -event->area.x, -event->area.y);
   }
-  swfdec_player_render (priv->player, cr,
-      event->area.x, event->area.y, event->area.width, event->area.height);
+  gdk_cairo_region (cr, event->region);
+  cairo_clip (cr);
+  swfdec_player_render (priv->player, cr);
   cairo_show_page (cr);
   cairo_destroy (cr);
 
   if (surface) {
     cairo_t *crw = gdk_cairo_create (gtkwidget->window);
-    cairo_set_source_surface (crw, surface, event->area.x, event->area.y);
+    cairo_set_source_surface (crw, surface, 0, 0);
     cairo_paint (crw);
     cairo_destroy (crw);
     cairo_surface_destroy (surface);
   }
 
+  swfdec_gtk_widget_clear_invalidations (widget);
+  gdk_window_end_paint (event->window);
   return FALSE;
 }
 
@@ -248,7 +313,7 @@ swfdec_gtk_widget_get_property (GObject *object, guint param_id, GValue *value,
       g_value_set_boolean (value, priv->renderer_set);
       break;
     case PROP_RENDERER:
-      g_value_set_uint (value, priv->renderer);
+      g_value_set_uint (value, priv->renderer_type);
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
@@ -264,18 +329,21 @@ swfdec_gtk_widget_set_property (GObject *object, guint param_id, const GValue *v
   SwfdecGtkWidgetPrivate *priv = widget->priv;
   
   switch (param_id) {
-    case PROP_PLAYER:
-      swfdec_gtk_widget_set_player (widget, g_value_get_object (value));
+    case PROP_FULLSCREEN_MODE:
+      priv->fullscreen_mode = g_value_get_boolean (value);
       break;
     case PROP_INTERACTIVE:
       swfdec_gtk_widget_set_interactive (widget, g_value_get_boolean (value));
       break;
+    case PROP_PLAYER:
+      swfdec_gtk_widget_set_player (widget, g_value_get_object (value));
+      break;
     case PROP_RENDERER_SET:
       priv->renderer_set = g_value_get_boolean (value);
       gtk_widget_queue_draw (GTK_WIDGET (widget));
       break;
     case PROP_RENDERER:
-      priv->renderer = g_value_get_uint (value);
+      priv->renderer_type = g_value_get_uint (value);
       if (priv->renderer_set)
 	gtk_widget_queue_draw (GTK_WIDGET (widget));
       break;
@@ -289,9 +357,17 @@ static void
 swfdec_gtk_widget_dispose (GObject *object)
 {
   SwfdecGtkWidget *widget = SWFDEC_GTK_WIDGET (object);
+  SwfdecGtkWidgetPrivate *priv = widget->priv;
 
   swfdec_gtk_widget_set_player (widget, NULL);
 
+  g_assert (priv->renderer == NULL);
+  if (priv->invalid) {
+    gdk_region_destroy (priv->invalid);
+    priv->invalid = NULL;
+  }
+  g_assert (priv->invalidator == 0);
+
   G_OBJECT_CLASS (swfdec_gtk_widget_parent_class)->dispose (object);
 }
 
@@ -303,7 +379,7 @@ swfdec_gtk_widget_size_allocate (GtkWidget *gtkwidget, GtkAllocation *allocation
 
   gtkwidget->allocation = *allocation;
 
-  if (priv->player && swfdec_player_is_initialized (priv->player))
+  if (priv->player && swfdec_player_is_initialized (priv->player) && !priv->fullscreen_mode)
     swfdec_player_set_size (priv->player, allocation->width, allocation->height);
   if (GTK_WIDGET_REALIZED (gtkwidget)) {
     gdk_window_move_resize (gtkwidget->window, 
@@ -319,12 +395,23 @@ swfdec_gtk_widget_size_request (GtkWidget *gtkwidget, GtkRequisition *req)
 
   if (priv->player == NULL) {
     req->width = req->height = 0;
+#if 0
+  } else if (priv->fullscreen_mode) {
+    GdkScreen *screen = gtk_widget_has_screen (gtkwidget) ? gtk_widget_get_screen (gtkwidget) : NULL;
+    req->width = screen ? gdk_screen_get_width (screen) : 0;
+    req->height = screen ? gdk_screen_get_height (screen) : 0;
+#endif
   } else {
     guint w, h;
     swfdec_player_get_default_size (priv->player, &w, &h);
-    /* FIXME: set some sane upper limit here? */
     req->width = MIN (w, G_MAXINT);
     req->height = MIN (h, G_MAXINT);
+    /* constrain the size - fix for huge (aka broken) Flash files */
+    if (gtk_widget_has_screen (gtkwidget)) {
+      GdkScreen *screen = gtk_widget_get_screen (gtkwidget);
+      req->width = MIN (req->width, gdk_screen_get_width (screen));
+      req->height = MIN (req->height, gdk_screen_get_height (screen));
+    }
   } 
 }
 
@@ -382,6 +469,69 @@ swfdec_gtk_widget_update_cursor (SwfdecGtkWidget *widget)
   }
 }
 
+static void
+swfdec_gtk_widget_update_renderer (SwfdecGtkWidget *widget)
+{
+  SwfdecGtkWidgetPrivate *priv = widget->priv;
+  gboolean needs_renderer;
+
+  needs_renderer = GTK_WIDGET_REALIZED (widget) &&
+    priv->player != NULL;
+
+  if (priv->renderer != NULL) {
+    swfdec_player_set_renderer (priv->player, NULL);
+    g_object_unref (priv->renderer);
+    priv->renderer = NULL;
+  }
+  if (needs_renderer) {
+    if (priv->fullscreen_mode) {
+      /* FIXME: We should really use the renderer of the app that 
+       * fullscreened us, but for now, we'll assume that app uses a
+       * similar renderer */
+      priv->renderer = swfdec_player_get_renderer (priv->player);
+      g_object_ref (priv->renderer);
+    } else if (priv->renderer_set) {
+      cairo_surface_t *surface = swfdec_gtk_widget_create_renderer (priv->renderer_type, 1, 1);
+      if (surface) {
+	priv->renderer = swfdec_renderer_new_for_player (
+	    surface, priv->player);
+	swfdec_player_set_renderer (priv->player, priv->renderer);
+	cairo_surface_destroy (surface);
+      }
+    } else {
+      cairo_t *cr = gdk_cairo_create (GTK_WIDGET (widget)->window);
+
+      priv->renderer = swfdec_renderer_new_for_player (
+	  cairo_get_target (cr), priv->player);
+      swfdec_player_set_renderer (priv->player, priv->renderer);
+      cairo_destroy (cr);
+    }
+  }
+}
+
+static void
+swfdec_gtk_widget_update_background (SwfdecGtkWidget *widget)
+{
+  GdkWindow *window = GTK_WIDGET (widget)->window;
+  GdkColor bgcolor;
+
+  if (window == NULL)
+    return;
+
+  if (widget->priv->player) {
+    guint bg = swfdec_player_get_background_color (widget->priv->player);
+    bgcolor.red = 0x101 * ((bg >> 16) & 0xFF);
+    bgcolor.green = 0x101 * ((bg >> 8) & 0xFF);
+    bgcolor.blue = 0x101 * (bg & 0xFF);
+  } else {
+    /* white */
+    bgcolor.red = bgcolor.green = bgcolor.blue = 0xFFFF;
+  }
+  gdk_rgb_find_color (gdk_drawable_get_colormap (window), &bgcolor);
+  gdk_window_set_background (window, &bgcolor);
+
+}
+
 static void
 swfdec_gtk_widget_realize (GtkWidget *widget)
 {
@@ -404,7 +554,8 @@ swfdec_gtk_widget_realize (GtkWidget *widget)
 			   GDK_POINTER_MOTION_MASK | 
 			   GDK_POINTER_MOTION_HINT_MASK |
 			   GDK_KEY_PRESS_MASK |
-			   GDK_KEY_RELEASE_MASK;
+			   GDK_KEY_RELEASE_MASK |
+			   GDK_FOCUS_CHANGE_MASK;
 
   attributes_mask = GDK_WA_X | GDK_WA_Y;
 
@@ -412,11 +563,47 @@ swfdec_gtk_widget_realize (GtkWidget *widget)
       &attributes, attributes_mask);
   gdk_window_set_user_data (widget->window, widget);
 
+  swfdec_gtk_widget_update_background (SWFDEC_GTK_WIDGET (widget));
   widget->style = gtk_style_attach (widget->style, widget->window);
 
   if (SWFDEC_GTK_WIDGET (widget)->priv->player) {
     swfdec_gtk_widget_update_cursor (SWFDEC_GTK_WIDGET (widget));
   }
+  swfdec_gtk_widget_update_renderer (SWFDEC_GTK_WIDGET (widget));
+}
+
+static void
+swfdec_gtk_widget_unrealize (GtkWidget *widget)
+{
+  GTK_WIDGET_CLASS (swfdec_gtk_widget_parent_class)->unrealize (widget);
+
+  swfdec_gtk_widget_update_renderer (SWFDEC_GTK_WIDGET (widget));
+}
+
+static void
+swfdec_gtk_widget_map (GtkWidget *gtkwidget)
+{
+  SwfdecGtkWidgetPrivate *priv = SWFDEC_GTK_WIDGET (gtkwidget)->priv;
+
+  g_assert (gdk_region_empty (priv->invalid));
+
+  GTK_WIDGET_CLASS (swfdec_gtk_widget_parent_class)->map (gtkwidget);
+}
+
+static void
+swfdec_gtk_widget_unmap (GtkWidget *gtkwidget)
+{
+  SwfdecGtkWidget *widget = SWFDEC_GTK_WIDGET (gtkwidget);
+
+  GTK_WIDGET_CLASS (swfdec_gtk_widget_parent_class)->unmap (gtkwidget);
+
+  swfdec_gtk_widget_clear_invalidations (widget);
+}
+
+static void
+swfdec_gtk_widget_style_set (GtkWidget *gtkwidget, GtkStyle *previous)
+{
+  /* do not set the window's background here */
 }
 
 static void
@@ -429,12 +616,15 @@ swfdec_gtk_widget_class_init (SwfdecGtkWidgetClass * g_class)
   object_class->get_property = swfdec_gtk_widget_get_property;
   object_class->set_property = swfdec_gtk_widget_set_property;
 
-  g_object_class_install_property (object_class, PROP_PLAYER,
-      g_param_spec_object ("player", "player", "player that is displayed",
-	  SWFDEC_TYPE_PLAYER, G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+  g_object_class_install_property (object_class, PROP_FULLSCREEN_MODE,
+      g_param_spec_boolean ("fullscreen-mode", "fullscreen mode", "special mode for fullscreen display",
+	  FALSE, G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
   g_object_class_install_property (object_class, PROP_INTERACTIVE,
       g_param_spec_boolean ("interactive", "interactive", "if mouse events are processed",
 	  TRUE, G_PARAM_READWRITE));
+  g_object_class_install_property (object_class, PROP_PLAYER,
+      g_param_spec_object ("player", "player", "player that is displayed",
+	  SWFDEC_TYPE_PLAYER, G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
   g_object_class_install_property (object_class, PROP_RENDERER_SET,
       g_param_spec_boolean ("renderer-set", "renderer set", "if an intermediate renderer should be used",
 	  TRUE, G_PARAM_READWRITE));
@@ -444,6 +634,10 @@ swfdec_gtk_widget_class_init (SwfdecGtkWidgetClass * g_class)
 	  0, G_MAXUINT, CAIRO_SURFACE_TYPE_IMAGE, G_PARAM_READWRITE));
 
   widget_class->realize = swfdec_gtk_widget_realize;
+  widget_class->unrealize = swfdec_gtk_widget_unrealize;
+  widget_class->map = swfdec_gtk_widget_map;
+  widget_class->unmap = swfdec_gtk_widget_unmap;
+  widget_class->style_set = swfdec_gtk_widget_style_set;
   widget_class->size_request = swfdec_gtk_widget_size_request;
   widget_class->size_allocate = swfdec_gtk_widget_size_allocate;
   widget_class->expose_event = swfdec_gtk_widget_expose;
@@ -453,39 +647,87 @@ swfdec_gtk_widget_class_init (SwfdecGtkWidgetClass * g_class)
   widget_class->leave_notify_event = swfdec_gtk_widget_leave_notify;
   widget_class->key_press_event = swfdec_gtk_widget_key_press;
   widget_class->key_release_event = swfdec_gtk_widget_key_release;
+  widget_class->focus_in_event = swfdec_gtk_widget_focus_inout;
+  widget_class->focus_out_event = swfdec_gtk_widget_focus_inout;
+  widget_class->focus = swfdec_gtk_widget_focus;
 
   g_type_class_add_private (object_class, sizeof (SwfdecGtkWidgetPrivate));
 }
 
 static void
-swfdec_gtk_widget_init (SwfdecGtkWidget * widget)
+swfdec_gtk_widget_init (SwfdecGtkWidget *widget)
 {
   SwfdecGtkWidgetPrivate *priv;
   
   priv = widget->priv = G_TYPE_INSTANCE_GET_PRIVATE (widget, SWFDEC_TYPE_GTK_WIDGET, SwfdecGtkWidgetPrivate);
 
   priv->interactive = TRUE;
-  priv->renderer = CAIRO_SURFACE_TYPE_IMAGE;
+  priv->renderer_type = CAIRO_SURFACE_TYPE_IMAGE;
+  priv->invalid = gdk_region_new ();
+
+  gtk_widget_set_double_buffered (GTK_WIDGET (widget), FALSE);
 
   GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_FOCUS);
 }
 
+static gboolean
+swfdec_gtk_widget_do_invalidate (gpointer widgetp)
+{
+  SwfdecGtkWidget *widget = widgetp;
+  SwfdecGtkWidgetPrivate *priv = widget->priv;
+
+  g_assert (GTK_WIDGET_REALIZED (widget));
+
+  gdk_window_invalidate_region (GTK_WIDGET (widget)->window, priv->invalid, FALSE);
+  swfdec_gtk_widget_clear_invalidations (widget);
+  return FALSE;
+}
+
 static void
-swfdec_gtk_widget_invalidate_cb (SwfdecPlayer *player, const SwfdecRectangle *extents,
+swfdec_gtk_widget_invalidate_cb (SwfdecPlayer *player,
     const SwfdecRectangle *rect, guint n_rects, SwfdecGtkWidget *widget)
 {
-  GdkRegion *region;
+  SwfdecGtkWidgetPrivate *priv = widget->priv;
   guint i;
 
-  if (!GTK_WIDGET_REALIZED (widget))
+  if (!GTK_WIDGET_MAPPED (widget) ||
+      priv->fullscreen_window != NULL)
     return;
 
-  region = gdk_region_new ();
   for (i = 0; i < n_rects; i++) {
-    gdk_region_union_with_rect (region, (GdkRectangle *) &rect[i]);
+    gdk_region_union_with_rect (priv->invalid, (GdkRectangle *) &rect[i]);
   }
-  gdk_window_invalidate_region (GTK_WIDGET (widget)->window, region, FALSE);
-  gdk_region_destroy (region);
+  if (priv->invalidator == 0) {
+    priv->invalidator = g_idle_add_full (SWFDEC_GTK_PRIORITY_REDRAW,
+	swfdec_gtk_widget_do_invalidate, widget, NULL);
+  }
+}
+
+static void
+swfdec_gtk_widget_do_fullscreen (SwfdecGtkWidget *widget, gboolean fullscreen)
+{
+  SwfdecGtkWidgetPrivate *priv = widget->priv;
+
+  if (priv->fullscreen_mode)
+    return;
+
+  if (fullscreen && priv->fullscreen_window == NULL) {
+    GtkWidget *window, *child;
+    
+    priv->fullscreen_window = window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+    child = swfdec_gtk_widget_new_fullscreen (priv->player);
+    gtk_container_add (GTK_CONTAINER (window), child);
+    gtk_widget_show_all (window);
+    gtk_window_fullscreen (GTK_WINDOW (window));
+    gtk_widget_grab_focus (child);
+    g_object_ref (window);
+  } else if (!fullscreen && priv->fullscreen_window != NULL) {
+    gtk_widget_destroy (priv->fullscreen_window);
+    g_object_unref (priv->fullscreen_window);
+    priv->fullscreen_window = NULL;
+  }
+  swfdec_gtk_widget_clear_invalidations (widget);
+  gtk_widget_queue_draw (GTK_WIDGET (widget));
 }
 
 static void
@@ -495,6 +737,17 @@ swfdec_gtk_widget_notify_cb (SwfdecPlayer *player, GParamSpec *pspec, SwfdecGtkW
     swfdec_gtk_widget_update_cursor (widget);
   } else if (g_str_equal (pspec->name, "initialized")) {
     gtk_widget_queue_resize (GTK_WIDGET (widget));
+  } else if (g_str_equal (pspec->name, "fullscreen")) {
+    swfdec_gtk_widget_do_fullscreen (widget, swfdec_player_get_fullscreen (player));
+  } else if (g_str_equal (pspec->name, "background-color")) {
+    swfdec_gtk_widget_update_background (widget);
+  } else if (g_str_equal (pspec->name, "selection")) {
+    const char *text = swfdec_player_get_selection (player);
+    if (GTK_WIDGET_HAS_FOCUS (widget) && text) {
+      GtkClipboard *clipboard = gtk_widget_get_clipboard (GTK_WIDGET (widget),
+	  GDK_SELECTION_PRIMARY);
+      gtk_clipboard_set_text (clipboard, text, -1);
+    }
   }
 }
 
@@ -510,28 +763,37 @@ swfdec_gtk_widget_notify_cb (SwfdecPlayer *player, GParamSpec *pspec, SwfdecGtkW
 void
 swfdec_gtk_widget_set_player (SwfdecGtkWidget *widget, SwfdecPlayer *player)
 {
-  SwfdecGtkWidgetPrivate *priv = widget->priv;
+  SwfdecGtkWidgetPrivate *priv;
 
   g_return_if_fail (SWFDEC_IS_GTK_WIDGET (widget));
   g_return_if_fail (player == NULL || SWFDEC_IS_PLAYER (player));
   
+  priv = widget->priv;
+  if (priv->player == player)
+    return;
+
+  swfdec_gtk_widget_clear_invalidations (widget);
+  if (player) {
+    g_signal_connect (player, "invalidate", G_CALLBACK (swfdec_gtk_widget_invalidate_cb), widget);
+    g_signal_connect (player, "notify", G_CALLBACK (swfdec_gtk_widget_notify_cb), widget);
+    g_object_ref (player);
+    swfdec_gtk_widget_update_cursor (widget);
+    swfdec_player_set_focus (player, GTK_WIDGET_HAS_FOCUS (GTK_WIDGET (widget)));
+  } else {
+    if (GTK_WIDGET (widget)->window)
+      gdk_window_set_cursor (GTK_WIDGET (widget)->window, NULL); 
+  }
   if (priv->player) {
     g_signal_handlers_disconnect_by_func (priv->player, swfdec_gtk_widget_invalidate_cb, widget);
     g_signal_handlers_disconnect_by_func (priv->player, swfdec_gtk_widget_notify_cb, widget);
     g_object_unref (priv->player);
   }
   priv->player = player;
-  if (player) {
-    g_signal_connect (player, "invalidate", G_CALLBACK (swfdec_gtk_widget_invalidate_cb), widget);
-    g_signal_connect (player, "notify", G_CALLBACK (swfdec_gtk_widget_notify_cb), widget);
-    g_object_ref (player);
-    swfdec_gtk_widget_update_cursor (widget);
-  } else {
-    if (GTK_WIDGET (widget)->window)
-      gdk_window_set_cursor (GTK_WIDGET (widget)->window, NULL); 
-  }
-  gtk_widget_queue_resize (GTK_WIDGET (widget));
   g_object_notify (G_OBJECT (widget), "player");
+  gtk_widget_queue_resize (GTK_WIDGET (widget));
+  swfdec_gtk_widget_update_renderer (widget);
+  swfdec_gtk_widget_update_background (widget);
+  swfdec_gtk_widget_do_fullscreen (widget, player ? swfdec_player_get_fullscreen (player) : FALSE);
 }
 
 /**
@@ -570,6 +832,29 @@ swfdec_gtk_widget_new (SwfdecPlayer *player)
   return GTK_WIDGET (widget);
 }
 
+/**
+ * swfdec_gtk_widget_new_fullscreen:
+ * @player: a #SwfdecPlayer
+ *
+ * Creates a new #SwfdecGtkWidget to display @player inside a fullscreen 
+ * window.
+ *
+ * Returns: the new widget that displays @player in fullscreen mode.
+ **/
+GtkWidget *
+swfdec_gtk_widget_new_fullscreen (SwfdecPlayer *player)
+{
+  SwfdecGtkWidget *widget;
+  
+  g_return_val_if_fail (SWFDEC_IS_PLAYER (player), NULL);
+
+  /* NB: We must set fullscreen mode before setting the player, or we 
+   * get inf-loops when setting the player creates fullscreen windows */
+  widget = g_object_new (SWFDEC_TYPE_GTK_WIDGET, "fullscreen-mode", TRUE, "player", player, NULL);
+
+  return GTK_WIDGET (widget);
+}
+
 /**
  * swfdec_gtk_widget_set_interactive:
  * @widget: a #SwfdecGtkWidget
@@ -620,12 +905,13 @@ swfdec_gtk_widget_set_renderer (SwfdecGtkWidget *widget, cairo_surface_type_t re
 {
   g_return_if_fail (SWFDEC_IS_GTK_WIDGET (widget));
 
-  widget->priv->renderer = renderer;
+  widget->priv->renderer_type = renderer;
   if (widget->priv->renderer_set == FALSE) {
     widget->priv->renderer_set = TRUE;
     g_object_notify (G_OBJECT (widget), "renderer-set");
   }
   g_object_notify (G_OBJECT (widget), "renderer");
+  swfdec_gtk_widget_update_renderer (widget);
 }
 
 /**
@@ -644,6 +930,7 @@ swfdec_gtk_widget_unset_renderer (SwfdecGtkWidget *widget)
     return;
   widget->priv->renderer_set = FALSE;
   g_object_notify (G_OBJECT (widget), "renderer-set");
+  swfdec_gtk_widget_update_renderer (widget);
 }
 
 /**
@@ -661,7 +948,7 @@ swfdec_gtk_widget_get_renderer (SwfdecGtkWidget *widget)
 {
   g_return_val_if_fail (SWFDEC_IS_GTK_WIDGET (widget), CAIRO_SURFACE_TYPE_IMAGE);
 
-  return widget->priv->renderer;
+  return widget->priv->renderer_type;
 }
 
 /**
