@@ -1,5 +1,5 @@
 /* Swfdec
- * Copyright (C) 2007 Benjamin Otte <otte@gnome.org>
+ * Copyright (C) 2007-2008 Benjamin Otte <otte@gnome.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -36,15 +36,14 @@
 #include "swfdec_as_stack.h"
 #include "swfdec_as_strings.h"
 #include "swfdec_as_types.h"
+#include "swfdec_constant_pool.h"
 #include "swfdec_debug.h"
+#include "swfdec_gc_object.h"
 #include "swfdec_internal.h" /* for swfdec_player_preinit_global() */
 #include "swfdec_script.h"
 
 /*** GARBAGE COLLECTION DOCS ***/
 
-#define SWFDEC_AS_GC_MARK (1 << 0)		/* only valid during GC */
-#define SWFDEC_AS_GC_ROOT (1 << 1)		/* for objects: rooted, for strings: static */
-
 /**
  * SECTION:Internals
  * @title: Internals and garbage collection
@@ -168,8 +167,8 @@ swfdec_as_context_abort (SwfdecAsContext *context, const char *reason)
 {
   g_return_if_fail (context);
 
-  SWFDEC_ERROR ("%s", reason);
   if (context->state != SWFDEC_AS_CONTEXT_ABORTED) {
+    SWFDEC_ERROR ("abort: %s", reason);
     context->state = SWFDEC_AS_CONTEXT_ABORTED;
     g_object_notify (G_OBJECT (context), "aborted");
   }
@@ -214,17 +213,20 @@ swfdec_as_context_try_use_mem (SwfdecAsContext *context, gsize bytes)
  * keeps track of the memory that script code consumes. If too much memory is
  * in use, this function may decide to stop the script engine with an out of
  * memory error.
- *
- * Returns: %TRUE if the memory could be allocated. %FALSE on OOM.
  **/
-gboolean
+void
 swfdec_as_context_use_mem (SwfdecAsContext *context, gsize bytes)
 {
-  g_return_val_if_fail (SWFDEC_IS_AS_CONTEXT (context), FALSE);
-  g_return_val_if_fail (bytes > 0, FALSE);
+  g_return_if_fail (SWFDEC_IS_AS_CONTEXT (context));
+  g_return_if_fail (bytes > 0);
 
   /* FIXME: Don't forget to abort on OOM */
-  return swfdec_as_context_try_use_mem (context, bytes);
+  if (!swfdec_as_context_try_use_mem (context, bytes)) {
+    swfdec_as_context_abort (context, "Out of memory");
+    /* add the memory anyway, as we're gonna make use of it. */
+    context->memory += bytes;
+    context->memory_since_gc += bytes;
+  }
 }
 
 /**
@@ -277,23 +279,18 @@ swfdec_as_context_remove_strings (gpointer key, gpointer value, gpointer data)
 static gboolean
 swfdec_as_context_remove_objects (gpointer key, gpointer value, gpointer debugger)
 {
-  SwfdecAsObject *object;
+  SwfdecGcObject *gc;
 
-  object = key;
+  gc = key;
   /* we only check for mark here, not root, since this works on destroy, too */
-  if (object->flags & SWFDEC_AS_GC_MARK) {
-    object->flags &= ~SWFDEC_AS_GC_MARK;
-    SWFDEC_LOG ("%s: %s %p", (object->flags & SWFDEC_AS_GC_ROOT) ? "rooted" : "marked",
-	G_OBJECT_TYPE_NAME (object), object);
+  if (gc->flags & SWFDEC_AS_GC_MARK) {
+    gc->flags &= ~SWFDEC_AS_GC_MARK;
+    SWFDEC_LOG ("%s: %s %p", (gc->flags & SWFDEC_AS_GC_ROOT) ? "rooted" : "marked",
+	G_OBJECT_TYPE_NAME (gc), gc);
     return FALSE;
   } else {
-    SWFDEC_LOG ("deleted: %s %p", G_OBJECT_TYPE_NAME (object), object);
-    if (debugger) {
-      SwfdecAsDebuggerClass *klass = SWFDEC_AS_DEBUGGER_GET_CLASS (debugger);
-      if (klass->remove)
-	klass->remove (debugger, object->context, object);
-    }
-    swfdec_as_object_collect (object);
+    SWFDEC_LOG ("deleted: %s %p", G_OBJECT_TYPE_NAME (gc), gc);
+    g_object_unref (gc);
     return TRUE;
   }
 }
@@ -310,28 +307,6 @@ swfdec_as_context_collect (SwfdecAsContext *context)
   SWFDEC_INFO (">> done collecting garbage");
 }
 
-/**
- * swfdec_as_object_mark:
- * @object: a #SwfdecAsObject
- *
- * Mark @object as being in use. Calling this function is only valid during
- * the marking phase of garbage collection.
- **/
-void
-swfdec_as_object_mark (SwfdecAsObject *object)
-{
-  SwfdecAsObjectClass *klass;
-
-  g_return_if_fail (SWFDEC_IS_AS_OBJECT (object));
-
-  if (object->flags & SWFDEC_AS_GC_MARK)
-    return;
-  object->flags |= SWFDEC_AS_GC_MARK;
-  klass = SWFDEC_AS_OBJECT_GET_CLASS (object);
-  g_assert (klass->mark);
-  klass->mark (object);
-}
-
 /**
  * swfdec_as_string_mark:
  * @string: a garbage-collected string
@@ -365,7 +340,7 @@ swfdec_as_value_mark (SwfdecAsValue *value)
   g_return_if_fail (SWFDEC_IS_AS_VALUE (value));
 
   if (SWFDEC_AS_VALUE_IS_OBJECT (value)) {
-    swfdec_as_object_mark (SWFDEC_AS_VALUE_GET_OBJECT (value));
+    swfdec_gc_object_mark (SWFDEC_AS_VALUE_GET_OBJECT (value));
   } else if (SWFDEC_AS_VALUE_IS_STRING (value)) {
     swfdec_as_string_mark (SWFDEC_AS_VALUE_GET_STRING (value));
   }
@@ -374,10 +349,24 @@ swfdec_as_value_mark (SwfdecAsValue *value)
 static void
 swfdec_as_context_mark_roots (gpointer key, gpointer value, gpointer data)
 {
-  SwfdecAsObject *object = key;
+  SwfdecGcObject *object = key;
 
   if ((object->flags & (SWFDEC_AS_GC_MARK | SWFDEC_AS_GC_ROOT)) == SWFDEC_AS_GC_ROOT)
-    swfdec_as_object_mark (object);
+    swfdec_gc_object_mark (object);
+}
+
+/* FIXME: replace this with refcounted strings? */
+static void 
+swfdec_as_context_mark_constant_pools (gpointer key, gpointer value, gpointer unused)
+{
+  SwfdecConstantPool *pool = value;
+  guint i;
+
+  for (i = 0; i < swfdec_constant_pool_size (pool); i++) {
+    const char *s = swfdec_constant_pool_get (pool, i);
+    if (s)
+      swfdec_as_string_mark (s);
+  }
 }
 
 static void
@@ -385,15 +374,16 @@ swfdec_as_context_do_mark (SwfdecAsContext *context)
 {
   /* This if is needed for SwfdecPlayer */
   if (context->global) {
-    swfdec_as_object_mark (context->global);
-    swfdec_as_object_mark (context->Function);
-    swfdec_as_object_mark (context->Function_prototype);
-    swfdec_as_object_mark (context->Object);
-    swfdec_as_object_mark (context->Object_prototype);
+    swfdec_gc_object_mark (context->global);
+    swfdec_gc_object_mark (context->Function);
+    swfdec_gc_object_mark (context->Function_prototype);
+    swfdec_gc_object_mark (context->Object);
+    swfdec_gc_object_mark (context->Object_prototype);
   }
   if (context->exception)
     swfdec_as_value_mark (&context->exception_value);
   g_hash_table_foreach (context->objects, swfdec_as_context_mark_roots, NULL);
+  g_hash_table_foreach (context->constant_pools, swfdec_as_context_mark_constant_pools, NULL);
 }
 
 /**
@@ -519,11 +509,16 @@ swfdec_as_context_dispose (GObject *object)
 
   while (context->stack)
     swfdec_as_stack_pop_segment (context);
+  /* We need to make sure there's no exception here. Otherwise collecting 
+   * frames that are inside a try block will assert */
+  swfdec_as_context_catch (context, NULL);
   swfdec_as_context_collect (context);
   if (context->memory != 0) {
     g_critical ("%zu bytes of memory left over\n", context->memory);
   }
   g_assert (g_hash_table_size (context->objects) == 0);
+  g_assert (g_hash_table_size (context->constant_pools) == 0);
+  g_hash_table_destroy (context->constant_pools);
   g_hash_table_destroy (context->objects);
   g_hash_table_destroy (context->strings);
   g_rand_free (context->rand);
@@ -583,6 +578,7 @@ swfdec_as_context_init (SwfdecAsContext *context)
 
   context->strings = g_hash_table_new (g_str_hash, g_str_equal);
   context->objects = g_hash_table_new (g_direct_hash, g_direct_equal);
+  context->constant_pools = g_hash_table_new (g_direct_hash, g_direct_equal);
 
   for (s = swfdec_as_strings; *s == '\2'; s += strlen (s) + 1) {
     g_hash_table_insert (context->strings, (char *) s + 1, (char *) s);
@@ -599,8 +595,10 @@ swfdec_as_context_create_string (SwfdecAsContext *context, const char *string, g
 {
   char *new;
   
-  if (!swfdec_as_context_use_mem (context, sizeof (char) * (2 + len)))
+  if (!swfdec_as_context_try_use_mem (context, sizeof (char) * (2 + len))) {
+    swfdec_as_context_abort (context, "Out of memory");
     return SWFDEC_AS_STR_EMPTY;
+  }
 
   new = g_slice_alloc (2 + len);
   memcpy (&new[1], string, len);
@@ -837,7 +835,9 @@ start:
     SwfdecAsValue rval = { 0, };
     if (frame->argc >= native->min_args && 
 	(native->type == 0 || 
-	 g_type_is_a (G_OBJECT_TYPE (frame->thisp), native->type))) {
+	 g_type_is_a (G_OBJECT_TYPE (frame->thisp), native->type)) &&
+	(!frame->construct || native->construct_type == 0 ||
+	 g_type_is_a (G_OBJECT_TYPE (frame->thisp), native->construct_type))) {
       SwfdecAsValue *argv;
       /* accumulate argv */
       if (frame->argc == 0 || frame->argv != NULL) {
@@ -868,6 +868,11 @@ start:
 	  argv, &rval);
       if (argv != frame->argv)
 	g_free (argv);
+    } else {
+      if (frame->construct && native->construct_type != 0 &&
+	 !g_type_is_a (G_OBJECT_TYPE (frame->thisp), native->construct_type)) {
+	SWFDEC_FIXME ("Ignoring call to native constructor with invalid type");
+      }
     }
     swfdec_as_frame_return (frame, &rval);
     goto start;
@@ -898,13 +903,17 @@ start:
       SWFDEC_ERROR ("pc %p not in valid range [%p, %p) anymore", pc, startpc, endpc);
       goto error;
     }
-    if (check_block && (pc < frame->block_start || pc >= frame->block_end)) {
+    while (check_block && (pc < frame->block_start || pc >= frame->block_end)) {
       SWFDEC_LOG ("code exited block");
       swfdec_as_frame_pop_block (frame);
       pc = frame->pc;
       if (frame != context->frame)
 	goto start;
+      if (context->exception)
+	break;
     }
+    if (context->exception)
+      continue;
 
     /* decode next action */
     action = *pc;
@@ -1005,183 +1014,6 @@ out:
   return;
 }
 
-/*** EVAL ***/
-
-static char *
-swfdec_as_slash_to_dot (const char *slash_str)
-{
-  const char *cur = slash_str;
-  GString *str = g_string_new ("");
-
-  if (*cur == '/') {
-    g_string_append (str, "_root");
-  } else {
-    goto start;
-  }
-  while (cur && *cur == '/') {
-    cur++;
-start:
-    if (str->len > 0)
-      g_string_append_c (str, '.');
-    if (cur[0] == '.' && cur[1] == '.') {
-      g_string_append (str, "_parent");
-      cur += 2;
-    } else {
-      char *slash = strchr (cur, '/');
-      if (slash) {
-	g_string_append_len (str, cur, slash - cur);
-	cur = slash;
-      } else {
-	g_string_append (str, cur);
-	cur = NULL;
-      }
-    }
-    /* cur should now point to the slash */
-  }
-  if (cur) {
-    if (*cur != '\0')
-      goto fail;
-  }
-  SWFDEC_DEBUG ("parsed slash-notated string \"%s\" into dot notation \"%s\"",
-      slash_str, str->str);
-  return g_string_free (str, FALSE);
-
-fail:
-  SWFDEC_WARNING ("failed to parse slash-notated string \"%s\" into dot notation", slash_str);
-  g_string_free (str, TRUE);
-  return NULL;
-}
-
-static void
-swfdec_as_context_eval_get_property (SwfdecAsContext *cx, 
-    SwfdecAsObject *obj, const char *name, SwfdecAsValue *ret)
-{
-  if (obj) {
-    swfdec_as_object_get_variable (obj, name, ret);
-  } else {
-    if (cx->frame) {
-      swfdec_as_frame_get_variable (cx->frame, name, ret);
-    } else {
-      SWFDEC_WARNING ("eval called without a frame");
-      swfdec_as_object_get_variable (cx->global, name, ret);
-    }
-  }
-}
-
-static void
-swfdec_as_context_eval_set_property (SwfdecAsContext *cx, 
-    SwfdecAsObject *obj, const char *name, const SwfdecAsValue *ret)
-{
-  if (obj == NULL) {
-    if (cx->frame == NULL) {
-      SWFDEC_ERROR ("no frame in eval_set?");
-      return;
-    }
-    swfdec_as_frame_set_variable (cx->frame, name, ret);
-  } else {
-    swfdec_as_object_set_variable (obj, name, ret);
-  }
-}
-
-static void
-swfdec_as_context_eval_internal (SwfdecAsContext *cx, SwfdecAsObject *obj, const char *str,
-        SwfdecAsValue *val, gboolean set)
-{
-  SwfdecAsValue cur;
-  char **varlist;
-  guint i;
-
-  SWFDEC_LOG ("eval called with \"%s\" on %p", str, obj);
-  if (strchr (str, '/')) {
-    char *work = swfdec_as_slash_to_dot (str);
-    if (!work) {
-      SWFDEC_AS_VALUE_SET_UNDEFINED (val);
-      return;
-    }
-    varlist = g_strsplit (work, ".", -1);
-    g_free (work);
-  } else {
-    varlist = g_strsplit (str, ".", -1);
-  }
-  for (i = 0; varlist[i] != NULL; i++) {
-    const char *dot = swfdec_as_context_get_string (cx, varlist[i]);
-    if (varlist[i+1] != NULL) {
-      swfdec_as_context_eval_get_property (cx, obj, dot, &cur);
-      if (!SWFDEC_AS_VALUE_IS_OBJECT (&cur)) {
-	SWFDEC_AS_VALUE_SET_UNDEFINED (&cur);
-	break;
-      }
-      obj = SWFDEC_AS_VALUE_GET_OBJECT (&cur);
-    } else {
-      if (set) {
-	swfdec_as_context_eval_set_property (cx, obj, dot, val);
-      } else {
-	swfdec_as_context_eval_get_property (cx, obj, dot, &cur);
-      }
-      goto finish;
-    }
-  }
-  if (obj == NULL) {
-    if (cx->frame)
-      obj = cx->frame->target;
-    else
-      obj = cx->global;
-  }
-  g_assert (obj != NULL);
-  SWFDEC_AS_VALUE_SET_OBJECT (&cur, obj);
-
-finish:
-  g_strfreev (varlist);
-  *val = cur;
-}
-
-/**
- * swfdec_as_context_eval:
- * @context: a #SwfdecAsContext
- * @obj: #SwfdecAsObject to use as source for evaluating or NULL for the 
- *       current frame's scope
- * @str: The string to evaluate
- * @val: location for the return value
- *
- * This function works like the Actionscript eval function used on @obj.
- * It handles both slash-style and dot-style notation. If an error occured
- * during evaluation, the return value will be the undefined value.
- **/
-void
-swfdec_as_context_eval (SwfdecAsContext *context, SwfdecAsObject *obj, const char *str, 
-    SwfdecAsValue *val)
-{
-  g_return_if_fail (SWFDEC_IS_AS_CONTEXT (context));
-  g_return_if_fail (obj == NULL || SWFDEC_IS_AS_OBJECT (obj));
-  g_return_if_fail (str != NULL);
-  g_return_if_fail (val != NULL);
-
-  swfdec_as_context_eval_internal (context, obj, str, val, FALSE);
-}
-
-/**
- * swfdec_as_context_eval_set:
- * @context: a #SwfdecAsContext
- * @obj: #SwfdecAsObject to use as source for evaluating or NULL for the 
- *       default object.
- * @str: The string to evaluate
- * @val: the value to set the variable to
- *
- * Sets the variable referenced by @str to @val. If @str does not reference 
- * a valid property, nothing happens.
- **/
-void
-swfdec_as_context_eval_set (SwfdecAsContext *context, SwfdecAsObject *obj, const char *str,
-    const SwfdecAsValue *val)
-{
-  g_return_if_fail (SWFDEC_IS_AS_CONTEXT (context));
-  g_return_if_fail (obj == NULL || SWFDEC_IS_AS_OBJECT (obj));
-  g_return_if_fail (str != NULL);
-  g_return_if_fail (val != NULL);
-
-  swfdec_as_context_eval_internal (context, obj, str, (SwfdecAsValue *) val, TRUE);
-}
-
 /*** AS CODE ***/
 
 static void
