@@ -470,10 +470,11 @@ swfdec_as_date_set_field (SwfdecAsContext *cx, SwfdecAsObject *object,
 	break;
       case FIELD_YEAR:
 	// NOTE: Test against double, not the integer
-	if (d >= 100 || d < 0)
-	  number -= 1900;
+	if (d >= 0 && d < 100)
+	  number += 1900;
 	// fall trough
       case FIELD_FULL_YEAR:
+	number -= 1900;
 	if (!isfinite (d)) {
 	  swfdec_as_date_set_brokentime_value (date, utc, field_offsets[field],
 	      cx, 0 - 1900);
@@ -757,12 +758,20 @@ swfdec_as_date_setTime (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   SwfdecAsDate *date;
+  double d;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_AS_DATE, &date, "");
 
-  if (argc > 0) {
-    swfdec_as_date_set_milliseconds_utc (date,
-	swfdec_as_value_to_integer (cx, &argv[0]));
+  if (argc > 0 &&
+      (cx->version > 6 || !SWFDEC_AS_VALUE_IS_UNDEFINED (&argv[0]))) {
+    d = swfdec_as_value_to_number (cx, &argv[0]);
+  } else {
+    d = NAN;
+  }
+  if (isfinite (d)) {
+    swfdec_as_date_set_milliseconds_utc (date, trunc (d));
+  } else {
+    swfdec_as_date_set_milliseconds_utc (date, NAN);
   }
 
   SWFDEC_AS_VALUE_SET_NUMBER (ret, date->milliseconds);
@@ -792,6 +801,9 @@ swfdec_as_date_setSeconds (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_SECONDS, FALSE);
+
+  if (argc > 1)
+    swfdec_as_date_setMilliseconds (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 128 + 14, swfdec_as_date_setUTCSeconds)
@@ -800,6 +812,9 @@ swfdec_as_date_setUTCSeconds (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_SECONDS, TRUE);
+
+  if (argc > 1)
+    swfdec_as_date_setUTCMilliseconds (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 13, swfdec_as_date_setMinutes)
@@ -808,6 +823,9 @@ swfdec_as_date_setMinutes (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_MINUTES, FALSE);
+
+  if (argc > 1)
+    swfdec_as_date_setSeconds (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 128 + 13, swfdec_as_date_setUTCMinutes)
@@ -816,6 +834,9 @@ swfdec_as_date_setUTCMinutes (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_MINUTES, TRUE);
+
+  if (argc > 1)
+    swfdec_as_date_setUTCSeconds (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 12, swfdec_as_date_setHours)
@@ -824,6 +845,9 @@ swfdec_as_date_setHours (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_HOURS, FALSE);
+
+  if (argc > 1)
+    swfdec_as_date_setMinutes (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 128 + 12, swfdec_as_date_setUTCHours)
@@ -832,6 +856,9 @@ swfdec_as_date_setUTCHours (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_HOURS, TRUE);
+
+  if (argc > 1)
+    swfdec_as_date_setUTCMinutes (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 11, swfdec_as_date_setDate)
@@ -858,6 +885,9 @@ swfdec_as_date_setMonth (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_MONTHS, FALSE);
+
+  if (argc > 1)
+    swfdec_as_date_setDate (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 128 + 10, swfdec_as_date_setUTCMonth)
@@ -866,6 +896,9 @@ swfdec_as_date_setUTCMonth (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_MONTHS, TRUE);
+
+  if (argc > 1)
+    swfdec_as_date_setUTCDate (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 20, swfdec_as_date_setYear)
@@ -874,6 +907,9 @@ swfdec_as_date_setYear (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *ret)
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_YEAR, FALSE);
+
+  if (argc > 1)
+    swfdec_as_date_setMonth (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 9, swfdec_as_date_setFullYear)
@@ -883,6 +919,9 @@ swfdec_as_date_setFullYear (SwfdecAsContext *cx, SwfdecAsObject *object,
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_FULL_YEAR,
       FALSE);
+
+  if (argc > 1)
+    swfdec_as_date_setMonth (cx, object, argc - 1, argv + 1, ret);
 }
 
 SWFDEC_AS_NATIVE (103, 128 + 9, swfdec_as_date_setUTCFullYear)
@@ -892,6 +931,9 @@ swfdec_as_date_setUTCFullYear (SwfdecAsContext *cx, SwfdecAsObject *object,
 {
   swfdec_as_date_set_field (cx, object, argc, argv, ret, FIELD_FULL_YEAR,
       TRUE);
+
+  if (argc > 1)
+    swfdec_as_date_setUTCMonth (cx, object, argc - 1, argv + 1, ret);
 }
 
 // Static methods
@@ -1016,10 +1058,7 @@ swfdec_as_date_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
 
   if (!cx->frame->construct) {
     SwfdecAsValue val;
-    if (!swfdec_as_context_use_mem (cx, sizeof (SwfdecAsDate)))
-      return;
-    object = g_object_new (SWFDEC_TYPE_AS_DATE, NULL);
-    swfdec_as_object_add (object, cx, sizeof (SwfdecAsDate));
+    object = g_object_new (SWFDEC_TYPE_AS_DATE, "context", cx, NULL);
     swfdec_as_object_get_variable (cx->global, SWFDEC_AS_STR_Date, &val);
     if (SWFDEC_AS_VALUE_IS_OBJECT (&val)) {
       swfdec_as_object_set_constructor (object,
@@ -1032,9 +1071,9 @@ swfdec_as_date_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
   date = SWFDEC_AS_DATE (object);
 
   /* FIXME: find a general solution here */
-  if (SWFDEC_IS_PLAYER (SWFDEC_AS_OBJECT (date)->context)) {
+  if (SWFDEC_IS_PLAYER (swfdec_gc_object_get_context (date))) {
     date->utc_offset =
-      SWFDEC_PLAYER (SWFDEC_AS_OBJECT (date)->context)->priv->system->utc_offset;
+      SWFDEC_PLAYER (swfdec_gc_object_get_context (date))->priv->system->utc_offset;
   }
 
   // don't accept arguments when not constructing
@@ -1054,8 +1093,11 @@ swfdec_as_date_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
     GTimeVal tv;
 
     swfdec_as_context_get_time (cx, &tv);
+    /* Use millisecond granularity here. Otherwise the value returned by 
+     * getTime() or toString() has a decimal point which breaks Flash files.
+     */
     swfdec_as_date_set_milliseconds_utc (date,
-	tv.tv_sec * 1000.0 + tv.tv_usec / 1000.0);
+	tv.tv_sec * 1000.0 + tv.tv_usec / 1000);
   }
   else if (argc == 1) // milliseconds from epoch, local
   {
@@ -1102,7 +1144,7 @@ swfdec_as_date_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
 	    &num)) {
 	brokentime.day_of_month = num;
       } else {
-	date->milliseconds = d;
+	date->milliseconds += d;
       }
     } else {
       brokentime.day_of_month = 1;
@@ -1113,7 +1155,7 @@ swfdec_as_date_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
 	    &num)) {
 	brokentime.hours = num;
       } else {
-	date->milliseconds = d;
+	date->milliseconds += d;
       }
     }
 
@@ -1122,7 +1164,7 @@ swfdec_as_date_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
 	    &num)) {
 	brokentime.minutes = num;
       } else {
-	date->milliseconds = d;
+	date->milliseconds += d;
       }
     }
 
@@ -1131,7 +1173,7 @@ swfdec_as_date_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
 	    &num)) {
 	brokentime.seconds = num;
       } else {
-	date->milliseconds = d;
+	date->milliseconds += d;
       }
     }
 
@@ -1146,7 +1188,7 @@ swfdec_as_date_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
 	    &num)) {
 	brokentime.milliseconds += num;
       } else {
-	date->milliseconds = d;
+	date->milliseconds += d;
       }
     }
 
@@ -1168,11 +1210,7 @@ swfdec_as_date_new (SwfdecAsContext *context, double milliseconds,
 
   g_return_val_if_fail (SWFDEC_IS_AS_CONTEXT (context), NULL);
 
-  if (!swfdec_as_context_use_mem (context, sizeof (SwfdecAsDate)))
-    return NULL;
-
-  ret = g_object_new (SWFDEC_TYPE_AS_DATE, NULL);
-  swfdec_as_object_add (ret, context, sizeof (SwfdecAsDate));
+  ret = g_object_new (SWFDEC_TYPE_AS_DATE, "context", context, NULL);
   swfdec_as_object_get_variable (context->global, SWFDEC_AS_STR_Date, &val);
   if (SWFDEC_AS_VALUE_IS_OBJECT (&val))
     swfdec_as_object_set_constructor (ret, SWFDEC_AS_VALUE_GET_OBJECT (&val));
