@@ -29,6 +29,7 @@
 #include "swfdec_as_strings.h"
 #include "swfdec_as_super.h"
 #include "swfdec_debug.h"
+#include "swfdec_movie.h"
 
 /**
  * SECTION:SwfdecAsFrame
@@ -86,7 +87,7 @@ swfdec_as_stack_iterator_init_arguments (SwfdecAsStackIterator *iter, SwfdecAsFr
     iter->current = NULL;
     return NULL;
   }
-  context = SWFDEC_AS_OBJECT (frame)->context;
+  context = swfdec_gc_object_get_context (frame);
   if (frame->argv) {
     iter->stack = NULL;
     iter->current = (SwfdecAsValue *) frame->argv;
@@ -133,7 +134,7 @@ swfdec_as_stack_iterator_init (SwfdecAsStackIterator *iter, SwfdecAsFrame *frame
   g_return_val_if_fail (iter != NULL, NULL);
   g_return_val_if_fail (SWFDEC_IS_AS_FRAME (frame), NULL);
 
-  context = SWFDEC_AS_OBJECT (frame)->context;
+  context = swfdec_gc_object_get_context (frame);
   iter->i = 0;
   stack = context->stack;
   if (context->frame == frame) {
@@ -278,13 +279,9 @@ swfdec_as_frame_dispose (GObject *object)
   /* clean up */
   g_slice_free1 (sizeof (SwfdecAsValue) * frame->n_registers, frame->registers);
   if (frame->constant_pool) {
-    swfdec_constant_pool_free (frame->constant_pool);
+    swfdec_constant_pool_unref (frame->constant_pool);
     frame->constant_pool = NULL;
   }
-  if (frame->constant_pool_buffer) {
-    swfdec_buffer_unref (frame->constant_pool_buffer);
-    frame->constant_pool_buffer = NULL;
-  }
   g_array_free (frame->blocks, TRUE);
   g_slist_free (frame->scope_chain);
   if (frame->script) {
@@ -296,27 +293,27 @@ swfdec_as_frame_dispose (GObject *object)
 }
 
 static void
-swfdec_as_frame_mark (SwfdecAsObject *object)
+swfdec_as_frame_mark (SwfdecGcObject *object)
 {
   SwfdecAsFrame *frame = SWFDEC_AS_FRAME (object);
   guint i;
 
   if (frame->next)
-    swfdec_as_object_mark (SWFDEC_AS_OBJECT (frame->next));
-  g_slist_foreach (frame->scope_chain, (GFunc) swfdec_as_object_mark, NULL);
+    swfdec_gc_object_mark (frame->next);
+  g_slist_foreach (frame->scope_chain, (GFunc) swfdec_gc_object_mark, NULL);
   if (frame->thisp)
-    swfdec_as_object_mark (frame->thisp);
+    swfdec_gc_object_mark (frame->thisp);
   if (frame->super)
-    swfdec_as_object_mark (frame->super);
-  swfdec_as_object_mark (frame->target);
-  swfdec_as_object_mark (frame->original_target);
+    swfdec_gc_object_mark (frame->super);
+  swfdec_gc_object_mark (frame->target);
+  swfdec_gc_object_mark (frame->original_target);
   if (frame->function)
-    swfdec_as_object_mark (SWFDEC_AS_OBJECT (frame->function));
+    swfdec_gc_object_mark (frame->function);
   for (i = 0; i < frame->n_registers; i++) {
     swfdec_as_value_mark (&frame->registers[i]);
   }
   /* don't mark argv, it's const, others have to take care of it */
-  SWFDEC_AS_OBJECT_CLASS (swfdec_as_frame_parent_class)->mark (object);
+  SWFDEC_GC_OBJECT_CLASS (swfdec_as_frame_parent_class)->mark (object);
 }
 
 static char *
@@ -359,11 +356,13 @@ static void
 swfdec_as_frame_class_init (SwfdecAsFrameClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
+  SwfdecGcObjectClass *gc_class = SWFDEC_GC_OBJECT_CLASS (klass);
   SwfdecAsObjectClass *asobject_class = SWFDEC_AS_OBJECT_CLASS (klass);
 
   object_class->dispose = swfdec_as_frame_dispose;
 
-  asobject_class->mark = swfdec_as_frame_mark;
+  gc_class->mark = swfdec_as_frame_mark;
+
   asobject_class->debug = swfdec_as_frame_debug;
 }
 
@@ -378,7 +377,7 @@ swfdec_as_frame_init (SwfdecAsFrame *frame)
 static void
 swfdec_as_frame_load (SwfdecAsFrame *frame)
 {
-  SwfdecAsContext *context = SWFDEC_AS_OBJECT (frame)->context;
+  SwfdecAsContext *context = swfdec_gc_object_get_context (frame);
 
   frame->stack_begin = context->cur;
   context->base = frame->stack_begin;
@@ -397,10 +396,7 @@ swfdec_as_frame_new (SwfdecAsContext *context, SwfdecScript *script)
   g_return_val_if_fail (script != NULL, NULL);
   
   size = sizeof (SwfdecAsFrame) + sizeof (SwfdecAsValue) * script->n_registers;
-  if (!swfdec_as_context_use_mem (context, size))
-    return NULL;
-  frame = g_object_new (SWFDEC_TYPE_AS_FRAME, NULL);
-  swfdec_as_object_add (SWFDEC_AS_OBJECT (frame), context, size);
+  frame = g_object_new (SWFDEC_TYPE_AS_FRAME, "context", context, NULL);
   frame->script = swfdec_script_ref (script);
   frame->function_name = script->name;
   SWFDEC_DEBUG ("new frame for function %s", frame->function_name);
@@ -409,12 +405,9 @@ swfdec_as_frame_new (SwfdecAsContext *context, SwfdecScript *script)
   frame->n_registers = script->n_registers;
   frame->registers = g_slice_alloc0 (sizeof (SwfdecAsValue) * frame->n_registers);
   if (script->constant_pool) {
-    frame->constant_pool_buffer = swfdec_buffer_ref (script->constant_pool);
-    frame->constant_pool = swfdec_constant_pool_new_from_action (
-	script->constant_pool->data, script->constant_pool->length, script->version);
-    if (frame->constant_pool) {
-      swfdec_constant_pool_attach_to_context (frame->constant_pool, context);
-    } else {
+    frame->constant_pool = swfdec_constant_pool_new (context, 
+	script->constant_pool, script->version);
+    if (frame->constant_pool == NULL) {
       SWFDEC_ERROR ("couldn't create constant pool");
     }
   }
@@ -431,11 +424,8 @@ swfdec_as_frame_new_native (SwfdecAsContext *context)
   g_return_val_if_fail (SWFDEC_IS_AS_CONTEXT (context), NULL);
   
   size = sizeof (SwfdecAsFrame);
-  if (!swfdec_as_context_use_mem (context, size))
-    return NULL;
-  frame = g_object_new (SWFDEC_TYPE_AS_FRAME, NULL);
+  frame = g_object_new (SWFDEC_TYPE_AS_FRAME, "context", context, NULL);
   SWFDEC_DEBUG ("new native frame");
-  swfdec_as_object_add (SWFDEC_AS_OBJECT (frame), context, size);
   swfdec_as_frame_load (frame);
   return frame;
 }
@@ -458,7 +448,7 @@ swfdec_as_frame_return (SwfdecAsFrame *frame, SwfdecAsValue *return_value)
   SwfdecAsFrame *next;
 
   g_return_if_fail (SWFDEC_IS_AS_FRAME (frame));
-  context = SWFDEC_AS_OBJECT (frame)->context;
+  context = swfdec_gc_object_get_context (frame);
   g_return_if_fail (frame == context->frame);
 
   /* save return value in case it was on the stack somewhere */
@@ -564,31 +554,42 @@ swfdec_as_frame_get_variable_and_flags (SwfdecAsFrame *frame, const char *variab
     SwfdecAsValue *value, guint *flags, SwfdecAsObject **pobject)
 {
   GSList *walk;
+  SwfdecAsContext *cx;
 
   g_return_val_if_fail (SWFDEC_IS_AS_FRAME (frame), NULL);
   g_return_val_if_fail (variable != NULL, NULL);
 
+  cx = swfdec_gc_object_get_context (frame);
+
   for (walk = frame->scope_chain; walk; walk = walk->next) {
     if (swfdec_as_object_get_variable_and_flags (walk->data, variable, value, 
 	  flags, pobject))
       return walk->data;
   }
   /* we've walked the scope chain down. Now look in the special objects. */
-  /* 1) the target */
-  if (swfdec_as_object_get_variable_and_flags (frame->target, variable, value, 
-	flags, pobject))
-    return frame->target;
+  /* 1) the target (if removed, use original target) */
+  if (SWFDEC_IS_MOVIE (frame->target) &&
+      SWFDEC_MOVIE(frame->target)->state < SWFDEC_MOVIE_STATE_DESTROYED) {
+    if (swfdec_as_object_get_variable_and_flags (frame->target, variable,
+	  value, flags, pobject))
+      return frame->target;
+  } else {
+    if (swfdec_as_object_get_variable_and_flags (frame->original_target,
+	  variable, value, flags, pobject))
+      return frame->original_target;
+  }
   /* 2) the global object */
-  if (swfdec_as_object_get_variable_and_flags (
-	SWFDEC_AS_OBJECT (frame)->context->global, variable, value, flags, pobject))
-    return SWFDEC_AS_OBJECT (frame)->context->global;
+  /* FIXME: ignored on version 4, but should it never be created instead? */
+  if (cx->version > 4 && swfdec_as_object_get_variable_and_flags (cx->global,
+	variable, value, flags, pobject))
+    return swfdec_gc_object_get_context (frame)->global;
 
   return NULL;
 }
 
 void
 swfdec_as_frame_set_variable_and_flags (SwfdecAsFrame *frame, const char *variable,
-    const SwfdecAsValue *value, guint default_flags)
+    const SwfdecAsValue *value, guint default_flags, gboolean overwrite, gboolean local)
 {
   SwfdecAsObject *pobject, *set;
   GSList *walk;
@@ -600,12 +601,24 @@ swfdec_as_frame_set_variable_and_flags (SwfdecAsFrame *frame, const char *variab
   for (walk = frame->scope_chain; walk; walk = walk->next) {
     if (swfdec_as_object_get_variable_and_flags (walk->data, variable, NULL, NULL, &pobject) &&
 	pobject == walk->data) {
+      if (!overwrite)
+	return;
       set = walk->data;
       break;
     }
   }
-  if (set == NULL)
-    set = frame->target;
+  if (set == NULL) {
+    if (local && frame->is_local) {
+      set = SWFDEC_AS_OBJECT (frame);
+    } else {
+      set = frame->target;
+    }
+  }
+
+  if (!overwrite) {
+    if (swfdec_as_object_get_variable (set, variable, NULL))
+      return;
+  }
 
   swfdec_as_object_set_variable_and_flags (set, variable, value, default_flags);
 }
@@ -630,7 +643,7 @@ swfdec_as_frame_delete_variable (SwfdecAsFrame *frame, const char *variable)
   if (ret)
     return ret;
   /* 2) the global object */
-  return swfdec_as_object_delete_variable (SWFDEC_AS_OBJECT (frame)->context->global, variable);
+  return swfdec_as_object_delete_variable (swfdec_gc_object_get_context (frame)->global, variable);
 }
 
 /**
@@ -671,7 +684,7 @@ swfdec_as_frame_preload (SwfdecAsFrame *frame)
 
   /* setup */
   object = SWFDEC_AS_OBJECT (frame);
-  context = object->context;
+  context = swfdec_gc_object_get_context (frame);
   script = frame->script;
   if (frame->script == NULL)
     goto out;
@@ -680,8 +693,6 @@ swfdec_as_frame_preload (SwfdecAsFrame *frame)
   if ((script->flags & (SWFDEC_SCRIPT_PRELOAD_ARGS | SWFDEC_SCRIPT_SUPPRESS_ARGS)) != SWFDEC_SCRIPT_SUPPRESS_ARGS) {
     SwfdecAsFrame *next;
     args = swfdec_as_array_new (context);
-    if (!args)
-      goto out;
     for (cur = swfdec_as_stack_iterator_init_arguments (&iter, frame); cur != NULL;
 	cur = swfdec_as_stack_iterator_next (&iter)) {
       swfdec_as_array_push (SWFDEC_AS_ARRAY (args), cur);
@@ -697,14 +708,16 @@ swfdec_as_frame_preload (SwfdecAsFrame *frame)
     } else {
       SWFDEC_AS_VALUE_SET_NULL (&val);
     }
-    swfdec_as_object_set_variable (args, SWFDEC_AS_STR_caller, &val);
+    swfdec_as_object_set_variable_and_flags (args, SWFDEC_AS_STR_caller, &val,
+	SWFDEC_AS_VARIABLE_HIDDEN | SWFDEC_AS_VARIABLE_PERMANENT);
 
     if (frame->function != NULL) {
       SWFDEC_AS_VALUE_SET_OBJECT (&val, SWFDEC_AS_OBJECT (frame->function));
     } else {
       SWFDEC_AS_VALUE_SET_NULL (&val);
     }
-    swfdec_as_object_set_variable (args, SWFDEC_AS_STR_callee, &val);
+    swfdec_as_object_set_variable_and_flags (args, SWFDEC_AS_STR_callee, &val,
+	SWFDEC_AS_VARIABLE_HIDDEN | SWFDEC_AS_VARIABLE_PERMANENT);
   } else {
     /* silence gcc */
     args = NULL;
@@ -811,7 +824,7 @@ swfdec_as_frame_handle_exception (SwfdecAsFrame *frame)
   SwfdecAsContext *cx;
 
   g_return_if_fail (SWFDEC_IS_AS_FRAME (frame));
-  cx = SWFDEC_AS_OBJECT (frame)->context;
+  cx = swfdec_gc_object_get_context (frame);
   g_return_if_fail (cx->exception);
 
   /* pop blocks in the hope that we are inside a Try block */
