@@ -491,6 +491,43 @@ const char swfdec_as_strings[] =
   SWFDEC_AS_CONSTANT_STRING ("height")
   SWFDEC_AS_CONSTANT_STRING ("deblocking")
   SWFDEC_AS_CONSTANT_STRING ("smoothing")
+  SWFDEC_AS_CONSTANT_STRING ("onKillFocus")
+  SWFDEC_AS_CONSTANT_STRING ("onSetFocus")
+  SWFDEC_AS_CONSTANT_STRING ("Selection")
+  SWFDEC_AS_CONSTANT_STRING ("tabEnabled")
+  SWFDEC_AS_CONSTANT_STRING ("focusEnabled")
+  SWFDEC_AS_CONSTANT_STRING ("ascent")
+  SWFDEC_AS_CONSTANT_STRING ("descent")
+  SWFDEC_AS_CONSTANT_STRING ("textFieldHeight")
+  SWFDEC_AS_CONSTANT_STRING ("textFieldWidth")
+  SWFDEC_AS_CONSTANT_STRING ("tabChildren")
+  SWFDEC_AS_CONSTANT_STRING ("tabIndex")
+  SWFDEC_AS_CONSTANT_STRING ("onChanged")
+  SWFDEC_AS_CONSTANT_STRING ("fullScreen")
+  SWFDEC_AS_CONSTANT_STRING ("onFullScreen")
+  SWFDEC_AS_CONSTANT_STRING ("_customHeaders")
+  SWFDEC_AS_CONSTANT_STRING ("Content-Type")
+  SWFDEC_AS_CONSTANT_STRING ("ll")
+  SWFDEC_AS_CONSTANT_STRING ("lr")
+  SWFDEC_AS_CONSTANT_STRING ("rl")
+  SWFDEC_AS_CONSTANT_STRING ("rr")
+  SWFDEC_AS_CONSTANT_STRING ("flash")
+  SWFDEC_AS_CONSTANT_STRING ("geom")
+  SWFDEC_AS_CONSTANT_STRING ("ColorTransform")
+  SWFDEC_AS_CONSTANT_STRING ("Transform")
+  SWFDEC_AS_CONSTANT_STRING ("bufferLength")
+  SWFDEC_AS_CONSTANT_STRING ("bufferTime")
+  SWFDEC_AS_CONSTANT_STRING ("audiocodec")
+  SWFDEC_AS_CONSTANT_STRING ("currentFps")
+  SWFDEC_AS_CONSTANT_STRING ("decodedFrames")
+  SWFDEC_AS_CONSTANT_STRING ("liveDelay")
+  SWFDEC_AS_CONSTANT_STRING ("videoCodec")
+  SWFDEC_AS_CONSTANT_STRING ("System")
+  SWFDEC_AS_CONSTANT_STRING ("__resolve")
+  SWFDEC_AS_CONSTANT_STRING ("Rectangle")
+  SWFDEC_AS_CONSTANT_STRING ("BitmapData")
+  SWFDEC_AS_CONSTANT_STRING ("tx")
+  SWFDEC_AS_CONSTANT_STRING ("ty")
   SWFDEC_AS_CONSTANT_STRING ("Invalid Date")
   /* add more here */
 ;
