@@ -1,6 +1,7 @@
 /* This is a generated file. See swfdec_as_strings.c for details. */
 #ifndef _SWFDEC_AS_STRINGS_H_
 #define _SWFDEC_AS_STRINGS_H_
+G_BEGIN_DECLS
 extern const char swfdec_as_strings[];
 #define SWFDEC_AS_STR_EMPTY &swfdec_as_strings[1]
 #define SWFDEC_AS_STR___proto__ &swfdec_as_strings[3]
@@ -470,4 +471,43 @@ extern const char swfdec_as_strings[];
 #define SWFDEC_AS_STR_height &swfdec_as_strings[4903]
 #define SWFDEC_AS_STR_deblocking &swfdec_as_strings[4911]
 #define SWFDEC_AS_STR_smoothing &swfdec_as_strings[4923]
+#define SWFDEC_AS_STR_onKillFocus &swfdec_as_strings[4934]
+#define SWFDEC_AS_STR_onSetFocus &swfdec_as_strings[4947]
+#define SWFDEC_AS_STR_Selection &swfdec_as_strings[4959]
+#define SWFDEC_AS_STR_tabEnabled &swfdec_as_strings[4970]
+#define SWFDEC_AS_STR_focusEnabled &swfdec_as_strings[4982]
+#define SWFDEC_AS_STR_ascent &swfdec_as_strings[4996]
+#define SWFDEC_AS_STR_descent &swfdec_as_strings[5004]
+#define SWFDEC_AS_STR_textFieldHeight &swfdec_as_strings[5013]
+#define SWFDEC_AS_STR_textFieldWidth &swfdec_as_strings[5030]
+#define SWFDEC_AS_STR_tabChildren &swfdec_as_strings[5046]
+#define SWFDEC_AS_STR_tabIndex &swfdec_as_strings[5059]
+#define SWFDEC_AS_STR_onChanged &swfdec_as_strings[5069]
+#define SWFDEC_AS_STR_fullScreen &swfdec_as_strings[5080]
+#define SWFDEC_AS_STR_onFullScreen &swfdec_as_strings[5092]
+#define SWFDEC_AS_STR__customHeaders &swfdec_as_strings[5106]
+#define SWFDEC_AS_STR_Content_Type &swfdec_as_strings[5122]
+#define SWFDEC_AS_STR_ll &swfdec_as_strings[5136]
+#define SWFDEC_AS_STR_lr &swfdec_as_strings[5140]
+#define SWFDEC_AS_STR_rl &swfdec_as_strings[5144]
+#define SWFDEC_AS_STR_rr &swfdec_as_strings[5148]
+#define SWFDEC_AS_STR_flash &swfdec_as_strings[5152]
+#define SWFDEC_AS_STR_geom &swfdec_as_strings[5159]
+#define SWFDEC_AS_STR_ColorTransform &swfdec_as_strings[5165]
+#define SWFDEC_AS_STR_Transform &swfdec_as_strings[5181]
+#define SWFDEC_AS_STR_bufferLength &swfdec_as_strings[5192]
+#define SWFDEC_AS_STR_bufferTime &swfdec_as_strings[5206]
+#define SWFDEC_AS_STR_audiocodec &swfdec_as_strings[5218]
+#define SWFDEC_AS_STR_currentFps &swfdec_as_strings[5230]
+#define SWFDEC_AS_STR_decodedFrames &swfdec_as_strings[5242]
+#define SWFDEC_AS_STR_liveDelay &swfdec_as_strings[5257]
+#define SWFDEC_AS_STR_videoCodec &swfdec_as_strings[5268]
+#define SWFDEC_AS_STR_System &swfdec_as_strings[5280]
+#define SWFDEC_AS_STR___resolve &swfdec_as_strings[5288]
+#define SWFDEC_AS_STR_Rectangle &swfdec_as_strings[5299]
+#define SWFDEC_AS_STR_BitmapData &swfdec_as_strings[5310]
+#define SWFDEC_AS_STR_tx &swfdec_as_strings[5322]
+#define SWFDEC_AS_STR_ty &swfdec_as_strings[5326]
+#define SWFDEC_AS_STR_Invalid_Date &swfdec_as_strings[5330]
+G_END_DECLS
 #endif
