@@ -1,5 +1,5 @@
 /* Swfdec
- * Copyright (C) 2007 Benjamin Otte <otte@gnome.org>
+ * Copyright (C) 2007-2008 Benjamin Otte <otte@gnome.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -34,6 +34,7 @@
 #include "swfdec_as_strings.h"
 #include "swfdec_as_super.h"
 #include "swfdec_debug.h"
+#include "swfdec_internal.h"
 #include "swfdec_movie.h"
 
 /*** GTK-DOC ***/
@@ -488,7 +489,7 @@ swfdec_as_value_to_number (SwfdecAsContext *context, const SwfdecAsValue *value)
 	// FIXME: We should most likely copy Tamarin's code here (MathUtils.cpp)
 	s = SWFDEC_AS_VALUE_GET_STRING (&tmp);
 	if (s == SWFDEC_AS_STR_EMPTY)
-	  return NAN;
+	  return (context->version >= 5) ? NAN : 0.0;
 	if (context->version > 5 && s[0] == '0' &&
 	    (s[1] == 'x' || s[1] == 'X')) {
 	  d = g_ascii_strtoll (s + 2, &end, 16);
@@ -497,16 +498,16 @@ swfdec_as_value_to_number (SwfdecAsContext *context, const SwfdecAsValue *value)
 	  d = g_ascii_strtoll (s, &end, 8);
 	} else {
 	  if (strpbrk (s, "xXiI") != NULL)
-	    return NAN;
+	    return (context->version >= 5) ? NAN : 0.0;
 	  d = g_ascii_strtod (s, &end);
 	}
-	if (*end == '\0')
+	if (*end == '\0' || context->version < 5)
 	  return d == -0.0 ? 0.0 : d;
 	else
 	  return NAN;
       }
     case SWFDEC_AS_TYPE_OBJECT:
-      return NAN;
+      return (context->version >= 5) ? NAN : 0.0;
     case SWFDEC_AS_TYPE_INT:
     default:
       g_assert_not_reached ();
@@ -678,3 +679,30 @@ swfdec_as_value_to_primitive (SwfdecAsValue *value)
   }
 }
 
+/* from swfdec_internal.h */
+gboolean
+swfdec_as_value_to_twips (SwfdecAsContext *context, const SwfdecAsValue *val, 
+    gboolean is_length, SwfdecTwips *result)
+{
+  double d;
+
+  g_return_val_if_fail (SWFDEC_IS_AS_CONTEXT (context), FALSE);
+  g_return_val_if_fail (val != NULL, FALSE);
+  g_return_val_if_fail (result != NULL, FALSE);
+
+  if (SWFDEC_AS_VALUE_IS_UNDEFINED (val) || SWFDEC_AS_VALUE_IS_NULL (val))
+    return FALSE;
+
+  d = swfdec_as_value_to_number (context, val);
+  if (isnan (d))
+    return FALSE;
+  if (is_length && d < 0)
+    return FALSE;
+
+  d *= SWFDEC_TWIPS_SCALE_FACTOR;
+  *result = swfdec_as_double_to_integer (d);
+  if (is_length)
+    *result = ABS (*result);
+  return TRUE;
+}
+
