@@ -46,11 +46,14 @@ struct _SwfdecColorTransform {
 #error "Unknown byte order"
 #endif
 
-#define SWFDEC_COLOR_COMBINE(r,g,b,a)	(((a)<<24) | ((r)<<16) | ((g)<<8) | (b))
-#define SWFDEC_COLOR_A(x)		(((x)>>24)&0xff)
-#define SWFDEC_COLOR_R(x)		(((x)>>16)&0xff)
-#define SWFDEC_COLOR_G(x)		(((x)>>8)&0xff)
-#define SWFDEC_COLOR_B(x)		((x)&0xff)
+#define SWFDEC_COLOR_COMBINE(r,g,b,a)	((SwfdecColor) (((a)<<24) | ((r)<<16) | ((g)<<8) | (b)))
+#define SWFDEC_COLOR_OPAQUE(color)	((SwfdecColor) ((color) | SWFDEC_COLOR_COMBINE (0, 0, 0, 0xFF)))
+#define SWFDEC_COLOR_ALPHA(x)		(((x)>>24)&0xff)
+#define SWFDEC_COLOR_RED(x)		(((x)>>16)&0xff)
+#define SWFDEC_COLOR_GREEN(x)		(((x)>>8)&0xff)
+#define SWFDEC_COLOR_BLUE(x)		((x)&0xff)
+
+#define SWFDEC_COLOR_WHITE		SWFDEC_COLOR_COMBINE (0xFF, 0xFF, 0xFF, 0xFF)
 
 SwfdecColor swfdec_color_apply_morph (SwfdecColor start, SwfdecColor end, guint ratio);
 void swfdec_color_set_source (cairo_t *cr, SwfdecColor color);
@@ -58,6 +61,7 @@ void swfdec_color_transform_init_identity (SwfdecColorTransform * trans);
 void swfdec_color_transform_init_mask (SwfdecColorTransform * trans);
 void swfdec_color_transform_init_color (SwfdecColorTransform *trans, SwfdecColor color);
 gboolean swfdec_color_transform_is_identity (const SwfdecColorTransform * trans);
+gboolean swfdec_color_transform_is_alpha (const SwfdecColorTransform * trans);
 #define swfdec_color_transform_is_mask(trans) ((trans)->mask)
 void swfdec_color_transform_chain (SwfdecColorTransform *dest,
     const SwfdecColorTransform *last, const SwfdecColorTransform *first);
