@@ -24,6 +24,7 @@
 
 #include <string.h>
 #include "swfdec_load_object.h"
+#include "swfdec_access.h"
 #include "swfdec_as_frame_internal.h"
 #include "swfdec_as_strings.h"
 #include "swfdec_debug.h"
@@ -37,7 +38,7 @@
 static SwfdecPlayer *
 swfdec_load_object_stream_target_get_player (SwfdecStreamTarget *target)
 {
-  return SWFDEC_PLAYER (SWFDEC_LOAD_OBJECT (target)->target->context);
+  return SWFDEC_PLAYER (swfdec_gc_object_get_context (SWFDEC_LOAD_OBJECT (target)->target));
 }
 
 static gboolean
@@ -75,7 +76,7 @@ swfdec_load_object_stream_target_error (SwfdecStreamTarget *target,
 
   /* unroot */
   swfdec_player_unroot (SWFDEC_PLAYER (
-	SWFDEC_AS_OBJECT (load_object->sandbox)->context), load_object);
+	swfdec_gc_object_get_context (load_object->sandbox)), load_object);
 }
 
 static void
@@ -97,7 +98,7 @@ swfdec_load_object_stream_target_close (SwfdecStreamTarget *target,
   swfdec_sandbox_use (load_object->sandbox);
   if (text != NULL) {
     load_object->finish (load_object->target, 
-	swfdec_as_context_give_string (load_object->target->context, text));
+	swfdec_as_context_give_string (swfdec_gc_object_get_context (load_object->target), text));
   } else {
     load_object->finish (load_object->target, SWFDEC_AS_STR_EMPTY);
   }
@@ -105,7 +106,7 @@ swfdec_load_object_stream_target_close (SwfdecStreamTarget *target,
 
   /* unroot */
   swfdec_player_unroot (SWFDEC_PLAYER (
-	SWFDEC_AS_OBJECT (load_object->sandbox)->context), load_object);
+	swfdec_gc_object_get_context (load_object->sandbox)), load_object);
 }
 
 static void
@@ -136,6 +137,8 @@ swfdec_load_object_dispose (GObject *object)
     swfdec_buffer_unref (load->buffer);
     load->buffer = NULL;
   }
+  g_strfreev (load->header_names);
+  g_strfreev (load->header_values);
 
   G_OBJECT_CLASS (swfdec_load_object_parent_class)->dispose (object);
 }
@@ -172,55 +175,30 @@ swfdec_load_object_load (SwfdecPlayer *player, gboolean allow, gpointer obj)
     return;
   }
 
-  load->loader = swfdec_player_load (player, load->url, load->request, load->buffer);
+  load->loader = swfdec_player_load_with_headers (player, load->url,
+      load->buffer, load->header_count, (const char **)load->header_names,
+      (const char **)load->header_values);
 
   swfdec_stream_set_target (SWFDEC_STREAM (load->loader), SWFDEC_STREAM_TARGET (load));
   swfdec_loader_set_data_type (load->loader, SWFDEC_LOADER_DATA_TEXT);
 }
 
+static const SwfdecAccessMatrix swfdec_load_object_matrix = {
+  { SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO },
+  { SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_YES, SWFDEC_ACCESS_POLICY },
+  { SWFDEC_ACCESS_YES, SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO },
+  { SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_POLICY },
+  { SWFDEC_ACCESS_YES, SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_POLICY }
+};
+
 /* perform security checks */
 static void
 swfdec_load_object_request (gpointer objectp, gpointer playerp)
 {
-  SwfdecLoadObject *load = SWFDEC_LOAD_OBJECT (objectp);
-  SwfdecPlayer *player = SWFDEC_PLAYER (playerp);
-  SwfdecURL *url;
+  SwfdecLoadObject *load = objectp;
 
-  if (swfdec_url_path_is_relative (load->url)) {
-    swfdec_load_object_load (player, 
-	load->sandbox->type != SWFDEC_SANDBOX_LOCAL_NETWORK, load);
-    return;
-  }
-  url = swfdec_player_create_url (player, load->url);
-  if (url == NULL) {
-    swfdec_load_object_load (player, FALSE, load);
-    return;
-  }
-  switch (load->sandbox->type) {
-    case SWFDEC_SANDBOX_REMOTE:
-    case SWFDEC_SANDBOX_LOCAL_NETWORK:
-    case SWFDEC_SANDBOX_LOCAL_TRUSTED:
-      if (swfdec_url_is_local (url)) {
-	swfdec_load_object_load (player, load->sandbox->type == SWFDEC_SANDBOX_LOCAL_TRUSTED, load);
-      } else {
-	SwfdecURL *load_url = swfdec_url_new_components (
-	    swfdec_url_get_protocol (url), swfdec_url_get_host (url), 
-	    swfdec_url_get_port (url), "crossdomain.xml", NULL);
-	swfdec_player_allow_or_load (player, url, load_url,
-	  swfdec_load_object_load, load);
-	swfdec_url_free (load_url);
-      }
-      break;
-    case SWFDEC_SANDBOX_LOCAL_FILE:
-      swfdec_load_object_load (player, swfdec_url_is_local (url), load);
-      break;
-    case SWFDEC_SANDBOX_NONE:
-    default:
-      g_assert_not_reached ();
-      break;
-  }
-
-  swfdec_url_free (url);
+  swfdec_player_allow_by_matrix (playerp, load->sandbox, load->url,
+      swfdec_load_object_matrix, swfdec_load_object_load, load);
 }
 
 static void
@@ -228,32 +206,37 @@ swfdec_load_object_mark (gpointer object, gpointer player)
 {
   SwfdecLoadObject *load = object;
 
-  swfdec_as_object_mark (SWFDEC_AS_OBJECT (load->sandbox));
+  swfdec_gc_object_mark (load->sandbox);
   if (load->url)
     swfdec_as_string_mark (load->url);
-  swfdec_as_object_mark (load->target);
+  swfdec_gc_object_mark (load->target);
 }
 
 void
 swfdec_load_object_create (SwfdecAsObject *target, const char *url,
-    SwfdecLoaderRequest request, SwfdecBuffer *data,
-    SwfdecLoadObjectProgress progress, SwfdecLoadObjectFinish finish)
+    SwfdecBuffer *data, guint header_count, char **header_names,
+    char **header_values, SwfdecLoadObjectProgress progress,
+    SwfdecLoadObjectFinish finish)
 {
   SwfdecPlayer *player;
   SwfdecLoadObject *load;
 
   g_return_if_fail (SWFDEC_IS_AS_OBJECT (target));
   g_return_if_fail (url != NULL);
+  g_return_if_fail (header_count == 0 || header_names != NULL);
+  g_return_if_fail (header_count == 0 || header_values != NULL);
   g_return_if_fail (finish != NULL);
 
-  player = SWFDEC_PLAYER (target->context);
+  player = SWFDEC_PLAYER (swfdec_gc_object_get_context (target));
   load = g_object_new (SWFDEC_TYPE_LOAD_OBJECT, NULL);
   swfdec_player_root_full (player, load, swfdec_load_object_mark, g_object_unref);
 
   load->target = target;
   load->url = url;
-  load->request = request;
   load->buffer = data;
+  load->header_count = header_count;
+  load->header_names = header_names;
+  load->header_values = header_values;
   load->progress = progress;
   load->finish = finish;
   /* get the current security */
