@@ -23,16 +23,20 @@
 
 #include <math.h>
 #include "swfdec_net_stream.h"
+#include "swfdec_access.h"
 #include "swfdec_amf.h"
 #include "swfdec_as_frame_internal.h"
 #include "swfdec_as_strings.h"
 #include "swfdec_audio_flv.h"
+#include "swfdec_cached_video.h"
 #include "swfdec_debug.h"
 #include "swfdec_loader_internal.h"
 #include "swfdec_player_internal.h"
+#include "swfdec_renderer_internal.h"
 #include "swfdec_resource.h"
 #include "swfdec_sandbox.h"
 #include "swfdec_stream_target.h"
+#include "swfdec_video_provider.h"
 
 /* NB: code and level must be rooted gc-strings */
 static void
@@ -42,11 +46,7 @@ swfdec_net_stream_onstatus (SwfdecNetStream *stream, const char *code, const cha
   SwfdecAsObject *object;
 
   swfdec_sandbox_use (stream->sandbox);
-  object = swfdec_as_object_new (SWFDEC_AS_OBJECT (stream)->context);
-  if (!object) {
-    swfdec_sandbox_unuse (stream->sandbox);
-    return;
-  }
+  object = swfdec_as_object_new (swfdec_gc_object_get_context (stream));
   SWFDEC_INFO ("emitting onStatus for %s %s", level, code);
   SWFDEC_AS_VALUE_SET_STRING (&val, code);
   swfdec_as_object_set_variable (object, SWFDEC_AS_STR_code, &val);
@@ -54,40 +54,55 @@ swfdec_net_stream_onstatus (SwfdecNetStream *stream, const char *code, const cha
   swfdec_as_object_set_variable (object, SWFDEC_AS_STR_level, &val);
 
   SWFDEC_AS_VALUE_SET_OBJECT (&val, object);
-  swfdec_as_object_call (SWFDEC_AS_OBJECT (stream), SWFDEC_AS_STR_onStatus, 1, &val, NULL);
+  if (!swfdec_as_object_call (SWFDEC_AS_OBJECT (stream),
+        SWFDEC_AS_STR_onStatus, 1, &val, NULL)) {
+    // if it's an error message and the stream object didn't have onStatus
+    // handler, call System.onStatus
+    if (level == SWFDEC_AS_STR_error) {
+      SwfdecAsValue system;
+
+      swfdec_as_object_get_variable (swfdec_gc_object_get_context (stream)->global,
+          SWFDEC_AS_STR_System, &system);
+      if (SWFDEC_AS_VALUE_IS_OBJECT (&system)) {
+        swfdec_as_object_call (SWFDEC_AS_VALUE_GET_OBJECT (&system),
+              SWFDEC_AS_STR_onStatus, 1, &val, NULL);
+      }
+    }
+  }
   swfdec_sandbox_unuse (stream->sandbox);
 }
 
-static cairo_surface_t *
-swfdec_net_stream_decode_video (SwfdecNetStream *stream, SwfdecBuffer *buffer)
+static void
+swfdec_net_stream_decode_video (SwfdecVideoDecoder *decoder, SwfdecBuffer *buffer)
 {
-  SwfdecVideoDecoder *decoder = stream->decoder;
-  cairo_surface_t *surface;
-
-  if (decoder == NULL)
-    return NULL;
-
   if (decoder->codec == SWFDEC_VIDEO_CODEC_VP6 ||
       decoder->codec == SWFDEC_VIDEO_CODEC_VP6_ALPHA) {
+    /* FIXME: This is somewhat nasty as we modify values in the decoder 
+     * directly. I know the current decoders don't mind, but if we expose 
+     * the decoder API... */
     guint wsub, hsub;
     SwfdecBuffer *tmp;
     if (buffer->length == 0) {
-      SWFDEC_ERROR ("0-byte VP6 video image buffer?");
-      return NULL;
+      swfdec_video_decoder_error (decoder, "0-byte VP6 video image buffer?");
+      return;
     }
     wsub = *buffer->data;
     hsub = wsub & 0xF;
     wsub >>= 4;
     tmp = swfdec_buffer_new_subbuffer (buffer, 1, buffer->length - 1);
-    surface = swfdec_video_decoder_decode (decoder, tmp);
+    swfdec_video_decoder_decode (decoder, tmp);
     swfdec_buffer_unref (tmp);
-    if (hsub || wsub) {
-      SWFDEC_FIXME ("need to subtract %ux%u pixels", wsub, hsub);
+    if (hsub >= decoder->height || wsub >= decoder->width) {
+      SWFDEC_ERROR ("can't reduce size by more than available");
+      decoder->width = 0;
+      decoder->height = 0;
+    } else {
+      decoder->width -= wsub;
+      decoder->height -= hsub;
     }
   } else {
-    surface = swfdec_video_decoder_decode (decoder, buffer);
+    swfdec_video_decoder_decode (decoder, buffer);
   }
-  return surface;
 }
 
 static void swfdec_net_stream_update_playing (SwfdecNetStream *stream);
@@ -119,19 +134,65 @@ swfdec_net_stream_video_goto (SwfdecNetStream *stream, guint timestamp)
   if (buffer == NULL) {
     SWFDEC_ERROR ("got no buffer - no video available?");
   } else {
-    if (format != stream->format) {
-      if (stream->decoder)
-	swfdec_video_decoder_free (stream->decoder);
-      stream->format = format;
-      stream->decoder = swfdec_video_decoder_new (format);
+    guint next;
+
+    if (stream->decoder && swfdec_video_decoder_get_codec (stream->decoder) != format) {
+      g_object_unref (stream->decoder);
+      stream->decoder = NULL;
     }
-    stream->surface = swfdec_net_stream_decode_video (stream, buffer);
-    if (stream->surface) {
-      GList *walk;
-      for (walk = stream->movies; walk; walk = walk->next) {
-	swfdec_video_movie_new_image (walk->data);
+    if (stream->decoder != NULL &&
+	(stream->decoder_time >= stream->current_time)) {
+      g_object_unref (stream->decoder);
+      stream->decoder = NULL;
+    }
+
+    if (stream->decoder == NULL) {
+      buffer = swfdec_flv_decoder_get_video (stream->flvdecoder, 
+	  stream->current_time, TRUE, &format, &stream->decoder_time,
+	  &next);
+      stream->decoder = swfdec_video_decoder_new (format);
+    } else {
+      swfdec_flv_decoder_get_video (stream->flvdecoder, 
+	  stream->decoder_time, FALSE, NULL, NULL, &next);
+      if (next != stream->current_time) {
+	guint key_time, key_next;
+	buffer = swfdec_flv_decoder_get_video (stream->flvdecoder, 
+	    stream->current_time, TRUE, &format, &key_time, &key_next);
+	if (key_time > stream->decoder_time) {
+	  stream->decoder_time = key_time;
+	  next = key_next;
+	} else {
+	  buffer = swfdec_flv_decoder_get_video (stream->flvdecoder, 
+	      next, FALSE, &format, &stream->decoder_time,
+	      &next);
+	}
+      } else {
+	buffer = swfdec_flv_decoder_get_video (stream->flvdecoder, 
+	    next, FALSE, &format, &stream->decoder_time,
+	    &next);
       }
     }
+
+    /* the following things hold:
+     * buffer: next buffer to decode
+     * format: format of that buffer
+     * stream->decoder_time: timestamp of buffer to decode
+     * stream->decoder: non-null, using stream->format
+     */
+    for (;;) {
+      if (format != swfdec_video_decoder_get_codec (stream->decoder)) {
+	g_object_unref (stream->decoder);
+	stream->decoder = swfdec_video_decoder_new (format);
+      }
+      swfdec_net_stream_decode_video (stream->decoder, buffer);
+      if (stream->decoder_time >= stream->current_time)
+	break;
+
+      buffer = swfdec_flv_decoder_get_video (stream->flvdecoder,
+	  next, FALSE, &format, &stream->decoder_time, &next);
+    }
+
+    swfdec_video_provider_new_image (SWFDEC_VIDEO_PROVIDER (stream));
   }
   if (stream->next_time <= stream->current_time) {
     if (swfdec_flv_decoder_is_eof (stream->flvdecoder)) {
@@ -154,7 +215,7 @@ swfdec_net_stream_video_goto (SwfdecNetStream *stream, guint timestamp)
       process_events_from++; /* increase so we get the next event next time */
       swfdec_bits_init (&bits, event);
       swfdec_sandbox_use (stream->sandbox);
-      if (swfdec_amf_parse (SWFDEC_AS_OBJECT (stream)->context, &bits, 2, 
+      if (swfdec_amf_parse (swfdec_gc_object_get_context (stream), &bits, 2, 
 	    SWFDEC_AMF_STRING, &name, SWFDEC_AMF_MIXED_ARRAY, &value) != 2) {
 	SWFDEC_ERROR ("could not parse data tag");
       } else {
@@ -179,14 +240,14 @@ swfdec_net_stream_timeout (SwfdecTimeout *timeout)
       stream->timeout.callback) {
     SWFDEC_LOG ("readding timeout");
     stream->timeout.timestamp += SWFDEC_MSECS_TO_TICKS (stream->next_time - stream->current_time);
-    swfdec_player_add_timeout (SWFDEC_PLAYER (SWFDEC_AS_OBJECT (stream)->context), &stream->timeout);
+    swfdec_player_add_timeout (SWFDEC_PLAYER (swfdec_gc_object_get_context (stream)), &stream->timeout);
   }
 }
 
 static void
 swfdec_net_stream_update_playing (SwfdecNetStream *stream)
 {
-  SwfdecPlayer *player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (stream)->context);
+  SwfdecPlayer *player = SWFDEC_PLAYER (swfdec_gc_object_get_context (stream));
   gboolean should_play;
     
   should_play = stream->playing; /* checks user-set play/pause */
@@ -226,7 +287,7 @@ swfdec_net_stream_update_playing (SwfdecNetStream *stream)
 static SwfdecPlayer *
 swfdec_net_stream_stream_target_get_player (SwfdecStreamTarget *target)
 {
-  return SWFDEC_PLAYER (SWFDEC_AS_OBJECT (target)->context);
+  return SWFDEC_PLAYER (swfdec_gc_object_get_context (target));
 }
 
 static void
@@ -271,7 +332,8 @@ swfdec_net_stream_stream_target_parse (SwfdecStreamTarget *target,
   if (ns->flvdecoder == NULL) {
     /* FIXME: add mp3 support */
     ns->flvdecoder = g_object_new (SWFDEC_TYPE_FLV_DECODER, NULL);
-    SWFDEC_DECODER (ns->flvdecoder)->player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (ns)->context);
+    g_signal_connect_swapped (ns->flvdecoder, "missing-plugin", 
+	G_CALLBACK (swfdec_player_add_missing_plugin), swfdec_gc_object_get_context (ns));
     swfdec_net_stream_onstatus (ns, SWFDEC_AS_STR_NetStream_Play_Start,
 	SWFDEC_AS_STR_status);
     swfdec_loader_set_data_type (SWFDEC_LOADER (stream), SWFDEC_LOADER_DATA_FLV);
@@ -326,38 +388,67 @@ swfdec_net_stream_stream_target_init (SwfdecStreamTargetInterface *iface)
   iface->error = swfdec_net_stream_stream_target_error;
 }
 
-/*** SWFDEC VIDEO MOVIE INPUT ***/
-
-static void
-swfdec_net_stream_input_connect (SwfdecVideoMovieInput *input, SwfdecVideoMovie *movie)
-{
-  SwfdecNetStream *stream = SWFDEC_NET_STREAM ((void *)((guchar *) input - G_STRUCT_OFFSET (SwfdecNetStream, input)));
-
-  stream->movies = g_list_prepend (stream->movies, movie);
-  g_object_ref (stream);
-}
-
-static void
-swfdec_net_stream_input_disconnect (SwfdecVideoMovieInput *input, SwfdecVideoMovie *movie)
-{
-  SwfdecNetStream *stream = SWFDEC_NET_STREAM ((void *)((guchar *) input - G_STRUCT_OFFSET (SwfdecNetStream, input)));
-
-  stream->movies = g_list_remove (stream->movies, movie);
-  g_object_unref (stream);
-}
+/*** SWFDEC VIDEO PROVIDER ***/
 
 static cairo_surface_t *
-swfdec_net_stream_input_get_image (SwfdecVideoMovieInput *input)
+swfdec_net_stream_video_provider_get_image (SwfdecVideoProvider *provider,
+    SwfdecRenderer *renderer, guint *width, guint *height)
 {
-  SwfdecNetStream *stream = SWFDEC_NET_STREAM ((void *)((guchar *) input - G_STRUCT_OFFSET (SwfdecNetStream, input)));
+  SwfdecNetStream *stream = SWFDEC_NET_STREAM (provider);
+  SwfdecCachedVideo *cached;
+  cairo_surface_t *surface;
 
-  return stream->surface;
+  cached = SWFDEC_CACHED_VIDEO (swfdec_renderer_get_cache (renderer, stream, NULL, NULL));
+  if (cached != NULL && swfdec_cached_video_get_frame (cached) == stream->current_time) {
+    swfdec_cached_use (SWFDEC_CACHED (cached));
+    swfdec_cached_video_get_size (cached, width, height);
+    return swfdec_cached_video_get_surface (cached);
+  }
+
+  if (stream->decoder == NULL)
+    return NULL;
+
+  surface = swfdec_video_decoder_get_image (stream->decoder, renderer);
+  if (surface == NULL)
+    return NULL;
+  *width = swfdec_video_decoder_get_width (stream->decoder);
+  *height = swfdec_video_decoder_get_height (stream->decoder);
+  cached = swfdec_cached_video_new (surface, *width * *height * 4);
+  swfdec_cached_video_set_frame (cached, stream->decoder_time);
+  swfdec_cached_video_set_size (cached, *width, *height);
+  swfdec_renderer_add_cache (renderer, TRUE, stream, SWFDEC_CACHED (cached));
+  g_object_unref (cached);
+
+  return surface;
+}
+
+static void
+swfdec_net_stream_video_provider_get_size (SwfdecVideoProvider *provider,
+    guint *width, guint *height)
+{
+  SwfdecNetStream *stream = SWFDEC_NET_STREAM (provider);
+
+  if (stream->decoder) {
+    *width = swfdec_video_decoder_get_width (stream->decoder);
+    *height = swfdec_video_decoder_get_height (stream->decoder);
+  } else {
+    *width = 0;
+    *height = 0;
+  }
+}
+
+static void
+swfdec_net_stream_video_provider_init (SwfdecVideoProviderInterface *iface)
+{
+  iface->get_image = swfdec_net_stream_video_provider_get_image;
+  iface->get_size = swfdec_net_stream_video_provider_get_size;
 }
 
 /*** SWFDEC_NET_STREAM ***/
 
 G_DEFINE_TYPE_WITH_CODE (SwfdecNetStream, swfdec_net_stream, SWFDEC_TYPE_AS_OBJECT,
-    G_IMPLEMENT_INTERFACE (SWFDEC_TYPE_STREAM_TARGET, swfdec_net_stream_stream_target_init))
+    G_IMPLEMENT_INTERFACE (SWFDEC_TYPE_STREAM_TARGET, swfdec_net_stream_stream_target_init)
+    G_IMPLEMENT_INTERFACE (SWFDEC_TYPE_VIDEO_PROVIDER, swfdec_net_stream_video_provider_init))
 
 static void
 swfdec_net_stream_dispose (GObject *object)
@@ -370,7 +461,7 @@ swfdec_net_stream_dispose (GObject *object)
     stream->surface = NULL;
   }
   if (stream->decoder) {
-    swfdec_video_decoder_free (stream->decoder);
+    g_object_unref (stream->decoder);
     stream->decoder = NULL;
   }
   swfdec_net_stream_set_loader (stream, NULL);
@@ -425,63 +516,57 @@ swfdec_net_stream_get_variable (SwfdecAsObject *object, SwfdecAsObject *orig,
     SWFDEC_AS_VALUE_SET_NUMBER (val, bytes);
     *flags = 0;
     return TRUE;
+  } else if (variable == SWFDEC_AS_STR_bufferLength) {
+    SWFDEC_STUB ("Netstream.bufferLength (get)");
+  } else if (variable == SWFDEC_AS_STR_bufferTime) {
+    SWFDEC_STUB ("Netstream.bufferTime (get)");
+  } else if (variable == SWFDEC_AS_STR_audiocodec) {
+    SWFDEC_STUB ("Netstream.audiocodec (get)");
+  } else if (variable == SWFDEC_AS_STR_currentFps) {
+    SWFDEC_STUB ("Netstream.currentFps (get)");
+  } else if (variable == SWFDEC_AS_STR_decodedFrames) {
+    SWFDEC_STUB ("Netstream.decodedFrames (get)");
+  } else if (variable == SWFDEC_AS_STR_liveDelay) {
+    SWFDEC_STUB ("Netstream.liveDelay (get)");
+  } else if (variable == SWFDEC_AS_STR_videoCodec) {
+    SWFDEC_STUB ("Netstream.videoCodec (get)");
   }
   return FALSE;
 }
 
 static void
-swfdec_net_stream_mark (SwfdecAsObject *object)
+swfdec_net_stream_mark (SwfdecGcObject *object)
 {
   SwfdecNetStream *stream = SWFDEC_NET_STREAM (object);
 
   if (stream->conn)
-    swfdec_as_object_mark (SWFDEC_AS_OBJECT (stream->conn));
+    swfdec_gc_object_mark (stream->conn);
   if (stream->sandbox)
-    swfdec_as_object_mark (SWFDEC_AS_OBJECT (stream->sandbox));
+    swfdec_gc_object_mark (stream->sandbox);
 
-  SWFDEC_AS_OBJECT_CLASS (swfdec_net_stream_parent_class)->mark (object);
+  SWFDEC_GC_OBJECT_CLASS (swfdec_net_stream_parent_class)->mark (object);
 }
 
 static void
 swfdec_net_stream_class_init (SwfdecNetStreamClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
+  SwfdecGcObjectClass *gc_class = SWFDEC_GC_OBJECT_CLASS (klass);
   SwfdecAsObjectClass *asobject_class = SWFDEC_AS_OBJECT_CLASS (klass);
 
   object_class->dispose = swfdec_net_stream_dispose;
 
+  gc_class->mark = swfdec_net_stream_mark;
+
   asobject_class->get = swfdec_net_stream_get_variable;
-  asobject_class->mark = swfdec_net_stream_mark;
 }
 
 static void
 swfdec_net_stream_init (SwfdecNetStream *stream)
 {
-  stream->input.connect = swfdec_net_stream_input_connect;
-  stream->input.disconnect = swfdec_net_stream_input_disconnect;
-  stream->input.get_image = swfdec_net_stream_input_get_image;
-
   stream->buffer_time = 100; /* msecs */
 }
 
-SwfdecNetStream *
-swfdec_net_stream_new (SwfdecNetConnection *conn)
-{
-  SwfdecAsContext *context;
-  SwfdecNetStream *stream;
-  
-  g_return_val_if_fail (SWFDEC_IS_NET_CONNECTION (conn), NULL);
-
-  context = SWFDEC_AS_OBJECT (conn)->context;
-  if (!swfdec_as_context_use_mem (context, sizeof (SwfdecNetStream)))
-    return NULL;
-  stream = g_object_new (SWFDEC_TYPE_NET_STREAM, NULL);
-  swfdec_as_object_add (SWFDEC_AS_OBJECT (stream), context, sizeof (SwfdecNetStream));
-  stream->conn = conn;
-
-  return stream;
-}
-
 static void
 swfdec_net_stream_load (SwfdecPlayer *player, gboolean allowed, gpointer streamp)
 {
@@ -489,8 +574,7 @@ swfdec_net_stream_load (SwfdecPlayer *player, gboolean allowed, gpointer streamp
   SwfdecLoader *loader;
 
   if (allowed) {
-    loader = swfdec_player_load (player, stream->requested_url, 
-	SWFDEC_LOADER_REQUEST_DEFAULT, NULL);
+    loader = swfdec_player_load (player, stream->requested_url, NULL);
     swfdec_net_stream_set_loader (stream, loader);
     g_object_unref (loader);
   } else {
@@ -502,17 +586,24 @@ swfdec_net_stream_load (SwfdecPlayer *player, gboolean allowed, gpointer streamp
   stream->requested_url = NULL;
 }
 
+static const SwfdecAccessMatrix swfdec_net_stream_matrix = {
+  { SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO },
+  { SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_YES, SWFDEC_ACCESS_POLICY },
+  { SWFDEC_ACCESS_YES, SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO },
+  { SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_POLICY },
+  { SWFDEC_ACCESS_YES, SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_POLICY }
+};
+
 void
-swfdec_net_stream_set_url (SwfdecNetStream *stream, SwfdecSandbox *sandbox, const char *url_string)
+swfdec_net_stream_set_url (SwfdecNetStream *stream, const char *url_string)
 {
   SwfdecPlayer *player;
   SwfdecAsContext *cx;
-  SwfdecURL *url;
 
   g_return_if_fail (SWFDEC_IS_NET_STREAM (stream));
-  g_return_if_fail (SWFDEC_IS_SANDBOX (sandbox));
+  g_return_if_fail (url_string != NULL);
 
-  cx = SWFDEC_AS_OBJECT (stream)->context;
+  cx = swfdec_gc_object_get_context (stream);
   player = SWFDEC_PLAYER (cx);
 
   if (stream->requested_url != NULL) {
@@ -521,45 +612,10 @@ swfdec_net_stream_set_url (SwfdecNetStream *stream, SwfdecSandbox *sandbox, cons
     return;
   }
   stream->requested_url = g_strdup (url_string);
-  stream->sandbox = sandbox;
-  if (swfdec_url_path_is_relative (url_string)) {
-    swfdec_net_stream_load (player, TRUE, stream);
-    return;
-  }
-  url = swfdec_player_create_url (player, url_string);
-  if (url == NULL) {
-    swfdec_net_stream_load (player, FALSE, stream);
-    return;
-  }
-  if (swfdec_url_is_local (url)) {
-    swfdec_net_stream_load (player, TRUE, stream);
-  } else {
-    switch (sandbox->type) {
-      case SWFDEC_SANDBOX_REMOTE:
-	swfdec_net_stream_load (player, TRUE, stream);
-	break;
-      case SWFDEC_SANDBOX_LOCAL_NETWORK:
-      case SWFDEC_SANDBOX_LOCAL_TRUSTED:
-	{
-	  SwfdecURL *load_url = swfdec_url_new_components (
-	      swfdec_url_get_protocol (url), swfdec_url_get_host (url), 
-	      swfdec_url_get_port (url), "crossdomain.xml", NULL);
-	  swfdec_player_allow_or_load (player, url, load_url,
-	    swfdec_net_stream_load, stream);
-	  swfdec_url_free (load_url);
-	}
-	break;
-      case SWFDEC_SANDBOX_LOCAL_FILE:
-	swfdec_net_stream_load (player, FALSE, stream);
-	break;
-      case SWFDEC_SANDBOX_NONE:
-      default:
-	g_assert_not_reached ();
-	break;
-    }
-  }
+  stream->sandbox = SWFDEC_SANDBOX (SWFDEC_AS_CONTEXT (player)->global);
 
-  swfdec_url_free (url);
+  swfdec_player_allow_by_matrix (player, stream->sandbox, url_string, 
+      swfdec_net_stream_matrix, swfdec_net_stream_load, stream);
 }
 
 void
@@ -571,11 +627,13 @@ swfdec_net_stream_set_loader (SwfdecNetStream *stream, SwfdecLoader *loader)
 
   if (stream->loader) {
     SwfdecStream *lstream = SWFDEC_STREAM (stream->loader);
-    swfdec_stream_close (lstream);
+    swfdec_stream_ensure_closed (lstream);
     swfdec_stream_set_target (lstream, NULL);
     g_object_unref (lstream);
   }
   if (stream->flvdecoder) {
+    g_signal_handlers_disconnect_by_func (stream->flvdecoder,
+	  swfdec_player_add_missing_plugin, swfdec_gc_object_get_context (stream));
     g_object_unref (stream->flvdecoder);
     stream->flvdecoder = NULL;
   }
@@ -643,6 +701,10 @@ swfdec_net_stream_seek (SwfdecNetStream *stream, double secs)
     SWFDEC_ERROR ("FIXME: implement seeking in audio only NetStream");
     return;
   }
+  if (stream->decoder) {
+    g_object_unref (stream->decoder);
+    stream->decoder = NULL;
+  }
   msecs = secs * 1000;
   msecs += first;
   if (msecs > last)
@@ -658,7 +720,7 @@ swfdec_net_stream_seek (SwfdecNetStream *stream, double secs)
     SWFDEC_WARNING ("FIXME: restarting audio after seek");
     swfdec_audio_remove (stream->audio);
     g_object_unref (stream->audio);
-    stream->audio = swfdec_audio_flv_new (SWFDEC_PLAYER (SWFDEC_AS_OBJECT (stream)->context), 
+    stream->audio = swfdec_audio_flv_new (SWFDEC_PLAYER (swfdec_gc_object_get_context (stream)), 
 	stream->flvdecoder, stream->current_time);
   }
 }
