@@ -31,6 +31,7 @@
 #include "swfdec_gradient_pattern.h"
 #include "swfdec_image.h"
 #include "swfdec_path.h"
+#include "swfdec_renderer_internal.h"
 #include "swfdec_stroke.h"
 
 /*** PATTERN ***/
@@ -48,7 +49,8 @@ swfdec_pattern_paint (SwfdecDraw *draw, cairo_t *cr, const SwfdecColorTransform
 {
   cairo_pattern_t *pattern;
 
-  pattern = swfdec_pattern_get_pattern (SWFDEC_PATTERN (draw), trans);
+  pattern = swfdec_pattern_get_pattern (SWFDEC_PATTERN (draw), swfdec_renderer_get (cr),
+      trans);
   if (pattern == NULL)
     return;
   cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
@@ -143,14 +145,15 @@ swfdec_color_pattern_morph (SwfdecDraw *dest, SwfdecDraw *source, guint ratio)
 }
 
 static cairo_pattern_t *
-swfdec_color_pattern_get_pattern (SwfdecPattern *pat, const SwfdecColorTransform *trans)
+swfdec_color_pattern_get_pattern (SwfdecPattern *pat, SwfdecRenderer *renderer,
+    const SwfdecColorTransform *trans)
 {
   SwfdecColor color = SWFDEC_COLOR_PATTERN (pat)->start_color;
 
   color = swfdec_color_apply_transform (color, trans);
   return cairo_pattern_create_rgba ( 
-      SWFDEC_COLOR_R (color) / 255.0, SWFDEC_COLOR_G (color) / 255.0,
-      SWFDEC_COLOR_B (color) / 255.0, SWFDEC_COLOR_A (color) / 255.0);
+      SWFDEC_COLOR_RED (color) / 255.0, SWFDEC_COLOR_GREEN (color) / 255.0,
+      SWFDEC_COLOR_BLUE (color) / 255.0, SWFDEC_COLOR_ALPHA (color) / 255.0);
 }
 
 static void
@@ -208,14 +211,43 @@ swfdec_image_pattern_morph (SwfdecDraw *dest, SwfdecDraw *source, guint ratio)
   SWFDEC_DRAW_CLASS (swfdec_image_pattern_parent_class)->morph (dest, source, ratio);
 }
 
+static void
+swfdec_image_pattern_paint (SwfdecDraw *draw, cairo_t *cr, const SwfdecColorTransform *trans)
+{
+  cairo_pattern_t *pattern;
+
+  if (swfdec_color_transform_is_alpha (trans)) {
+    SwfdecColorTransform identity;
+    swfdec_color_transform_init_identity (&identity);
+    pattern = swfdec_pattern_get_pattern (SWFDEC_PATTERN (draw),
+	swfdec_renderer_get (cr), &identity);
+    if (pattern == NULL)
+      return;
+    cairo_save (cr);
+    cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
+    cairo_set_fill_rule (cr, CAIRO_FILL_RULE_EVEN_ODD);
+    cairo_append_path (cr, &draw->path);
+    cairo_clip (cr);
+    cairo_set_source (cr, pattern);
+    cairo_pattern_destroy (pattern);
+    cairo_paint_with_alpha (cr, trans->aa / 256.0);
+    cairo_restore (cr);
+  } else {
+    SWFDEC_DRAW_CLASS (swfdec_image_pattern_parent_class)->paint (draw, cr, trans);
+  }
+}
+
 static cairo_pattern_t *
-swfdec_image_pattern_get_pattern (SwfdecPattern *pat, const SwfdecColorTransform *trans)
+swfdec_image_pattern_get_pattern (SwfdecPattern *pat, SwfdecRenderer *renderer,
+    const SwfdecColorTransform *trans)
 {
   SwfdecImagePattern *image = SWFDEC_IMAGE_PATTERN (pat);
   cairo_pattern_t *pattern;
   cairo_surface_t *surface;
   
-  surface = swfdec_image_create_surface_transformed (image->image, trans);
+  if (swfdec_color_transform_is_mask (trans))
+    return cairo_pattern_create_rgb (0, 0, 0);
+  surface = swfdec_image_create_surface_transformed (image->image, renderer, trans);
   if (surface == NULL)
     return NULL;
   pattern = cairo_pattern_create_for_surface (surface);
@@ -230,6 +262,7 @@ static void
 swfdec_image_pattern_class_init (SwfdecImagePatternClass *klass)
 {
   SWFDEC_DRAW_CLASS (klass)->morph = swfdec_image_pattern_morph;
+  SWFDEC_DRAW_CLASS (klass)->paint = swfdec_image_pattern_paint;
 
   SWFDEC_PATTERN_CLASS (klass)->get_pattern = swfdec_image_pattern_get_pattern;
 }
@@ -324,13 +357,7 @@ swfdec_pattern_do_parse (SwfdecBits *bits, SwfdecSwfDecoder *dec, gboolean rgba)
       if (paint_style_type == 0x40 || paint_style_type == 0x42) {
 	SWFDEC_IMAGE_PATTERN (pattern)->extend = CAIRO_EXTEND_REPEAT;
       } else {
-#if 0
-	/* not implemented yet in cairo */
 	SWFDEC_IMAGE_PATTERN (pattern)->extend = CAIRO_EXTEND_PAD;
-#else
-	SWFDEC_FIXME ("CAIRO_EXTEND_PAD is not yet implemented");
-	SWFDEC_IMAGE_PATTERN (pattern)->extend = CAIRO_EXTEND_NONE;
-#endif
       }
       if (paint_style_type == 0x40 || paint_style_type == 0x41) {
 	SWFDEC_IMAGE_PATTERN (pattern)->filter = CAIRO_FILTER_BILINEAR;
@@ -470,13 +497,7 @@ swfdec_pattern_parse_morph (SwfdecBits *bits, SwfdecSwfDecoder *dec)
       if (paint_style_type == 0x40 || paint_style_type == 0x42) {
 	SWFDEC_IMAGE_PATTERN (pattern)->extend = CAIRO_EXTEND_REPEAT;
       } else {
-#if 0
-	/* not implemented yet in cairo */
 	SWFDEC_IMAGE_PATTERN (pattern)->extend = CAIRO_EXTEND_PAD;
-#else
-	SWFDEC_FIXME ("CAIRO_EXTEND_PAD is not yet implemented");
-	SWFDEC_IMAGE_PATTERN (pattern)->extend = CAIRO_EXTEND_NONE;
-#endif
       }
       if (paint_style_type == 0x40 || paint_style_type == 0x41) {
 	SWFDEC_IMAGE_PATTERN (pattern)->filter = CAIRO_FILTER_BILINEAR;
@@ -524,7 +545,7 @@ swfdec_pattern_to_string (SwfdecPattern *pattern)
   if (SWFDEC_IS_IMAGE_PATTERN (pattern)) {
     SwfdecImagePattern *image = SWFDEC_IMAGE_PATTERN (pattern);
     if (image->image->width == 0)
-      cairo_surface_destroy (swfdec_image_create_surface (image->image));
+      cairo_surface_destroy (swfdec_image_create_surface (image->image, NULL));
     return g_strdup_printf ("%ux%u image %u (%s, %s)", image->image->width,
 	image->image->height, SWFDEC_CHARACTER (image->image)->id,
 	image->extend == CAIRO_EXTEND_REPEAT ? "repeat" : "no repeat",
@@ -545,15 +566,17 @@ swfdec_pattern_to_string (SwfdecPattern *pattern)
 }
 
 cairo_pattern_t *
-swfdec_pattern_get_pattern (SwfdecPattern *pattern, const SwfdecColorTransform *trans)
+swfdec_pattern_get_pattern (SwfdecPattern *pattern, SwfdecRenderer *renderer,
+    const SwfdecColorTransform *trans)
 {
   SwfdecPatternClass *klass;
 
   g_return_val_if_fail (SWFDEC_IS_PATTERN (pattern), NULL);
+  g_return_val_if_fail (SWFDEC_IS_RENDERER (renderer), NULL);
   g_return_val_if_fail (trans != NULL, NULL);
 
   klass = SWFDEC_PATTERN_GET_CLASS (pattern);
   g_assert (klass->get_pattern);
-  return klass->get_pattern (pattern, trans);
+  return klass->get_pattern (pattern, renderer, trans);
 }
 
