@@ -1,5 +1,5 @@
 /* Swfdec
- * Copyright (C) 2006-2007 Benjamin Otte <otte@gnome.org>
+ * Copyright (C) 2006-2008 Benjamin Otte <otte@gnome.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -25,6 +25,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include "swfdec_resource.h"
+#include "swfdec_access.h"
 #include "swfdec_as_object.h"
 #include "swfdec_as_internal.h"
 #include "swfdec_as_interpret.h"
@@ -56,19 +57,19 @@ swfdec_resource_is_root (SwfdecResource *resource)
   g_return_val_if_fail (SWFDEC_IS_RESOURCE (resource), FALSE);
 
   return
-    resource->movie == SWFDEC_PLAYER (SWFDEC_AS_OBJECT (resource)->context)->priv->roots->data;
+    resource->movie == SWFDEC_PLAYER (swfdec_gc_object_get_context (resource))->priv->roots->data;
 }
 
 static SwfdecPlayer *
 swfdec_resource_stream_target_get_player (SwfdecStreamTarget *target)
 {
-  return SWFDEC_PLAYER (SWFDEC_AS_OBJECT (target)->context);
+  return SWFDEC_PLAYER (swfdec_gc_object_get_context (target));
 }
 
 static void
 swfdec_resource_stream_target_image (SwfdecResource *instance)
 {
-  SwfdecPlayer *player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (instance)->context);
+  SwfdecPlayer *player = SWFDEC_PLAYER (swfdec_gc_object_get_context (instance));
   SwfdecSpriteMovie *movie = instance->movie;
 
   if (movie->sprite != NULL)
@@ -111,28 +112,31 @@ swfdec_resource_emit_signal (SwfdecResource *resource, const char *name, gboolea
     SwfdecAsValue *args, guint n_args)
 {
   SwfdecAsContext *cx;
-  SwfdecAsObject *movie;
+  SwfdecMovie *movie;
   guint skip = progress ? 4 : 2;
   SwfdecAsValue vals[n_args + skip];
 
   if (resource->clip_loader == NULL)
     return;
-  cx = SWFDEC_AS_OBJECT (resource->clip_loader)->context;
-  g_assert (resource->target);
-  movie = swfdec_action_lookup_object (cx, SWFDEC_PLAYER (cx)->priv->roots->data, 
-      resource->target, resource->target + strlen (resource->target));
-  if (!SWFDEC_IS_SPRITE_MOVIE (movie)) {
-    SWFDEC_DEBUG ("no movie, not emitting signal");
-    return;
-  }
-  if (name == SWFDEC_AS_STR_onLoadInit &&
-      movie != SWFDEC_AS_OBJECT (resource->movie)) {
-    SWFDEC_INFO ("not emitting onLoadInit - the movie is different");
-    return;
-  }
+  cx = swfdec_gc_object_get_context (resource->clip_loader);
 
   SWFDEC_AS_VALUE_SET_STRING (&vals[0], name);
-  SWFDEC_AS_VALUE_SET_OBJECT (&vals[1], movie);
+  if (resource->movie) {
+    movie = swfdec_movie_resolve (SWFDEC_MOVIE (resource->movie));
+    if (movie == NULL) {
+      SWFDEC_DEBUG ("no movie, not emitting signal");
+      return;
+    }
+    if (name == SWFDEC_AS_STR_onLoadInit &&
+	movie != SWFDEC_MOVIE (resource->movie)) {
+      SWFDEC_INFO ("not emitting onLoadInit - the movie is different");
+      return;
+    }
+    SWFDEC_AS_VALUE_SET_OBJECT (&vals[1], SWFDEC_AS_OBJECT (movie));
+  } else {
+    SWFDEC_AS_VALUE_SET_UNDEFINED (&vals[1]);
+    movie = NULL;
+  }
   if (progress) {
     SwfdecResource *res;
     
@@ -176,58 +180,27 @@ swfdec_resource_replace_movie (SwfdecSpriteMovie *movie, SwfdecResource *resourc
   SwfdecMovie *mov = SWFDEC_MOVIE (movie);
   SwfdecMovie *copy;
   
-  copy = swfdec_movie_new (SWFDEC_PLAYER (SWFDEC_AS_OBJECT (movie)->context), 
+  copy = swfdec_movie_new (SWFDEC_PLAYER (swfdec_gc_object_get_context (movie)), 
       mov->depth, mov->parent, resource, NULL, mov->name);
   if (copy == NULL)
     return FALSE;
+  swfdec_movie_begin_update_matrix (copy);
   copy->matrix = mov->matrix;
   copy->original_name = mov->original_name;
   copy->modified = mov->modified;
   copy->xscale = mov->xscale;
   copy->yscale = mov->yscale;
   copy->rotation = mov->rotation;
+  copy->lockroot = mov->lockroot;
+  swfdec_movie_end_update_matrix (copy);
   /* FIXME: are events copied? If so, wouldn't that be a security issue? */
   swfdec_movie_set_static_properties (copy, &mov->original_transform,
-      &mov->original_ctrans, mov->original_ratio, mov->clip_depth, 
+      &mov->color_transform, mov->original_ratio, mov->clip_depth, 
       mov->blend_mode, NULL);
   swfdec_movie_remove (mov);
-  swfdec_movie_queue_update (copy, SWFDEC_MOVIE_INVALID_MATRIX);
   return SWFDEC_SPRITE_MOVIE (copy);
 }
 
-static gboolean
-swfdec_resource_create_movie (SwfdecResource *resource)
-{
-  SwfdecPlayer *player;
-  SwfdecSpriteMovie *movie;
-  int level = -1;
-
-  if (resource->movie)
-    return TRUE;
-  player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (resource)->context);
-  movie = (SwfdecSpriteMovie *) swfdec_action_lookup_object (SWFDEC_AS_CONTEXT (player),
-      player->priv->roots->data, resource->target, resource->target + strlen (resource->target));
-  if (!SWFDEC_IS_SPRITE_MOVIE (movie)) {
-    level = swfdec_player_get_level (player, resource->target);
-    if (level < 0) {
-      SWFDEC_WARNING ("%s does not reference a movie, not loading %s", resource->target,
-	  swfdec_url_get_url (swfdec_loader_get_url (resource->loader)));
-      swfdec_stream_close (SWFDEC_STREAM (resource->loader));
-      swfdec_player_unroot (player, resource);
-      return FALSE;
-    }
-    movie = swfdec_player_get_movie_at_level (player, level);
-  }
-  if (movie == NULL) {
-    movie = swfdec_player_create_movie_at_level (player, resource, level);
-  } else {
-    /* FIXME: is this correct? */
-    movie = swfdec_resource_replace_movie (movie, resource);
-  }
-  swfdec_player_unroot (player, resource);
-  return TRUE;
-}
-
 static void
 swfdec_resource_stream_target_open (SwfdecStreamTarget *target, SwfdecStream *stream)
 {
@@ -235,8 +208,7 @@ swfdec_resource_stream_target_open (SwfdecStreamTarget *target, SwfdecStream *st
   SwfdecResource *instance = SWFDEC_RESOURCE (target);
   const char *query;
 
-  if (!swfdec_resource_create_movie (instance))
-    return;
+  g_assert (instance->movie);
   query = swfdec_url_get_query (swfdec_loader_get_url (loader));
   if (query) {
     SWFDEC_INFO ("set url query movie variables: %s", query);
@@ -266,14 +238,15 @@ swfdec_resource_stream_target_parse (SwfdecStreamTarget *target, SwfdecStream *s
     if (swfdec_buffer_queue_get_depth (queue) < SWFDEC_DECODER_DETECT_LENGTH)
       return FALSE;
     buffer = swfdec_buffer_queue_peek (queue, 4);
-    dec =
-      swfdec_decoder_new (SWFDEC_PLAYER (SWFDEC_AS_OBJECT (resource)->context), buffer);
+    dec = swfdec_decoder_new (buffer);
     swfdec_buffer_unref (buffer);
     if (dec == NULL) {
       SWFDEC_ERROR ("no decoder found for format");
     } else {
       glong total;
       resource->decoder = dec;
+      g_signal_connect_swapped (dec, "missing-plugin", 
+	  G_CALLBACK (swfdec_player_add_missing_plugin), swfdec_gc_object_get_context (resource));
       total = swfdec_loader_get_size (loader);
       if (total >= 0)
 	dec->bytes_total = total;
@@ -309,7 +282,7 @@ swfdec_resource_stream_target_parse (SwfdecStreamTarget *target, SwfdecStream *s
       if (SWFDEC_IS_SWF_DECODER (dec))
 	resource->version = SWFDEC_SWF_DECODER (dec)->version;
       if (swfdec_resource_is_root (resource)) {
-	swfdec_player_initialize (SWFDEC_PLAYER (SWFDEC_AS_OBJECT (resource)->context),
+	swfdec_player_initialize (SWFDEC_PLAYER (swfdec_gc_object_get_context (resource)),
 	    dec->rate, dec->width, dec->height);
       }
     }
@@ -322,14 +295,15 @@ swfdec_resource_stream_target_parse (SwfdecStreamTarget *target, SwfdecStream *s
   return FALSE;
 }
 
-static void
+static gboolean
 swfdec_resource_abort_if_not_initialized (SwfdecResource *resource)
 {
   if (resource->sandbox)
-    return;
+    return FALSE;
 
-  swfdec_as_context_abort (SWFDEC_AS_OBJECT (resource)->context,
+  swfdec_as_context_abort (swfdec_gc_object_get_context (resource),
       "This is not a Flash file");
+  return TRUE;
 }
 
 static void
@@ -345,10 +319,15 @@ swfdec_resource_stream_target_close (SwfdecStreamTarget *target, SwfdecStream *s
     if (dec->data_type != SWFDEC_LOADER_DATA_UNKNOWN)
       swfdec_loader_set_data_type (SWFDEC_LOADER (stream), dec->data_type);
   }
+
   SWFDEC_AS_VALUE_SET_INT (&val, 0); /* FIXME */
   swfdec_resource_emit_signal (resource, SWFDEC_AS_STR_onLoadComplete, FALSE, &val, 1);
   resource->state = SWFDEC_RESOURCE_COMPLETE;
-  swfdec_resource_abort_if_not_initialized (resource);
+  if (swfdec_resource_abort_if_not_initialized (resource))
+    return;
+
+  if (resource->movie != NULL)
+    swfdec_actor_queue_script (SWFDEC_ACTOR (resource->movie), SWFDEC_EVENT_LOAD);
 }
 
 static void
@@ -387,18 +366,20 @@ swfdec_resource_stream_target_init (SwfdecStreamTargetInterface *iface)
 }
 
 static void
-swfdec_resource_mark (SwfdecAsObject *object)
+swfdec_resource_mark (SwfdecGcObject *object)
 {
   SwfdecResource *resource = SWFDEC_RESOURCE (object);
 
   if (resource->clip_loader) {
-    swfdec_as_object_mark (SWFDEC_AS_OBJECT (resource->clip_loader));
-    swfdec_as_object_mark (SWFDEC_AS_OBJECT (resource->clip_loader_sandbox));
+    swfdec_gc_object_mark (resource->clip_loader);
+    swfdec_gc_object_mark (resource->clip_loader_sandbox);
   }
   if (resource->sandbox)
-    swfdec_as_object_mark (SWFDEC_AS_OBJECT (resource->sandbox));
+    swfdec_gc_object_mark (resource->sandbox);
+  if (resource->target)
+    swfdec_gc_object_mark (resource->target);
 
-  SWFDEC_AS_OBJECT_CLASS (swfdec_resource_parent_class)->mark (object);
+  SWFDEC_GC_OBJECT_CLASS (swfdec_resource_parent_class)->mark (object);
 }
 
 static void
@@ -412,10 +393,11 @@ swfdec_resource_dispose (GObject *object)
     resource->loader = NULL;
   }
   if (resource->decoder) {
+    g_signal_handlers_disconnect_by_func (resource->decoder,
+	  swfdec_player_add_missing_plugin, swfdec_gc_object_get_context (resource));
     g_object_unref (resource->decoder);
     resource->decoder = NULL;
   }
-  g_free (resource->target);
   g_free (resource->variables);
   g_hash_table_destroy (resource->exports);
   g_hash_table_destroy (resource->export_names);
@@ -427,11 +409,11 @@ static void
 swfdec_resource_class_init (SwfdecResourceClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
-  SwfdecAsObjectClass *asobject_class = SWFDEC_AS_OBJECT_CLASS (klass);
+  SwfdecGcObjectClass *gc_class = SWFDEC_GC_OBJECT_CLASS (klass);
 
   object_class->dispose = swfdec_resource_dispose;
 
-  asobject_class->mark = swfdec_resource_mark;
+  gc_class->mark = swfdec_resource_mark;
 }
 
 static void
@@ -459,16 +441,11 @@ SwfdecResource *
 swfdec_resource_new (SwfdecPlayer *player, SwfdecLoader *loader, const char *variables)
 {
   SwfdecResource *resource;
-  guint size;
 
   g_return_val_if_fail (SWFDEC_IS_PLAYER (player), NULL);
   g_return_val_if_fail (SWFDEC_IS_LOADER (loader), NULL);
 
-  size = sizeof (SwfdecResource);
-  if (!swfdec_as_context_use_mem (SWFDEC_AS_CONTEXT (player), size))
-    size = 0;
-  resource = g_object_new (SWFDEC_TYPE_RESOURCE, NULL);
-  swfdec_as_object_add (SWFDEC_AS_OBJECT (resource), SWFDEC_AS_CONTEXT (player), size);
+  resource = g_object_new (SWFDEC_TYPE_RESOURCE, "context", player, NULL);
   resource->version = 8;
   resource->variables = g_strdup (variables);
   swfdec_resource_set_loader (resource, loader);
@@ -510,12 +487,11 @@ swfdec_resource_add_export (SwfdecResource *instance, SwfdecCharacter *character
 typedef struct _SwfdecResourceLoad SwfdecResourceLoad;
 struct _SwfdecResourceLoad {
   SwfdecSandbox *		sandbox;
-  char *			target;
+  char *			target_string;
+  SwfdecSpriteMovie *		target_movie;
   char *			url;
-  SwfdecLoaderRequest		request;
   SwfdecBuffer *		buffer;
   SwfdecMovieClipLoader *	loader;
-  gboolean			target_is_movie;
 };
 
 static void
@@ -523,9 +499,8 @@ swfdec_resource_load_free (gpointer loadp)
 {
   SwfdecResourceLoad *load = loadp;
 
-  swfdec_player_unroot (SWFDEC_PLAYER (SWFDEC_AS_OBJECT (load->sandbox)->context), load);
   g_free (load->url);
-  g_free (load->target);
+  g_free (load->target_string);
   if (load->buffer)
     swfdec_buffer_unref (load->buffer);
   g_slice_free (SwfdecResourceLoad, load);
@@ -536,9 +511,46 @@ swfdec_resource_load_mark (gpointer loadp, gpointer playerp)
 {
   SwfdecResourceLoad *load = loadp;
 
-  swfdec_as_object_mark (SWFDEC_AS_OBJECT (load->sandbox));
+  swfdec_gc_object_mark (load->sandbox);
   if (load->loader)
-    swfdec_as_object_mark (SWFDEC_AS_OBJECT (load->loader));
+    swfdec_gc_object_mark (load->loader);
+  if (load->target_movie)
+    swfdec_gc_object_mark (load->target_movie);
+}
+
+static gboolean
+swfdec_resource_create_movie (SwfdecResource *resource, SwfdecResourceLoad *load)
+{
+  SwfdecPlayer *player;
+  SwfdecSpriteMovie *movie;
+
+  if (resource->movie)
+    return TRUE;
+  player = SWFDEC_PLAYER (swfdec_gc_object_get_context (resource));
+  if (load->target_movie) {
+    movie = (SwfdecSpriteMovie *) swfdec_movie_resolve (SWFDEC_MOVIE (load->target_movie));
+    if (SWFDEC_IS_SPRITE_MOVIE (movie))
+      movie = swfdec_resource_replace_movie (movie, resource);
+    else
+      movie = NULL;
+  } else {
+    int level = swfdec_player_get_level (player, load->target_string, 7);
+    if (level >= 0) {
+      movie = swfdec_player_get_movie_at_level (player, level);
+      if (movie)
+	swfdec_movie_remove (SWFDEC_MOVIE (movie));
+      movie = swfdec_player_create_movie_at_level (player, resource, level);
+    } else {
+      movie = NULL;
+    }
+  }
+  if (movie == NULL) {
+    SWFDEC_WARNING ("target does not reference a movie, not loading %s", load->url);
+    return FALSE;
+  }
+  /* FIXME: does this belong here? */
+  SWFDEC_ACTOR (movie)->focusrect = SWFDEC_FLASH_YES;
+  return TRUE;
 }
 
 static void
@@ -548,139 +560,178 @@ swfdec_resource_do_load (SwfdecPlayer *player, gboolean allowed, gpointer loadp)
   SwfdecResource *resource;
   SwfdecLoader *loader;
   
-  if (!swfdec_as_context_use_mem (SWFDEC_AS_CONTEXT (player), sizeof (SwfdecResource)))
-    return;
-  resource = g_object_new (SWFDEC_TYPE_RESOURCE, NULL);
-  swfdec_as_object_add (SWFDEC_AS_OBJECT (resource), SWFDEC_AS_CONTEXT (player), sizeof (SwfdecResource));
+  resource = g_object_new (SWFDEC_TYPE_RESOURCE, "context", player, NULL);
   resource->version = 8;
-  resource->target = g_strdup (load->target);
   if (load->loader) {
     resource->clip_loader = load->loader;
     resource->clip_loader_sandbox = load->sandbox;
   }
   resource->sandbox = load->sandbox;
-
   if (!allowed) {
     SWFDEC_WARNING ("SECURITY: no access to %s from %s",
 	load->url, swfdec_url_get_url (load->sandbox->url));
     /* FIXME: is replacing correct? */
-    swfdec_resource_emit_error (resource, SWFDEC_AS_STR_IllegalRequest);
+    if (load->target_movie) {
+      resource->movie = SWFDEC_SPRITE_MOVIE (swfdec_movie_resolve (SWFDEC_MOVIE (load->target_movie)));
+      swfdec_resource_emit_error (resource, SWFDEC_AS_STR_IllegalRequest);
+    }
+    swfdec_player_unroot (player, load);
     return;
   }
 
-  swfdec_player_root (player, resource, (GFunc) swfdec_as_object_mark);
-  loader = swfdec_player_load (player, load-> url, load->request, load->buffer);
-  swfdec_resource_set_loader (resource, loader);
-  g_object_unref (loader);
+  /* FIXME: load nonetheless, even if there's no movie? */
+  if (swfdec_resource_create_movie (resource, load)) {
+    loader = swfdec_player_load (player, load->url, load->buffer);
+    swfdec_resource_set_loader (resource, loader);
+    g_object_unref (loader);
+  }
+  swfdec_player_unroot (player, load);
 }
 
+static const SwfdecAccessMatrix swfdec_resource_matrix = {
+  { SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO },
+  { SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_YES, SWFDEC_ACCESS_YES },
+  { SWFDEC_ACCESS_YES, SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_NO },
+  { SWFDEC_ACCESS_YES, SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_POLICY },
+  { SWFDEC_ACCESS_YES, SWFDEC_ACCESS_NO,  SWFDEC_ACCESS_POLICY }
+};
+
 static void
 swfdec_resource_load_request (gpointer loadp, gpointer playerp)
 {
   SwfdecResourceLoad *load = loadp;
   SwfdecPlayer *player = playerp;
-  SwfdecURL *url;
 
   /* empty URL means unload (yay!) */
   if (load->url[0] == '\0') {
     SwfdecSpriteMovie *movie;
       
-    movie = (SwfdecSpriteMovie *) swfdec_action_lookup_object (
-	SWFDEC_AS_CONTEXT (player), player->priv->roots->data, 
-	load->target, load->target + strlen (load->target));
-    if (!SWFDEC_IS_SPRITE_MOVIE (movie)) {
+    movie = load->target_movie ? (SwfdecSpriteMovie *) swfdec_movie_resolve (SWFDEC_MOVIE (load->target_movie)) : NULL;
+    if (SWFDEC_IS_SPRITE_MOVIE (movie)) {
+      swfdec_resource_replace_movie (movie, SWFDEC_MOVIE (movie)->resource);
+    } else {
       SWFDEC_DEBUG ("no movie, not unloading");
-      return;
     }
-    swfdec_resource_replace_movie (movie, SWFDEC_MOVIE (movie)->resource);
+    swfdec_player_unroot (player, load);
     return;
   }
 
   /* fscommand? */
   if (g_ascii_strncasecmp (load->url, "FSCommand:", 10) == 0) {
     char *command = load->url + 10;
-    g_signal_emit_by_name (player, "fscommand", command, load->target);
+    if (load->target_movie) {
+      char *target = swfdec_movie_get_path (SWFDEC_MOVIE (load->target_movie), TRUE);
+      SWFDEC_FIXME ("Adobe 9.0.124.0 and later don't emit fscommands here. "
+	  "We just do for compatibility reasons with the testsuite.");
+      g_signal_emit_by_name (player, "fscommand", command, target);
+      g_free (target);
+    } else {
+      g_signal_emit_by_name (player, "fscommand", command, load->target_string);
+    }
+    swfdec_player_unroot (player, load);
     return;
   }
 
   /* LAUNCH command (aka getURL) */
-  if (!load->target_is_movie && swfdec_player_get_level (player, load->target) < 0) {
-    swfdec_player_launch (player, load->request, load->url,load->target, load->buffer);
+  if (load->target_string && swfdec_player_get_level (player, load->target_string, 7) < 0) {
+    swfdec_player_launch (player, load->url, load->target_string, load->buffer);
+    swfdec_player_unroot (player, load);
     return;
   }
 
-  if (swfdec_url_path_is_relative (load->url)) {
-    swfdec_resource_do_load (player, TRUE, load);
-    return;
-  }
-  url = swfdec_player_create_url (player, load->url);
-  if (url == NULL) {
-    swfdec_resource_do_load (player, FALSE, load);
-    return;
-  }
-  switch (load->sandbox->type) {
-    case SWFDEC_SANDBOX_REMOTE:
-      swfdec_resource_do_load (player, !swfdec_url_is_local (url), load);
-      break;
-    case SWFDEC_SANDBOX_LOCAL_NETWORK:
-    case SWFDEC_SANDBOX_LOCAL_TRUSTED:
-      if (!swfdec_url_is_local (url)) {
-	SWFDEC_FIXME ("Adobe claims you need to be allowed by policy files now, "
-	    "we don't check that though");
-      }
-      swfdec_resource_do_load (player, TRUE, load);
-      break;
-    case SWFDEC_SANDBOX_LOCAL_FILE:
-      swfdec_resource_do_load (player, swfdec_url_is_local (url), load);
-      break;
-    case SWFDEC_SANDBOX_NONE:
-    default:
-      g_assert_not_reached ();
-      break;
-  }
-  swfdec_url_free (url);
+  swfdec_player_allow_by_matrix (player, load->sandbox, load->url,
+      swfdec_resource_matrix, swfdec_resource_do_load, load);
 }
 
 /* NB: must be called from a script */
-/* FIXME: 7 arguments?! */
-void
-swfdec_resource_load (SwfdecPlayer *player, const char *target, const char *url, 
-    SwfdecLoaderRequest request, SwfdecBuffer *buffer, SwfdecMovieClipLoader *loader,
-    gboolean target_is_movie)
+/* FIXME: 6 arguments?! */
+static void
+swfdec_resource_load_internal (SwfdecPlayer *player,
+    SwfdecSpriteMovie *target_movie, const char *target_string,
+    const char *url, SwfdecBuffer *buffer, SwfdecMovieClipLoader *loader)
 {
-  SwfdecSpriteMovie *movie;
   SwfdecResourceLoad *load;
 
-  g_return_if_fail (SWFDEC_IS_PLAYER (player));
-  g_return_if_fail (target != NULL);
-  g_return_if_fail (url != NULL);
-  g_return_if_fail (loader == NULL || SWFDEC_IS_MOVIE_CLIP_LOADER (loader));
-
   g_assert (SWFDEC_AS_CONTEXT (player)->frame != NULL);
   load = g_slice_new (SwfdecResourceLoad);
 
-  if (target_is_movie) {
-    movie = (SwfdecSpriteMovie *) swfdec_player_get_movie_from_string (player, target);
-    if (SWFDEC_IS_SPRITE_MOVIE (movie)) {
-      load->target = swfdec_movie_get_path (SWFDEC_MOVIE (movie), TRUE);
-    } else if (swfdec_player_get_level (player, target) >= 0) {
-      load->target = g_strdup (target);
-    } else {
-      SWFDEC_WARNING ("%s does not reference a movie, not loading %s", target, url);
-      return;
-    }
-  } else {
-    load->target = g_strdup (target);
-  }
   load->sandbox = SWFDEC_SANDBOX (SWFDEC_AS_CONTEXT (player)->global);
   load->url = g_strdup (url);
-  load->request = request;
+  load->target_movie = target_movie;
+  load->target_string = g_strdup (target_string);
   load->buffer = buffer;
   load->loader = loader;
-  load->target_is_movie = target_is_movie;
 
-  swfdec_player_root (player, load, swfdec_resource_load_mark);
-  swfdec_player_request_resource (player, swfdec_resource_load_request, load, swfdec_resource_load_free);
+  swfdec_player_root_full (player, load, swfdec_resource_load_mark, swfdec_resource_load_free);
+  swfdec_player_request_resource (player, swfdec_resource_load_request, load, NULL);
+}
+
+gboolean
+swfdec_resource_load_movie (SwfdecPlayer *player, const SwfdecAsValue *target, 
+    const char *url, SwfdecBuffer *buffer, SwfdecMovieClipLoader *loader)
+{
+  SwfdecMovie *movie;
+  const char *s;
+
+  g_return_val_if_fail (SWFDEC_IS_PLAYER (player), FALSE);
+  g_return_val_if_fail (target != NULL, FALSE);
+  g_return_val_if_fail (url != NULL, FALSE);
+  g_return_val_if_fail (loader == NULL || SWFDEC_IS_MOVIE_CLIP_LOADER (loader), FALSE);
+
+  if (SWFDEC_AS_VALUE_IS_OBJECT (target)) {
+    SwfdecAsObject *object = SWFDEC_AS_VALUE_GET_OBJECT (target);
+    if (SWFDEC_IS_SPRITE_MOVIE (object)) {
+      swfdec_resource_load_internal (player, SWFDEC_SPRITE_MOVIE (object),
+	  NULL, url, buffer, loader);
+      return TRUE;
+    }
+  }
+
+  if (loader) {
+    if (SWFDEC_AS_VALUE_IS_NUMBER (target)) {
+      int i = swfdec_as_double_to_integer (SWFDEC_AS_VALUE_GET_NUMBER (target));
+      if (i < 0)
+	return FALSE;
+      s = swfdec_as_context_give_string (swfdec_gc_object_get_context (loader),
+	  g_strdup_printf ("_level%d", i));
+      swfdec_resource_load_internal (player, NULL, s, url, buffer, loader);
+      return TRUE;
+    } else if (SWFDEC_AS_VALUE_IS_STRING (target) ||
+	(SWFDEC_AS_VALUE_IS_OBJECT (target) && SWFDEC_IS_MOVIE (SWFDEC_AS_VALUE_GET_OBJECT (target)))) {
+      s = swfdec_as_value_to_string (SWFDEC_AS_CONTEXT (player), target);
+    } else {
+      SWFDEC_WARNING ("target does not reference a movie, not loading %s", url);
+      return FALSE;
+    }
+  } else {
+    s = swfdec_as_value_to_string (SWFDEC_AS_CONTEXT (player), target);
+  }
+  if (swfdec_player_get_level (player, s, SWFDEC_AS_CONTEXT (player)->version) >= 0) {
+    /* lowercase the string, so we can do case insensitive level lookups later on */
+    char *tmp = g_ascii_strdown (s, -1);
+    swfdec_resource_load_internal (player, NULL, tmp, url, buffer, NULL);
+    g_free (tmp);
+    return TRUE;
+  }
+  movie = swfdec_player_get_movie_from_string (player, s);
+  if (SWFDEC_IS_SPRITE_MOVIE (movie)) {
+    swfdec_resource_load_internal (player, SWFDEC_SPRITE_MOVIE (movie),
+	NULL, url, buffer, loader);
+    return TRUE;
+  }
+  SWFDEC_WARNING ("%s does not reference a movie, not loading %s", s, url);
+  return FALSE;
+}
+
+void
+swfdec_resource_load (SwfdecPlayer *player, const char *target, 
+    const char *url, SwfdecBuffer *buffer)
+{
+  g_return_if_fail (SWFDEC_IS_PLAYER (player));
+  g_return_if_fail (target != NULL);
+  g_return_if_fail (url != NULL);
+
+  swfdec_resource_load_internal (player, NULL, target, url, buffer, NULL);
 }
 
 gboolean
