@@ -1,5 +1,5 @@
 /* Swfdec
- * Copyright (C) 2006 Benjamin Otte <otte@gnome.org>
+ * Copyright (C) 2006-2008 Benjamin Otte <otte@gnome.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -26,6 +26,7 @@
 #include "swfdec_sprite_movie.h"
 #include "swfdec_as_internal.h"
 #include "swfdec_as_strings.h"
+#include "swfdec_audio_swf_stream.h"
 #include "swfdec_audio_event.h"
 #include "swfdec_audio_stream.h"
 #include "swfdec_debug.h"
@@ -66,7 +67,7 @@ static gboolean
 swfdec_sprite_movie_perform_old_place (SwfdecSpriteMovie *movie,
     SwfdecBits *bits, guint tag)
 {
-  SwfdecPlayer *player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (movie)->context);
+  SwfdecPlayer *player = SWFDEC_PLAYER (swfdec_gc_object_get_context (movie));
   SwfdecMovie *mov = SWFDEC_MOVIE (movie);
   SwfdecMovie *cur;
   SwfdecSwfDecoder *dec;
@@ -121,9 +122,12 @@ swfdec_sprite_movie_perform_old_place (SwfdecSpriteMovie *movie,
   cur = swfdec_movie_new (player, depth, mov, mov->resource, graphic, NULL);
   swfdec_movie_set_static_properties (cur, &transform,
       has_ctrans ? &ctrans : NULL, -1, 0, 0, NULL);
-  swfdec_movie_queue_script (cur, SWFDEC_EVENT_INITIALIZE);
-  swfdec_movie_queue_script (cur, SWFDEC_EVENT_CONSTRUCT);
-  swfdec_movie_queue_script (cur, SWFDEC_EVENT_LOAD);
+  if (SWFDEC_IS_ACTOR (cur)) {
+    SwfdecActor *actor = SWFDEC_ACTOR (cur);
+    swfdec_actor_queue_script (actor, SWFDEC_EVENT_INITIALIZE);
+    swfdec_actor_queue_script (actor, SWFDEC_EVENT_CONSTRUCT);
+    swfdec_actor_queue_script (actor, SWFDEC_EVENT_LOAD);
+  }
   swfdec_movie_initialize (cur);
 
   return TRUE;
@@ -133,7 +137,7 @@ swfdec_sprite_movie_perform_old_place (SwfdecSpriteMovie *movie,
 static gboolean
 swfdec_sprite_movie_perform_place (SwfdecSpriteMovie *movie, SwfdecBits *bits, guint tag)
 {
-  SwfdecPlayer *player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (movie)->context);
+  SwfdecPlayer *player = SWFDEC_PLAYER (swfdec_gc_object_get_context (movie));
   SwfdecMovie *mov = SWFDEC_MOVIE (movie);
   SwfdecMovie *cur;
   SwfdecSwfDecoder *dec;
@@ -230,8 +234,12 @@ swfdec_sprite_movie_perform_place (SwfdecSpriteMovie *movie, SwfdecBits *bits, g
 
   if (has_name) {
     char *s = swfdec_bits_get_string (bits, version);
-    name = swfdec_as_context_give_string (SWFDEC_AS_CONTEXT (player), s);
-    SWFDEC_LOG ("  name = %s", name);
+    if (s) {
+      name = swfdec_as_context_give_string (SWFDEC_AS_CONTEXT (player), s);
+      SWFDEC_LOG ("  name = %s", name);
+    } else {
+      name = NULL;
+    }
   } else {
     name = NULL;
   }
@@ -244,7 +252,7 @@ swfdec_sprite_movie_perform_place (SwfdecSpriteMovie *movie, SwfdecBits *bits, g
   }
 
   if (has_filter) {
-    GSList *filters = swfdec_filter_parse (player, bits);
+    GSList *filters = swfdec_filter_parse (bits);
     g_slist_free (filters);
   }
 
@@ -259,7 +267,7 @@ swfdec_sprite_movie_perform_place (SwfdecSpriteMovie *movie, SwfdecBits *bits, g
     int reserved, clip_event_flags, event_flags, key_code;
     char *script_name;
 
-    events = swfdec_event_list_new (player);
+    events = swfdec_event_list_new ();
     reserved = swfdec_bits_get_u16 (bits);
     clip_event_flags = swfdec_get_clipeventflags (mov, bits);
 
@@ -281,7 +289,7 @@ swfdec_sprite_movie_perform_place (SwfdecSpriteMovie *movie, SwfdecBits *bits, g
 
       SWFDEC_INFO ("clip event with flags 0x%X, key code %d", event_flags, key_code);
 #define SWFDEC_UNIMPLEMENTED_EVENTS \
-  ((1<< SWFDEC_EVENT_DATA) | (1<<SWFDEC_EVENT_KEY_PRESS))
+  ((1<< SWFDEC_EVENT_DATA))
       if (event_flags & SWFDEC_UNIMPLEMENTED_EVENTS) {
 	SWFDEC_ERROR ("using non-implemented clip events %u", event_flags & SWFDEC_UNIMPLEMENTED_EVENTS);
       }
@@ -326,9 +334,12 @@ swfdec_sprite_movie_perform_place (SwfdecSpriteMovie *movie, SwfdecBits *bits, g
     cur = swfdec_movie_new (player, depth, mov, mov->resource, graphic, name);
     swfdec_movie_set_static_properties (cur, has_transform ? &transform : NULL, 
 	has_ctrans ? &ctrans : NULL, ratio, clip_depth, blend_mode, events);
-    swfdec_movie_queue_script (cur, SWFDEC_EVENT_INITIALIZE);
-    swfdec_movie_queue_script (cur, SWFDEC_EVENT_CONSTRUCT);
-    swfdec_movie_queue_script (cur, SWFDEC_EVENT_LOAD);
+    if (SWFDEC_IS_ACTOR (cur)) {
+      SwfdecActor *actor = SWFDEC_ACTOR (cur);
+      swfdec_actor_queue_script (actor, SWFDEC_EVENT_INITIALIZE);
+      swfdec_actor_queue_script (actor, SWFDEC_EVENT_CONSTRUCT);
+      swfdec_actor_queue_script (actor, SWFDEC_EVENT_LOAD);
+    }
     swfdec_movie_initialize (cur);
   }
 
@@ -348,7 +359,7 @@ swfdec_sprite_movie_start_sound (SwfdecMovie *movie, SwfdecBits *bits)
   chunk = swfdec_sound_parse_chunk (SWFDEC_SWF_DECODER (movie->resource->decoder), bits, id);
   if (chunk) {
     SwfdecAudio *audio = swfdec_audio_event_new_from_chunk (SWFDEC_PLAYER (
-	  SWFDEC_AS_OBJECT (movie)->context), chunk);
+	  swfdec_gc_object_get_context (movie)), chunk);
     if (audio)
       g_object_unref (audio);
   }
@@ -356,10 +367,11 @@ swfdec_sprite_movie_start_sound (SwfdecMovie *movie, SwfdecBits *bits)
 
 static gboolean
 swfdec_sprite_movie_perform_one_action (SwfdecSpriteMovie *movie, guint tag, SwfdecBuffer *buffer,
-    gboolean skip_scripts, gboolean first_time)
+    gboolean fast_forward, gboolean first_time)
 {
   SwfdecMovie *mov = SWFDEC_MOVIE (movie);
-  SwfdecPlayer *player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (mov)->context);
+  SwfdecActor *actor = SWFDEC_ACTOR (movie);
+  SwfdecPlayer *player = SWFDEC_PLAYER (swfdec_gc_object_get_context (mov));
   SwfdecBits bits;
 
   g_assert (mov->resource);
@@ -368,13 +380,20 @@ swfdec_sprite_movie_perform_one_action (SwfdecSpriteMovie *movie, guint tag, Swf
   SWFDEC_LOG ("%p: executing %uth tag %s in frame %u", movie, movie->next_action - 1, 
       swfdec_swf_decoder_get_tag_name (tag), movie->frame);
   switch (tag) {
+    case SWFDEC_TAG_SETBACKGROUNDCOLOR:
+      swfdec_player_set_background_color (player, swfdec_bits_get_color (&bits));
+      return TRUE;
     case SWFDEC_TAG_DOACTION:
       SWFDEC_LOG ("SCRIPT action");
-      if (!skip_scripts) {
+      if (!fast_forward) {
 	SwfdecScript *script = swfdec_swf_decoder_get_script (
 	    SWFDEC_SWF_DECODER (mov->resource->decoder), buffer->data);
-	g_assert (script);
-	swfdec_player_add_action_script (player, mov, script, 2);
+	if (script) {
+	  swfdec_player_add_action_script (player, actor, script, 
+	      SWFDEC_PLAYER_ACTION_QUEUE_NORMAL);
+	} else {
+	  SWFDEC_ERROR ("Failed to locate script for DoAction tag");
+	}
       }
       return TRUE;
     case SWFDEC_TAG_PLACEOBJECT:
@@ -398,7 +417,8 @@ swfdec_sprite_movie_perform_one_action (SwfdecSpriteMovie *movie, guint tag, Swf
       }
       return TRUE;
     case SWFDEC_TAG_STARTSOUND:
-      swfdec_sprite_movie_start_sound (mov, &bits);
+      if (!fast_forward) 
+	swfdec_sprite_movie_start_sound (mov, &bits);
       return TRUE;
     case SWFDEC_TAG_SHOWFRAME:
       if (movie->frame < movie->n_frames) {
@@ -447,7 +467,6 @@ swfdec_sprite_movie_perform_one_action (SwfdecSpriteMovie *movie, guint tag, Swf
       {
 	guint id;
 	SwfdecSprite *sprite;
-	char *name;
 
 	id = swfdec_bits_get_u16 (&bits);
 	SWFDEC_LOG ("InitAction");
@@ -461,15 +480,29 @@ swfdec_sprite_movie_perform_one_action (SwfdecSpriteMovie *movie, guint tag, Swf
 	  SWFDEC_ERROR ("sprite %u already has an init action", id);
 	  return TRUE;
 	}
-	name = g_strdup_printf ("InitAction %u", id);
-	sprite->init_action = swfdec_script_new_from_bits (&bits, name, 
-	    swfdec_movie_get_version (mov));
-	g_free (name);
+	sprite->init_action = swfdec_script_ref (swfdec_swf_decoder_get_script (
+	    SWFDEC_SWF_DECODER (mov->resource->decoder), buffer->data + 2));
 	if (sprite->init_action) {
-	  swfdec_player_add_action_script (player, mov, sprite->init_action, 0);
+	  swfdec_player_add_action_script (player, actor, sprite->init_action, 
+	      SWFDEC_PLAYER_ACTION_QUEUE_INIT);
+	} else {
+	  SWFDEC_ERROR ("Failed to locate script for InitAction of Sprite %u", id);
 	}
       }
       return TRUE;
+    case SWFDEC_TAG_SOUNDSTREAMHEAD:
+    case SWFDEC_TAG_SOUNDSTREAMHEAD2:
+      /* ignore, those are handled by the sound stream */
+      return TRUE;
+    case SWFDEC_TAG_SOUNDSTREAMBLOCK:
+      if (!fast_forward) {
+	if (movie->sound_stream == NULL) {
+	  movie->sound_stream = swfdec_audio_swf_stream_new (player, movie->sprite,
+	      movie->next_action - 1);
+	}
+	movie->sound_active = TRUE;
+      }
+      return TRUE;
     default:
       g_assert_not_reached ();
       return FALSE;
@@ -513,6 +546,7 @@ swfdec_sprite_movie_goto (SwfdecSpriteMovie *movie, guint goto_frame)
   SwfdecPlayer *player;
   GList *old;
   guint n;
+  gboolean remove_audio;
 
   g_return_if_fail (SWFDEC_IS_SPRITE_MOVIE (movie));
 
@@ -528,7 +562,7 @@ swfdec_sprite_movie_goto (SwfdecSpriteMovie *movie, guint goto_frame)
     return;
   }
 
-  player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (movie)->context);
+  player = SWFDEC_PLAYER (swfdec_gc_object_get_context (movie));
   SWFDEC_LOG ("doing goto %u for %p %d", goto_frame, movie, 
       SWFDEC_CHARACTER (movie->sprite)->id);
 
@@ -553,11 +587,21 @@ swfdec_sprite_movie_goto (SwfdecSpriteMovie *movie, guint goto_frame)
     mov->list = g_list_concat (mov->list, walk);
     n = goto_frame;
     movie->next_action = 0;
+    remove_audio = TRUE;
   } else {
     /* NB: this path is also taken on init */
     old = NULL;
     n = goto_frame - movie->frame;
+    remove_audio = n > 1;
   }
+  /* remove audio after seeks */
+  if (remove_audio && movie->sound_stream) {
+    swfdec_audio_remove (movie->sound_stream);
+    g_object_unref (movie->sound_stream);
+    movie->sound_stream = NULL;
+  }
+  remove_audio = !movie->sound_active;
+  movie->sound_active = FALSE;
   while (n) {
     guint tag;
     gboolean first_time;
@@ -599,8 +643,8 @@ swfdec_sprite_movie_goto (SwfdecSpriteMovie *movie, guint goto_frame)
 	if (klass->replace)
 	  klass->replace (prev, cur->graphic);
 	swfdec_movie_set_static_properties (prev, &cur->original_transform,
-	    &cur->original_ctrans, cur->original_ratio, cur->clip_depth, 
-	    cur->blend_mode, cur->events);
+	    &cur->color_transform, cur->original_ratio, cur->clip_depth, 
+	    cur->blend_mode, SWFDEC_IS_ACTOR (cur) ? SWFDEC_ACTOR (cur)->events : NULL);
 	swfdec_movie_destroy (cur);
 	cur = prev;
 	continue;
@@ -613,11 +657,18 @@ out:
     }
     g_list_free (old);
   }
+
+  /* after two frames without SoundStreamBlock, audio apparently gets removed */
+  if (!movie->sound_active && remove_audio && movie->sound_stream != NULL) {
+    swfdec_audio_remove (movie->sound_stream);
+    g_object_unref (movie->sound_stream);
+    movie->sound_stream = NULL;
+  }
 }
 
 /*** MOVIE ***/
 
-G_DEFINE_TYPE (SwfdecSpriteMovie, swfdec_sprite_movie, SWFDEC_TYPE_MOVIE)
+G_DEFINE_TYPE (SwfdecSpriteMovie, swfdec_sprite_movie, SWFDEC_TYPE_ACTOR)
 
 static void
 swfdec_sprite_movie_dispose (GObject *object)
@@ -639,31 +690,44 @@ swfdec_sprite_movie_init_movie (SwfdecMovie *mov)
   swfdec_sprite_movie_goto (movie, 1);
 }
 
-static void
-swfdec_sprite_movie_add (SwfdecAsObject *object)
+static GObject *
+swfdec_sprite_movie_constructor (GType type, guint n_construct_properties,
+    GObjectConstructParam *construct_properties)
 {
-  SwfdecMovie *movie = SWFDEC_MOVIE (object);
+  GObject *object;
+  SwfdecMovie *movie;
 
-  if (movie->resource->sandbox)
-    swfdec_as_object_set_constructor (object, movie->resource->sandbox->MovieClip);
+  object = G_OBJECT_CLASS (swfdec_sprite_movie_parent_class)->constructor (type, 
+      n_construct_properties, construct_properties);
 
-  SWFDEC_AS_OBJECT_CLASS (swfdec_sprite_movie_parent_class)->add (object);
+  movie = SWFDEC_MOVIE (object);
+  if (movie->resource->sandbox) {
+    swfdec_as_object_set_constructor (SWFDEC_AS_OBJECT (movie), 
+	movie->resource->sandbox->MovieClip);
+  }
+
+  if (movie->graphic) {
+    SwfdecSpriteMovie *smovie = SWFDEC_SPRITE_MOVIE (object);
+    smovie->sprite = SWFDEC_SPRITE (movie->graphic);
+    smovie->n_frames = smovie->sprite->n_frames;
+  }
+
+  return object;
 }
 
 static void
-swfdec_sprite_movie_iterate (SwfdecMovie *mov)
+swfdec_sprite_movie_iterate (SwfdecActor *actor)
 {
-  SwfdecSpriteMovie *movie = SWFDEC_SPRITE_MOVIE (mov);
-  SwfdecPlayer *player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (mov)->context);
+  SwfdecSpriteMovie *movie = SWFDEC_SPRITE_MOVIE (actor);
   guint goto_frame;
 
-  if (mov->state >= SWFDEC_MOVIE_STATE_REMOVED)
+  if (SWFDEC_MOVIE (actor)->state >= SWFDEC_MOVIE_STATE_REMOVED)
     return;
 
   if (movie->sprite && movie->frame == (guint) -1)
     movie->frame = 0;
 
-  swfdec_player_add_action (player, mov, SWFDEC_EVENT_ENTER, 2);
+  swfdec_actor_queue_script (actor, SWFDEC_EVENT_ENTER);
   if (movie->playing && movie->sprite != NULL) {
     if (movie->frame == movie->n_frames)
       goto_frame = 1;
@@ -675,73 +739,13 @@ swfdec_sprite_movie_iterate (SwfdecMovie *mov)
   }
 }
 
-/* FIXME: This function is a mess */
-static gboolean
-swfdec_sprite_movie_iterate_end (SwfdecMovie *mov)
-{
-  SwfdecSpriteMovie *movie = SWFDEC_SPRITE_MOVIE (mov);
-  SwfdecSpriteFrame *last;
-  SwfdecSpriteFrame *current;
-  SwfdecPlayer *player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (mov)->context);
-
-  if (!SWFDEC_MOVIE_CLASS (swfdec_sprite_movie_parent_class)->iterate_end (mov))
-    return FALSE;
-  
-  if (movie->sprite == NULL)
-    return TRUE;
-  g_assert (movie->frame <= movie->n_frames);
-  if (movie->frame == 0)
-    return TRUE;
-  current = &movie->sprite->frames[movie->frame - 1];
-
-  /* then do the streaming thing */
-  if (current->sound_head == NULL ||
-      !movie->playing) {
-    if (movie->sound_stream) {
-      swfdec_audio_remove (movie->sound_stream);
-      g_object_unref (movie->sound_stream);
-      movie->sound_stream = NULL;
-    }
-    goto exit;
-  }
-  if (movie->sound_stream == NULL && current->sound_block == NULL)
-    goto exit;
-  SWFDEC_LOG ("iterating audio (from %u to %u)", movie->sound_frame, movie->frame);
-  if (movie->sound_frame + 1 != movie->frame)
-    goto new_decoder;
-  if (movie->sound_frame == (guint) -1)
-    goto new_decoder;
-  if (current->sound_head && movie->sound_stream == NULL)
-    goto new_decoder;
-  last = &movie->sprite->frames[movie->sound_frame];
-  if (last->sound_head != current->sound_head)
-    goto new_decoder;
-exit:
-  movie->sound_frame = movie->frame;
-  return TRUE;
-
-new_decoder:
-  if (movie->sound_stream) {
-    swfdec_audio_remove (movie->sound_stream);
-    g_object_unref (movie->sound_stream);
-    movie->sound_stream = NULL;
-  }
-
-  if (current->sound_block) {
-    movie->sound_stream = swfdec_audio_stream_new (player, 
-	movie->sprite, movie->frame - 1);
-    movie->sound_frame = movie->frame;
-  }
-  return TRUE;
-}
-
 static void
 swfdec_sprite_movie_finish_movie (SwfdecMovie *mov)
 {
   SwfdecSpriteMovie *movie = SWFDEC_SPRITE_MOVIE (mov);
-  SwfdecPlayer *player = SWFDEC_PLAYER (SWFDEC_AS_OBJECT (mov)->context);
+  SwfdecPlayer *player = SWFDEC_PLAYER (swfdec_gc_object_get_context (mov));
 
-  swfdec_player_remove_all_actions (player, mov);
+  swfdec_player_remove_all_actions (player, SWFDEC_ACTOR (mov));
   if (movie->sound_stream) {
     swfdec_audio_remove (movie->sound_stream);
     g_object_unref (movie->sound_stream);
@@ -750,35 +754,58 @@ swfdec_sprite_movie_finish_movie (SwfdecMovie *mov)
 }
 
 static void
-swfdec_sprite_movie_mark (SwfdecAsObject *object)
+swfdec_sprite_movie_mark (SwfdecGcObject *object)
 {
   GList *walk;
 
   for (walk = SWFDEC_MOVIE (object)->list; walk; walk = walk->next) {
     SwfdecAsObject *child = walk->data;
     g_assert (child->properties != NULL);
-    swfdec_as_object_mark (child);
+    swfdec_gc_object_mark (child);
   }
 
-  SWFDEC_AS_OBJECT_CLASS (swfdec_sprite_movie_parent_class)->mark (object);
+  SWFDEC_GC_OBJECT_CLASS (swfdec_sprite_movie_parent_class)->mark (object);
+}
+
+static void
+swfdec_sprite_movie_property_get (SwfdecMovie *mov, guint prop_id, SwfdecAsValue *val)
+{
+  SwfdecSpriteMovie *movie = SWFDEC_SPRITE_MOVIE (mov);
+
+  switch (prop_id) {
+    case SWFDEC_MOVIE_PROPERTY_CURRENTFRAME:
+      SWFDEC_AS_VALUE_SET_INT (val, movie->frame);
+      break;
+    case SWFDEC_MOVIE_PROPERTY_FRAMESLOADED:
+      SWFDEC_AS_VALUE_SET_INT (val, swfdec_sprite_movie_get_frames_loaded (movie));
+      break;
+    case SWFDEC_MOVIE_PROPERTY_TOTALFRAMES:
+      SWFDEC_AS_VALUE_SET_INT (val, swfdec_sprite_movie_get_frames_total (movie));
+      break;
+    default:
+      SWFDEC_MOVIE_CLASS (swfdec_sprite_movie_parent_class)->property_get (mov, prop_id, val);
+      break;
+  }
 }
 
 static void
 swfdec_sprite_movie_class_init (SwfdecSpriteMovieClass * g_class)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (g_class);
-  SwfdecAsObjectClass *asobject_class = SWFDEC_AS_OBJECT_CLASS (g_class);
+  SwfdecGcObjectClass *gc_class = SWFDEC_GC_OBJECT_CLASS (g_class);
   SwfdecMovieClass *movie_class = SWFDEC_MOVIE_CLASS (g_class);
+  SwfdecActorClass *actor_class = SWFDEC_ACTOR_CLASS (g_class);
 
   object_class->dispose = swfdec_sprite_movie_dispose;
+  object_class->constructor = swfdec_sprite_movie_constructor;
 
-  asobject_class->add = swfdec_sprite_movie_add;
-  asobject_class->mark = swfdec_sprite_movie_mark;
+  gc_class->mark = swfdec_sprite_movie_mark;
 
   movie_class->init_movie = swfdec_sprite_movie_init_movie;
   movie_class->finish_movie = swfdec_sprite_movie_finish_movie;
-  movie_class->iterate_start = swfdec_sprite_movie_iterate;
-  movie_class->iterate_end = swfdec_sprite_movie_iterate_end;
+  movie_class->property_get = swfdec_sprite_movie_property_get;
+  
+  actor_class->iterate_start = swfdec_sprite_movie_iterate;
 }
 
 static void
@@ -808,7 +835,7 @@ swfdec_sprite_movie_unload (SwfdecSpriteMovie *movie)
   g_return_if_fail (SWFDEC_IS_SPRITE_MOVIE (movie));
 
   mov = SWFDEC_MOVIE (movie);
-  swfdec_sprite_movie_clear (SWFDEC_AS_OBJECT (movie)->context, 
+  swfdec_sprite_movie_clear (swfdec_gc_object_get_context (movie), 
       SWFDEC_AS_OBJECT (movie), 0, NULL, &hack);
   /* FIXME: destroy or unload? */
   while (mov->list)
