@@ -1,7 +1,7 @@
 /* Swfdec
  * Copyright (C) 2003-2006 David Schleef <ds@schleef.org>
  *		 2005-2006 Eric Anholt <eric@anholt.net>
- *		 2006-2007 Benjamin Otte <otte@gnome.org>
+ *		 2006-2008 Benjamin Otte <otte@gnome.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -26,6 +26,8 @@
 #include "swfdec_movie.h"
 #include "swfdec_as_internal.h"
 #include "swfdec_as_strings.h"
+#include "swfdec_bitmap_data.h"
+#include "swfdec_bitmap_movie.h"
 #include "swfdec_bits.h"
 #include "swfdec_debug.h"
 #include "swfdec_decoder.h"
@@ -34,6 +36,7 @@
 #include "swfdec_sprite.h"
 #include "swfdec_sprite_movie.h"
 #include "swfdec_swf_decoder.h"
+#include "swfdec_transform_as.h"
 #include "swfdec_resource.h"
 #include "swfdec_utils.h"
 #include "swfdec_as_internal.h"
@@ -59,7 +62,11 @@ void
 swfdec_sprite_movie_get__lockroot (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *rval)
 {
-  SWFDEC_STUB ("MovieClip._lockroot (get)");
+  SwfdecMovie *movie;
+
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, &movie, "");
+
+  SWFDEC_AS_VALUE_SET_BOOLEAN (rval, movie->lockroot);
 }
 
 SWFDEC_AS_NATIVE (900, 301, swfdec_sprite_movie_set__lockroot)
@@ -67,7 +74,12 @@ void
 swfdec_sprite_movie_set__lockroot (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *rval)
 {
-  SWFDEC_STUB ("MovieClip._lockroot (set)");
+  SwfdecMovie *movie;
+  gboolean lockroot;
+
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, &movie, "b", &lockroot);
+
+  movie->lockroot = lockroot;
 }
 
 SWFDEC_AS_NATIVE (900, 401, swfdec_sprite_movie_get_cacheAsBitmap)
@@ -139,7 +151,12 @@ void
 swfdec_sprite_movie_get_transform (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *rval)
 {
-  SWFDEC_STUB ("MovieClip.transform (get)");
+  SwfdecMovie *movie;
+
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, &movie, "");
+
+  SWFDEC_AS_VALUE_SET_OBJECT (rval,
+      SWFDEC_AS_OBJECT (swfdec_transform_as_new (cx, movie)));
 }
 
 SWFDEC_AS_NATIVE (900, 420, swfdec_sprite_movie_set_transform)
@@ -306,8 +323,11 @@ swfdec_sprite_movie_getInstanceAtDepth (SwfdecAsContext *cx,
     return;
 
   movie = swfdec_movie_find (movie, depth);
-  if (movie != NULL)
+  if (movie != NULL) {
+    if (!swfdec_movie_is_scriptable (movie))
+      movie = movie->parent;
     SWFDEC_AS_VALUE_SET_OBJECT (rval, SWFDEC_AS_OBJECT (movie));
+  }
 }
 
 SWFDEC_AS_NATIVE (900, 24, swfdec_sprite_movie_getSWFVersion)
@@ -328,7 +348,28 @@ void
 swfdec_sprite_movie_attachBitmap (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *rval)
 {
-  SWFDEC_STUB ("MovieClip.attachBitmap");
+  SwfdecMovie *parent;
+  SwfdecBitmapData *bitmap;
+  const char *snapping;
+  gboolean smoothing;
+  int depth;
+  SwfdecMovie *movie;
+
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_SPRITE_MOVIE, &parent, "oi|sb", 
+      &bitmap, &depth, &snapping, &smoothing);
+
+  if (!SWFDEC_IS_BITMAP_DATA (bitmap))
+    return;
+  if (swfdec_depth_classify (depth) == SWFDEC_DEPTH_CLASS_EMPTY)
+    return;
+
+  movie = swfdec_movie_find (parent, depth);
+  if (movie)
+    swfdec_movie_remove (movie);
+
+  swfdec_bitmap_movie_new (parent, bitmap, depth);
+  SWFDEC_LOG ("created new BitmapMovie to parent %s at depth %d", 
+      parent->name, depth);
 }
 
 SWFDEC_AS_NATIVE (900, 26, swfdec_sprite_movie_getRect)
@@ -439,7 +480,7 @@ swfdec_sprite_movie_do_goto (SwfdecSpriteMovie *movie, SwfdecAsValue *target)
       return;
     frame++;
   } else {
-    frame = swfdec_as_value_to_integer (SWFDEC_AS_OBJECT (movie)->context, target);
+    frame = swfdec_as_value_to_integer (swfdec_gc_object_get_context (movie), target);
   }
   /* FIXME: how to handle overflow? */
   frame = CLAMP (frame, 1, (int) movie->n_frames);
@@ -508,17 +549,18 @@ swfdec_sprite_movie_hitTest (SwfdecAsContext *cx, SwfdecAsObject *object,
 {
   SwfdecMovie *movie;
 
-  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, (gpointer)&movie, "");
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, &movie, "");
   
   if (argc == 1) {
     SwfdecMovie *other;
     SwfdecRect movie_rect, other_rect;
 
-    other = swfdec_player_get_movie_from_value (SWFDEC_PLAYER (cx), &argv[0]);
-    if (other == NULL) {
+    if (!SWFDEC_AS_VALUE_IS_OBJECT (&argv[0]) ||
+	!SWFDEC_IS_MOVIE (SWFDEC_AS_VALUE_GET_OBJECT (&argv[0]))) {
       SWFDEC_AS_VALUE_SET_BOOLEAN (rval, FALSE);
       return;
     }
+    other = SWFDEC_MOVIE (SWFDEC_AS_VALUE_GET_OBJECT (&argv[0]));
     swfdec_movie_update (movie);
     swfdec_movie_update (other);
     movie_rect = movie->extents;
@@ -536,11 +578,14 @@ swfdec_sprite_movie_hitTest (SwfdecAsContext *cx, SwfdecAsObject *object,
     y = swfdec_as_value_to_number (cx, &argv[1]) * SWFDEC_TWIPS_SCALE_FACTOR;
     shape = (argc >= 3 && swfdec_as_value_to_boolean (cx, &argv[2]));
 
-    swfdec_movie_global_to_local (movie, &x, &y);
-
     if (shape) {
+      if (movie->parent)
+	swfdec_movie_global_to_local (movie->parent, &x, &y);
       ret = swfdec_movie_contains (movie, x, y);
     } else {
+      if (movie->cache_state >= SWFDEC_MOVIE_INVALID_EXTENTS)
+	  swfdec_movie_update (movie);
+      swfdec_movie_global_to_local (movie, &x, &y);
       ret = swfdec_rect_contains (&movie->original_extents, x, y);
     }
     SWFDEC_AS_VALUE_SET_BOOLEAN (rval, ret);
@@ -552,11 +597,11 @@ void
 swfdec_sprite_movie_startDrag (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *rval)
 {
-  SwfdecMovie *movie;
+  SwfdecActor *actor;
   SwfdecPlayer *player = SWFDEC_PLAYER (cx);
   gboolean center = FALSE;
 
-  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, (gpointer)&movie, "");
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_ACTOR, &actor, "");
 
   if (argc > 0) {
     center = swfdec_as_value_to_boolean (cx, &argv[0]);
@@ -568,9 +613,9 @@ swfdec_sprite_movie_startDrag (SwfdecAsContext *cx, SwfdecAsObject *object,
     rect.x1 = swfdec_as_value_to_number (cx, &argv[3]);
     rect.y1 = swfdec_as_value_to_number (cx, &argv[4]);
     swfdec_rect_scale (&rect, &rect, SWFDEC_TWIPS_SCALE_FACTOR);
-    swfdec_player_set_drag_movie (player, movie, center, &rect);
+    swfdec_player_set_drag_movie (player, actor, center, &rect);
   } else {
-    swfdec_player_set_drag_movie (player, movie, center, NULL);
+    swfdec_player_set_drag_movie (player, actor, center, NULL);
   }
 }
 
@@ -589,23 +634,23 @@ void
 swfdec_sprite_movie_swapDepths (SwfdecAsContext *cx, SwfdecAsObject *object,
     guint argc, SwfdecAsValue *argv, SwfdecAsValue *rval)
 {
-  SwfdecMovie *movie;
-  SwfdecMovie *other;
+  SwfdecMovie *movie, *other;
+  SwfdecAsValue value;
   int depth;
 
-  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, (gpointer)&movie, "");
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, (gpointer)&movie, "v", &value);
 
-  if (argc < 1)
-    return;
+  if (movie->parent == NULL)
+    SWFDEC_FIXME ("swapDepths on root movie, should do something weird");
 
-  if (SWFDEC_AS_VALUE_IS_OBJECT (&argv[0])) {
-    other = (SwfdecMovie *) SWFDEC_AS_VALUE_GET_OBJECT (&argv[0]);
+  if (SWFDEC_AS_VALUE_IS_OBJECT (&value)) {
+    other = (SwfdecMovie *) SWFDEC_AS_VALUE_GET_OBJECT (&value);
     if (!SWFDEC_IS_MOVIE (other) ||
 	other->parent != movie->parent)
       return;
     depth = other->depth;
   } else {
-    depth = swfdec_as_value_to_integer (cx, &argv[0]);
+    depth = swfdec_as_value_to_integer (cx, &value);
     if (movie->parent) {
       other = swfdec_movie_find (movie->parent, depth);
     } else {
@@ -613,6 +658,13 @@ swfdec_sprite_movie_swapDepths (SwfdecAsContext *cx, SwfdecAsObject *object,
       other = NULL;
     }
   }
+
+  // FIXME: one different than the reserved range, should the classify function
+  // be changed instead?
+  if (swfdec_depth_classify (depth) == SWFDEC_DEPTH_CLASS_EMPTY ||
+      depth >= 2130690045)
+    return;
+
   if (other)
     swfdec_movie_set_depth (other, movie->depth);
   swfdec_movie_set_depth (movie, depth);
@@ -633,16 +685,33 @@ swfdec_sprite_movie_createEmptyMovieClip (SwfdecAsContext *cx, SwfdecAsObject *o
   if (movie)
     swfdec_movie_remove (movie);
   movie = swfdec_movie_new (SWFDEC_PLAYER (cx), depth, parent, parent->resource, NULL, name);
-  swfdec_movie_initialize (movie);
+
+  if (SWFDEC_IS_SPRITE_MOVIE (movie)) {
+    SwfdecSandbox *sandbox = SWFDEC_SANDBOX (cx->global);
+    SwfdecActor *actor = SWFDEC_ACTOR (movie);
+    swfdec_sandbox_unuse (sandbox);
+    swfdec_movie_initialize (movie);
+    swfdec_actor_execute (actor, SWFDEC_EVENT_CONSTRUCT, 0);
+    swfdec_sandbox_use (sandbox);
+  } else {
+    swfdec_movie_initialize (movie);
+  }
+
   SWFDEC_AS_VALUE_SET_OBJECT (rval, SWFDEC_AS_OBJECT (movie));
 }
 
 static void
 swfdec_sprite_movie_copy_props (SwfdecMovie *target, SwfdecMovie *src)
 {
-  swfdec_movie_queue_update (target, SWFDEC_MOVIE_INVALID_MATRIX);
+  swfdec_movie_begin_update_matrix (target);
   target->matrix = src->matrix;
+  target->modified = src->modified;
+  target->xscale = src->xscale;
+  target->yscale = src->yscale;
+  target->rotation = src->rotation;
+  target->lockroot = src->lockroot;
   target->color_transform = src->color_transform;
+  swfdec_movie_end_update_matrix (target);
 }
 
 static gboolean
@@ -653,6 +722,11 @@ swfdec_sprite_movie_foreach_copy_properties (SwfdecAsObject *object,
 
   g_return_val_if_fail (SWFDEC_IS_AS_OBJECT (target), FALSE);
 
+  /* FIXME: We likely need better flag handling here.
+   * We might even want to fix swfdec_as_object_foreach() */
+  if (flags & SWFDEC_AS_VARIABLE_HIDDEN)
+    return TRUE;
+
   swfdec_as_object_set_variable (target, variable, value);
 
   return TRUE;
@@ -666,20 +740,28 @@ swfdec_sprite_movie_init_from_object (SwfdecMovie *movie,
   g_return_if_fail (initObject == NULL || SWFDEC_IS_AS_OBJECT (initObject));
 
   if (initObject != NULL) {
-    swfdec_as_object_foreach (initObject,
-	swfdec_sprite_movie_foreach_copy_properties, SWFDEC_AS_OBJECT (movie));
+    SwfdecAsContext *cx = swfdec_gc_object_get_context (movie);
+    if (cx->version <= 6) {
+      swfdec_as_object_foreach (initObject,
+	  swfdec_sprite_movie_foreach_copy_properties, SWFDEC_AS_OBJECT (movie));
+      swfdec_movie_initialize (movie);
+    } else {
+      swfdec_movie_initialize (movie);
+      swfdec_as_object_foreach (initObject,
+	  swfdec_sprite_movie_foreach_copy_properties, SWFDEC_AS_OBJECT (movie));
+    }
+  } else {
+    swfdec_movie_initialize (movie);
   }
 
   if (SWFDEC_IS_SPRITE_MOVIE (movie)) {
     SwfdecSandbox *sandbox = movie->resource->sandbox;
-    swfdec_movie_queue_script (movie, SWFDEC_EVENT_INITIALIZE);
-    swfdec_movie_queue_script (movie, SWFDEC_EVENT_LOAD);
+    SwfdecActor *actor = SWFDEC_ACTOR (movie);
+    swfdec_actor_queue_script (actor, SWFDEC_EVENT_INITIALIZE);
+    swfdec_actor_queue_script (actor, SWFDEC_EVENT_LOAD);
     swfdec_sandbox_unuse (sandbox);
-    swfdec_movie_initialize (movie);
-    swfdec_movie_execute (movie, SWFDEC_EVENT_CONSTRUCT);
+    swfdec_actor_execute (actor, SWFDEC_EVENT_CONSTRUCT, 0);
     swfdec_sandbox_use (sandbox);
-  } else {
-    swfdec_movie_initialize (movie);
   }
 }
 
@@ -690,12 +772,12 @@ swfdec_sprite_movie_attachMovie (SwfdecAsContext *cx, SwfdecAsObject *object,
 {
   SwfdecMovie *movie;
   SwfdecMovie *ret;
-  SwfdecAsObject *initObject;
+  SwfdecAsObject *initObject, *constructor;
   const char *name, *export;
   int depth;
   SwfdecGraphic *sprite;
 
-  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, (gpointer)&movie, "ssi", &export, &name, &depth);
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, &movie, "ssi", &export, &name, &depth);
 
   if (argc > 3 && SWFDEC_AS_VALUE_IS_OBJECT (&argv[3])) {
     initObject = SWFDEC_AS_VALUE_GET_OBJECT ((&argv[3]));
@@ -716,10 +798,16 @@ swfdec_sprite_movie_attachMovie (SwfdecAsContext *cx, SwfdecAsObject *object,
   ret = swfdec_movie_find (movie, depth);
   if (ret)
     swfdec_movie_remove (ret);
-  ret = swfdec_movie_new (SWFDEC_PLAYER (object->context), depth, movie, movie->resource, sprite, name);
+  ret = swfdec_movie_new (SWFDEC_PLAYER (swfdec_gc_object_get_context (object)),
+      depth, movie, movie->resource, sprite, name);
   SWFDEC_LOG ("attached %s (%u) as %s to depth %u", export, SWFDEC_CHARACTER (sprite)->id,
       ret->name, ret->depth);
   /* run init and construct */
+  constructor = swfdec_player_get_export_class (SWFDEC_PLAYER (cx), export);
+  if (constructor == NULL)
+    constructor = movie->resource->sandbox->MovieClip;
+  swfdec_as_object_set_constructor (SWFDEC_AS_OBJECT (ret), constructor);
+
   swfdec_sprite_movie_init_from_object (ret, initObject);
   SWFDEC_AS_VALUE_SET_OBJECT (rval, SWFDEC_AS_OBJECT (ret));
 }
@@ -781,11 +869,9 @@ swfdec_sprite_movie_getBounds (SwfdecAsContext *cx, SwfdecAsObject *object,
   SwfdecAsObject *obj;
   SwfdecMovie *movie;
 
-  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, (gpointer)&movie, "");
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, &movie, "");
 
   obj= swfdec_as_object_new_empty (cx);
-  if (obj== NULL)
-    return;
 
   swfdec_movie_update (movie);
   if (swfdec_rect_is_empty (&movie->extents)) {
@@ -805,7 +891,7 @@ swfdec_sprite_movie_getBounds (SwfdecAsContext *cx, SwfdecAsObject *object,
 
     if (movie->parent)
       swfdec_movie_rect_local_to_global (movie->parent, &rect);
-    swfdec_movie_rect_global_to_local ((other != NULL ? other : movie), &rect);
+    swfdec_movie_rect_global_to_local (other, &rect);
 
     x0 = rect.x0;
     y0 = rect.y0;
@@ -878,12 +964,8 @@ swfdec_sprite_movie_init_context (SwfdecPlayer *player)
 
   movie = SWFDEC_AS_OBJECT (swfdec_as_object_add_function (context->global, 
       SWFDEC_AS_STR_MovieClip, 0, NULL, 0));
-  if (movie == NULL)
-    return;
   SWFDEC_SANDBOX (context->global)->MovieClip = movie;
   proto = swfdec_as_object_new (context);
-  if (!proto)
-    return;
   SWFDEC_AS_VALUE_SET_OBJECT (&val, proto);
   swfdec_as_object_set_variable_and_flags (movie,
       SWFDEC_AS_STR_prototype, &val, SWFDEC_AS_VARIABLE_HIDDEN |
