@@ -74,12 +74,14 @@ tag_func_frame_label (SwfdecSwfDecoder * s, guint tag)
 {
   SwfdecSpriteFrame *frame = &s->parse_sprite->frames[s->parse_sprite->parse_frame];
   
-  if (frame->label) {
-    SWFDEC_WARNING ("frame %d already has a label (%s)", s->parse_sprite->parse_frame, frame->label);
-    g_free (frame->label);
+  if (frame->labels) {
+    SWFDEC_WARNING ("adding another label for frame %d (%s)",
+        s->parse_sprite->parse_frame, (char *)frame->labels->data);
   }
-  frame->label = swfdec_bits_get_string (&s->b, s->version);
-  SWFDEC_LOG ("frame %d named %s", s->parse_sprite->parse_frame, frame->label);
+  frame->labels = g_slist_prepend (frame->labels,
+      swfdec_bits_get_string (&s->b, s->version));
+  SWFDEC_LOG ("frame %d named %s", s->parse_sprite->parse_frame,
+      (char *)frame->labels->data);
 
   return SWFDEC_STATUS_OK;
 }
@@ -201,7 +203,7 @@ tag_func_define_sprite (SwfdecSwfDecoder * s, guint define_sprite_tag)
     SwfdecTagFunc func;
 
     x = swfdec_bits_get_u16 (&parse);
-    tag = (x >> 6) & 0x3ff;
+    tag = x >> 6;
     tag_len = x & 0x3f;
     if (tag_len == 0x3f) {
       tag_len = swfdec_bits_get_u32 (&parse);
@@ -252,190 +254,6 @@ tag_func_define_sprite (SwfdecSwfDecoder * s, guint define_sprite_tag)
   return SWFDEC_STATUS_OK;
 }
 
-#if 0
-static int
-tag_func_define_button_2 (SwfdecSwfDecoder * s, guint tag)
-{
-  SwfdecBits bits;
-  int id, reserved;
-  guint length;
-  SwfdecButton *button;
-  char *script_name;
-
-  id = swfdec_bits_get_u16 (&s->b);
-  button = swfdec_swf_decoder_create_character (s, id, SWFDEC_TYPE_BUTTON);
-  if (!button)
-    return SWFDEC_STATUS_OK;
-
-  SWFDEC_LOG ("  ID: %d", id);
-
-  reserved = swfdec_bits_getbits (&s->b, 7);
-  button->menubutton = swfdec_bits_getbit (&s->b) ? TRUE : FALSE;
-  length = swfdec_bits_get_u16 (&s->b);
-
-  SWFDEC_LOG ("  reserved = %d", reserved);
-  SWFDEC_LOG ("  menu = %d", button->menubutton);
-  SWFDEC_LOG ("  length of region = %d", length);
-
-  if (length)
-    swfdec_bits_init_bits (&bits, &s->b, length > 2 ? length - 2 : 0);
-  else
-    swfdec_bits_init_bits (&bits, &s->b, swfdec_bits_left (&s->b) / 8);
-  while (swfdec_bits_peek_u8 (&bits)) {
-    guint character;
-    guint depth;
-    guint states;
-    gboolean has_blend_mode, has_filters;
-    SwfdecContent *content;
-
-    if (s->version >= 8) {
-      reserved = swfdec_bits_getbits (&bits, 2);
-      has_blend_mode = swfdec_bits_getbit (&bits);
-      has_filters = swfdec_bits_getbit (&bits);
-      SWFDEC_LOG ("  reserved = %d", reserved);
-      SWFDEC_LOG ("  has_blend_mode = %d", has_blend_mode);
-      SWFDEC_LOG ("  has_filters = %d", has_filters);
-    } else {
-      reserved = swfdec_bits_getbits (&bits, 4);
-      has_blend_mode = 0;
-      has_filters = 0;
-      SWFDEC_LOG ("  reserved = %d", reserved);
-    }
-    states = swfdec_bits_getbits (&bits, 4);
-    character = swfdec_bits_get_u16 (&bits);
-    depth = swfdec_bits_get_u16 (&bits);
-
-    SWFDEC_LOG ("  states: %s%s%s%scharacter=%u layer=%u",
-        states & (1 << SWFDEC_BUTTON_HIT) ? "HIT " : "", 
-	states & (1 << SWFDEC_BUTTON_DOWN) ? "DOWN " : "", 
-        states & (1 << SWFDEC_BUTTON_OVER) ? "OVER " : "",
-	states & (1 << SWFDEC_BUTTON_UP) ? "UP " : "", 
-	character, depth);
-    content = swfdec_content_new (depth);
-
-    swfdec_bits_get_matrix (&bits, &content->transform, NULL);
-    content->has_transform = TRUE;
-    SWFDEC_LOG ("matrix: %g %g  %g %g   %g %g",
-	content->transform.xx, content->transform.yy, 
-	content->transform.xy, content->transform.yx,
-	content->transform.x0, content->transform.y0);
-    swfdec_bits_get_color_transform (&bits, &content->color_transform);
-    content->has_color_transform = TRUE;
-
-    content->graphic = swfdec_swf_decoder_get_character (s, character);
-    if (has_blend_mode) {
-      content->blend_mode = swfdec_bits_get_u8 (&bits);
-      SWFDEC_LOG ("  blend mode = %u", content->blend_mode);
-    }
-    if (has_filters) {
-      GSList *list = swfdec_filter_parse (SWFDEC_DECODER (s)->player, &bits);
-      g_slist_free (list);
-    }
-    if (!SWFDEC_IS_GRAPHIC (content->graphic)) {
-      SWFDEC_ERROR ("id %u does not reference a graphic, ignoring", character);
-      swfdec_content_free (content);
-    } else {
-      swfdec_button_append_content (button, states, content);
-    }
-  }
-  swfdec_bits_get_u8 (&bits);
-  if (swfdec_bits_left (&bits)) {
-    SWFDEC_WARNING ("%u bytes left when parsing button records", swfdec_bits_left (&bits) / 8);
-  }
-
-  script_name = g_strdup_printf ("Button%u", SWFDEC_CHARACTER (button)->id);
-  while (length != 0) {
-    guint condition, key;
-
-    length = swfdec_bits_get_u16 (&s->b);
-    if (length)
-      swfdec_bits_init_bits (&bits, &s->b, length > 2 ? length - 2 : 0);
-    else
-      swfdec_bits_init_bits (&bits, &s->b, swfdec_bits_left (&s->b) / 8);
-    condition = swfdec_bits_get_u16 (&bits);
-    key = condition >> 9;
-    condition &= 0x1FF;
-
-    SWFDEC_LOG (" length = %d", length);
-
-    if (button->events == NULL)
-      button->events = swfdec_event_list_new (SWFDEC_DECODER (s)->player);
-    SWFDEC_LOG ("  new event for condition %u (key %u)", condition, key);
-    swfdec_event_list_parse (button->events, &bits, s->version, condition, key,
-	script_name);
-    if (swfdec_bits_left (&bits)) {
-      SWFDEC_WARNING ("%u bytes left after parsing script", swfdec_bits_left (&bits) / 8);
-    }
-  }
-  g_free (script_name);
-
-  return SWFDEC_STATUS_OK;
-}
-
-static int
-tag_func_define_button (SwfdecSwfDecoder * s, guint tag)
-{
-  SwfdecBits *bits = &s->b;
-  int id;
-  SwfdecButton *button;
-  char *script_name;
-
-  id = swfdec_bits_get_u16 (bits);
-  button = swfdec_swf_decoder_create_character (s, id, SWFDEC_TYPE_BUTTON);
-  if (!button)
-    return SWFDEC_STATUS_OK;
-
-  SWFDEC_LOG ("  ID: %d", id);
-
-  while (swfdec_bits_peek_u8 (bits)) {
-    int reserved;
-    guint character;
-    guint depth;
-    guint states;
-    SwfdecContent *content;
-
-    swfdec_bits_syncbits (bits);
-    reserved = swfdec_bits_getbits (bits, 4);
-    states = swfdec_bits_getbits (bits, 4);
-    character = swfdec_bits_get_u16 (bits);
-    depth = swfdec_bits_get_u16 (bits);
-
-    SWFDEC_LOG ("  reserved = %d", reserved);
-    SWFDEC_LOG ("states: %s%s%s%scharacter=%u layer=%u",
-        states & (1 << SWFDEC_BUTTON_HIT) ? "HIT " : "", 
-	states & (1 << SWFDEC_BUTTON_DOWN) ? "DOWN " : "", 
-        states & (1 << SWFDEC_BUTTON_OVER) ? "OVER " : "",
-	states & (1 << SWFDEC_BUTTON_UP) ? "UP " : "", 
-	character, depth);
-    content = swfdec_content_new (depth);
-
-    swfdec_bits_get_matrix (bits, &content->transform, NULL);
-    content->has_transform = TRUE;
-    SWFDEC_LOG ("matrix: %g %g  %g %g   %g %g",
-	content->transform.xx, content->transform.yy, 
-	content->transform.xy, content->transform.yx,
-	content->transform.x0, content->transform.y0);
-
-    content->graphic = swfdec_swf_decoder_get_character (s, character);
-    if (!SWFDEC_IS_GRAPHIC (content->graphic)) {
-      SWFDEC_ERROR ("id %u does not reference a graphic, ignoring", character);
-      swfdec_content_free (content);
-    } else {
-      swfdec_button_append_content (button, states, content);
-    }
-  }
-  swfdec_bits_get_u8 (bits);
-
-  script_name = g_strdup_printf ("Button%u", SWFDEC_CHARACTER (button)->id);
-  button->events = swfdec_event_list_new (SWFDEC_DECODER (s)->player);
-  swfdec_event_list_parse (button->events, &s->b, s->version, 
-      SWFDEC_BUTTON_OVER_UP_TO_OVER_DOWN, 0, script_name);
-  g_free (script_name);
-
-  return SWFDEC_STATUS_OK;
-}
-#endif
-
 static int
 tag_func_file_attributes (SwfdecSwfDecoder *s, guint tag)
 {
@@ -474,12 +292,6 @@ tag_func_show_frame (SwfdecSwfDecoder * s, guint tag)
       SWFDEC_CHARACTER (s->parse_sprite)->id);
 
   s->parse_sprite->parse_frame++;
-  if (s->parse_sprite->parse_frame < s->parse_sprite->n_frames) {
-    SwfdecSpriteFrame *old = &s->parse_sprite->frames[s->parse_sprite->parse_frame - 1];
-    SwfdecSpriteFrame *new = &s->parse_sprite->frames[s->parse_sprite->parse_frame];
-    if (old->sound_head)
-      new->sound_head = g_object_ref (old->sound_head);
-  }
   tag_func_enqueue (s, tag);
 
   return SWFDEC_STATUS_IMAGE;
@@ -509,6 +321,81 @@ tag_func_do_action (SwfdecSwfDecoder * s, guint tag)
   return SWFDEC_STATUS_OK;
 }
 
+static int
+tag_func_do_init_action (SwfdecSwfDecoder * s, guint tag)
+{
+  SwfdecScript *script;
+  SwfdecBits bits;
+  char *name;
+  guint id;
+
+  bits = s->b;
+  id = swfdec_bits_get_u16 (&bits);
+  if (swfdec_bits_left (&bits) == 0) {
+    SWFDEC_WARNING ("empty script, ignoring");
+    return SWFDEC_STATUS_OK;
+  }
+  name = g_strdup_printf ("Init %u", id);
+  script = swfdec_script_new_from_bits (&bits, name, s->version);
+  g_free (name);
+  if (script) {
+    swfdec_swf_decoder_add_script (s, script);
+    tag_func_enqueue (s, tag);
+  }
+
+  return SWFDEC_STATUS_OK;
+}
+
+/* only needed for the codec finding stuff */
+static int
+tag_func_sound_stream_head (SwfdecSwfDecoder *s, guint tag)
+{
+  SwfdecBits bits;
+  SwfdecAudioFormat playback_format, format;
+  guint playback_codec, codec;
+  int n_samples;
+
+  bits = s->b;
+
+  /* we don't care about playback suggestions */
+  playback_codec = swfdec_bits_getbits (&bits, 4);
+  playback_format = swfdec_audio_format_parse (&bits);
+  SWFDEC_LOG ("  suggested playback format: %s", swfdec_audio_format_to_string (playback_format));
+
+  codec = swfdec_bits_getbits (&bits, 4);
+  format = swfdec_audio_format_parse (&bits);
+  n_samples = swfdec_bits_get_u16 (&bits);
+  SWFDEC_LOG ("  codec: %u", codec);
+  SWFDEC_LOG ("  format: %s", swfdec_audio_format_to_string (format));
+  SWFDEC_LOG ("  samples: %u", n_samples);
+
+  swfdec_decoder_use_audio_codec (SWFDEC_DECODER (s), codec, format);
+
+  return tag_func_enqueue (s, tag);
+}
+
+static int
+tag_func_metadata (SwfdecSwfDecoder *s, guint tag)
+{
+  char *meta;
+
+  meta = swfdec_bits_get_string (&s->b, s->version);
+  if (meta == NULL) {
+    SWFDEC_ERROR ("invalid metadata contents");
+    return SWFDEC_STATUS_OK;
+  }
+  SWFDEC_LOG ("metadata: %s", meta);
+
+  if (s->metadata) {
+    SWFDEC_ERROR ("Duplicated metadata tag, ignoring");
+    /* FIXME: or use this one? */
+    g_free (meta);
+  } else {
+    s->metadata = meta;
+  }
+  return SWFDEC_STATUS_OK;
+}
+
 struct tag_func_struct
 {
   const char *name;
@@ -526,7 +413,7 @@ static struct tag_func_struct tag_funcs[] = {
   [SWFDEC_TAG_DEFINEBUTTON] = {"DefineButton", tag_func_define_button, 0},
   [SWFDEC_TAG_JPEGTABLES] = {"JPEGTables", swfdec_image_jpegtables, 0},
   [SWFDEC_TAG_SETBACKGROUNDCOLOR] =
-      {"SetBackgroundColor", tag_func_set_background_color, SWFDEC_TAG_DEFINE_SPRITE },
+      {"SetBackgroundColor", tag_func_enqueue, SWFDEC_TAG_DEFINE_SPRITE },
   [SWFDEC_TAG_DEFINEFONT] = {"DefineFont", tag_func_define_font, 0},
   [SWFDEC_TAG_DEFINETEXT] = {"DefineText", tag_func_define_text, 0},
   [SWFDEC_TAG_DOACTION] = {"DoAction", tag_func_do_action, SWFDEC_TAG_DEFINE_SPRITE },
@@ -536,7 +423,7 @@ static struct tag_func_struct tag_funcs[] = {
   [SWFDEC_TAG_DEFINEBUTTONSOUND] =
       {"DefineButtonSound", tag_func_define_button_sound, 0},
   [SWFDEC_TAG_SOUNDSTREAMHEAD] = {"SoundStreamHead", tag_func_sound_stream_head, SWFDEC_TAG_DEFINE_SPRITE },
-  [SWFDEC_TAG_SOUNDSTREAMBLOCK] = {"SoundStreamBlock", tag_func_sound_stream_block, SWFDEC_TAG_DEFINE_SPRITE },
+  [SWFDEC_TAG_SOUNDSTREAMBLOCK] = {"SoundStreamBlock", tag_func_enqueue, SWFDEC_TAG_DEFINE_SPRITE },
   [SWFDEC_TAG_DEFINEBITSLOSSLESS] =
       {"DefineBitsLossless", tag_func_define_bits_lossless, 0},
   [SWFDEC_TAG_DEFINEBITSJPEG2] = {"DefineBitsJPEG2", tag_func_define_bits_jpeg_2, 0},
@@ -555,7 +442,7 @@ static struct tag_func_struct tag_funcs[] = {
   [SWFDEC_TAG_DEFINEMOVIE] = {"DefineMovie", NULL, 0},
   [SWFDEC_TAG_DEFINESPRITE] = {"DefineSprite", tag_func_define_sprite, 0},
   [SWFDEC_TAG_NAMECHARACTER] = {"NameCharacter", NULL, 0},
-  [SWFDEC_TAG_SERIALNUMBER] = {"SerialNumber", NULL, 0},
+  [SWFDEC_TAG_PRODUCTINFO] = {"ProductInfo", NULL, 0},
   [SWFDEC_TAG_GENERATORTEXT] = {"GeneratorText", NULL, 0},
   [SWFDEC_TAG_FRAMELABEL] = {"FrameLabel", tag_func_frame_label, SWFDEC_TAG_DEFINE_SPRITE },
   [SWFDEC_TAG_SOUNDSTREAMHEAD2] = {"SoundStreamHead2", tag_func_sound_stream_head, SWFDEC_TAG_DEFINE_SPRITE },
@@ -568,7 +455,7 @@ static struct tag_func_struct tag_funcs[] = {
   [SWFDEC_TAG_EXPORTASSETS] = {"ExportAssets", tag_func_enqueue, 0},
   [SWFDEC_TAG_IMPORTASSETS] = {"ImportAssets", NULL, 0},
   [SWFDEC_TAG_ENABLEDEBUGGER] = {"EnableDebugger", NULL, 0},
-  [SWFDEC_TAG_DOINITACTION] = {"DoInitAction", tag_func_enqueue, SWFDEC_TAG_DEFINE_SPRITE },
+  [SWFDEC_TAG_DOINITACTION] = {"DoInitAction", tag_func_do_init_action, SWFDEC_TAG_DEFINE_SPRITE },
   [SWFDEC_TAG_DEFINEVIDEOSTREAM] = {"DefineVideoStream", tag_func_define_video, SWFDEC_TAG_DEFINE_SPRITE },
   [SWFDEC_TAG_VIDEOFRAME] = {"VideoFrame", tag_func_video_frame, SWFDEC_TAG_DEFINE_SPRITE },
   [SWFDEC_TAG_DEFINEFONTINFO2] = {"DefineFontInfo2", tag_func_define_font_info, 0},
@@ -581,17 +468,17 @@ static struct tag_func_struct tag_funcs[] = {
   [SWFDEC_TAG_IMPORTASSETS2] = {"ImportAssets2", NULL, 0},
   [SWFDEC_TAG_DEFINEFONTALIGNZONES] = {"DefineFontAlignZones", NULL, 0},
   [SWFDEC_TAG_CSMTEXTSETTINGS] = {"CSMTextSettings", NULL, 0},
-  [SWFDEC_TAG_DEFINEFONT3] = {"DefineFont3", tag_func_define_font_3, 0},
-  [SWFDEC_TAG_AVM2DECL] = {"AVM2Decl", NULL, SWFDEC_TAG_DEFINE_SPRITE },
-  [SWFDEC_TAG_METADATA] = {"Metadata", NULL, 0},
+  [SWFDEC_TAG_DEFINEFONT3] = {"DefineFont3", tag_func_define_font_2, 0},
+  [SWFDEC_TAG_SYMBOLCLASS] = {"SymbolClass", NULL, SWFDEC_TAG_DEFINE_SPRITE },
+  [SWFDEC_TAG_METADATA] = {"Metadata", tag_func_metadata, 0},
   [SWFDEC_TAG_DEFINESCALINGGRID] = {"DefineScalingGrid", NULL, 0},
-  [SWFDEC_TAG_AVM2ACTION] = {"AVM2Action", NULL, SWFDEC_TAG_DEFINE_SPRITE },
+  [SWFDEC_TAG_DOABC] = {"DoAbc", NULL, SWFDEC_TAG_DEFINE_SPRITE },
   [SWFDEC_TAG_DEFINESHAPE4] = {"DefineShape4", tag_define_shape_4, 0},
   [SWFDEC_TAG_DEFINEMORPHSHAPE2] = {"DefineMorphShape2", NULL, 0},
   [SWFDEC_TAG_PRIVATE_IMAGE] = { "PrivateImage", NULL, 0},
   [SWFDEC_TAG_DEFINESCENEDATA] = { "DefineSceneData", NULL, 0},
   [SWFDEC_TAG_DEFINEBINARYDATA] = { "DefineBinaryData", NULL, 0},
-  [SWFDEC_TAG_DEFINEFONTNAME] = { "DefineFontName", NULL, 0},
+  [SWFDEC_TAG_DEFINEFONTNAME] = { "DefineFontName", tag_func_define_font_name, 0},
   [SWFDEC_TAG_STARTSOUND2] = {"StartSound2", NULL, SWFDEC_TAG_DEFINE_SPRITE }
 };
 
