@@ -1,5 +1,5 @@
 /* Swfdec
- * Copyright (C) 2007 Benjamin Otte <otte@gnome.org>
+ * Copyright (C) 2007-2008 Benjamin Otte <otte@gnome.org>
  *               2007 Pekka Lampila <pekka.lampila@iki.fi>
  *
  * This library is free software; you can redistribute it and/or
@@ -22,6 +22,7 @@
 #include "config.h"
 #endif
 
+#include <math.h>
 #include <string.h>
 #include <pango/pangocairo.h>
 
@@ -90,9 +91,10 @@ swfdec_text_field_movie_do_set_text (SwfdecAsContext *cx,
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "s", &value);
 
   swfdec_text_field_movie_set_text (text, value, FALSE);
+  swfdec_text_field_movie_update_layout (text);
 
   if (text->variable != NULL) {
-    if (text->text->html) {
+    if (text->html) {
       swfdec_text_field_movie_set_listen_variable_text (text,
 	  swfdec_text_field_movie_get_html_text (text));
     } else {
@@ -110,7 +112,7 @@ swfdec_text_field_movie_get_html (SwfdecAsContext *cx, SwfdecAsObject *object,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->text->html);
+  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->html);
 }
 
 static void
@@ -124,7 +126,7 @@ swfdec_text_field_movie_set_html (SwfdecAsContext *cx, SwfdecAsObject *object,
 
   swfdec_as_value_to_number (cx, &argv[0]);
 
-  text->text->html = value;
+  text->html = value;
 
   // FIXME: resize? invalidate?
 }
@@ -140,7 +142,7 @@ swfdec_text_field_movie_get_htmlText (SwfdecAsContext *cx,
 
   if (text->style_sheet_input) {
     SWFDEC_AS_VALUE_SET_STRING (ret, text->style_sheet_input);
-  } else if (text->text->html) {
+  } else if (text->html) {
     SWFDEC_AS_VALUE_SET_STRING (ret,
 	swfdec_text_field_movie_get_html_text (text));
   } else {
@@ -159,10 +161,10 @@ swfdec_text_field_movie_set_htmlText (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "s", &value);
 
-  swfdec_text_field_movie_set_text (text, value, text->text->html);
+  swfdec_text_field_movie_set_text (text, value, text->html);
 
   if (text->variable != NULL) {
-    if (text->text->html) {
+    if (text->html) {
       swfdec_text_field_movie_set_listen_variable_text (text,
 	  swfdec_text_field_movie_get_html_text (text));
     } else {
@@ -180,7 +182,7 @@ swfdec_text_field_movie_get_length (SwfdecAsContext *cx, SwfdecAsObject *object,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_INT (ret, g_utf8_strlen (text->input->str, -1));
+  SWFDEC_AS_VALUE_SET_INT (ret, g_utf8_strlen (swfdec_text_buffer_get_text (text->text), -1));
 }
 
 /*
@@ -222,8 +224,8 @@ swfdec_text_field_movie_get_maxChars (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  if (text->text->max_chars != 0) {
-    SWFDEC_AS_VALUE_SET_INT (ret, text->text->max_chars);
+  if (text->max_chars != 0) {
+    SWFDEC_AS_VALUE_SET_INT (ret, text->max_chars);
   } else {
     SWFDEC_AS_VALUE_SET_NULL (ret);
   }
@@ -242,7 +244,7 @@ swfdec_text_field_movie_set_maxChars (SwfdecAsContext *cx,
     return;
 
   swfdec_as_value_to_number (cx, &argv[0]);
-  text->text->max_chars = swfdec_as_value_to_integer (cx, &argv[0]);
+  text->max_chars = swfdec_as_value_to_integer (cx, &argv[0]);
 }
 
 static void
@@ -254,7 +256,7 @@ swfdec_text_field_movie_get_multiline (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->text->multiline);
+  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->multiline);
 }
 
 static void
@@ -269,7 +271,7 @@ swfdec_text_field_movie_set_multiline (SwfdecAsContext *cx,
 
   swfdec_as_value_to_number (cx, &argv[0]);
 
-  text->text->multiline = value;
+  text->multiline = value;
 }
 
 static void
@@ -301,6 +303,9 @@ swfdec_text_field_movie_set_restrict (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "s", &value);
 
+  if (text->restrict_ == NULL && value == SWFDEC_AS_STR_EMPTY)
+    return;
+
   if (SWFDEC_AS_VALUE_IS_UNDEFINED (&argv[0]) ||
       SWFDEC_AS_VALUE_IS_NULL (&argv[0])) {
     text->restrict_ = NULL;
@@ -318,7 +323,7 @@ swfdec_text_field_movie_get_selectable (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->text->selectable);
+  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->selectable);
 }
 
 static void
@@ -333,7 +338,7 @@ swfdec_text_field_movie_set_selectable (SwfdecAsContext *cx,
 
   swfdec_as_value_to_number (cx, &argv[0]);
 
-  text->text->selectable = value;
+  text->selectable = value;
 
   // FIXME: invalidate
 }
@@ -347,7 +352,7 @@ swfdec_text_field_movie_do_get_type (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  if (text->text->editable) {
+  if (text->editable) {
     SWFDEC_AS_VALUE_SET_STRING (ret, SWFDEC_AS_STR_input);
   } else {
     SWFDEC_AS_VALUE_SET_STRING (ret, SWFDEC_AS_STR_dynamic);
@@ -368,10 +373,11 @@ swfdec_text_field_movie_do_set_type (SwfdecAsContext *cx,
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "s", &value);
 
   if (!g_strcasecmp (value, SWFDEC_AS_STR_input)) {
-    text->text->editable = TRUE;
+    text->editable = TRUE;
   } else if (!g_strcasecmp (value, SWFDEC_AS_STR_dynamic)) {
-    text->text->editable = FALSE;
+    text->editable = FALSE;
   }
+  /* else ignore */
 
   // FIXME: invalidate
 }
@@ -422,12 +428,13 @@ swfdec_text_field_movie_get_textHeight (SwfdecAsContext *cx,
     SwfdecAsValue *ret)
 {
   SwfdecTextFieldMovie *text;
-  int height;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  swfdec_text_field_movie_get_text_size (text, NULL, &height);
-  SWFDEC_AS_VALUE_SET_NUMBER (ret, SWFDEC_TWIPS_TO_DOUBLE (height));
+  swfdec_movie_update (SWFDEC_MOVIE (text));
+  swfdec_text_field_movie_update_layout (text);
+  SWFDEC_AS_VALUE_SET_INT (ret, floor (text->layout_height * 
+	text->from_layout.yy / SWFDEC_TWIPS_SCALE_FACTOR));
 }
 
 static void
@@ -436,12 +443,13 @@ swfdec_text_field_movie_get_textWidth (SwfdecAsContext *cx,
     SwfdecAsValue *ret)
 {
   SwfdecTextFieldMovie *text;
-  int width;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  swfdec_text_field_movie_get_text_size (text, &width, NULL);
-  SWFDEC_AS_VALUE_SET_NUMBER (ret, SWFDEC_TWIPS_TO_DOUBLE (width));
+  swfdec_movie_update (SWFDEC_MOVIE (text));
+  swfdec_text_field_movie_update_layout (text);
+  SWFDEC_AS_VALUE_SET_INT (ret, floor (text->layout_width *
+	text->from_layout.xx / SWFDEC_TWIPS_SCALE_FACTOR));
 }
 
 /*
@@ -456,7 +464,7 @@ swfdec_text_field_movie_get_background (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->text->background);
+  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->background);
 }
 
 static void
@@ -471,8 +479,8 @@ swfdec_text_field_movie_set_background (SwfdecAsContext *cx,
 
   swfdec_as_value_to_number (cx, &argv[0]);
 
-  if (text->text->background != value) {
-    text->text->background = value;
+  if (text->background != value) {
+    text->background = value;
     swfdec_movie_invalidate_last (SWFDEC_MOVIE (text));
   }
 }
@@ -487,9 +495,9 @@ swfdec_text_field_movie_get_backgroundColor (SwfdecAsContext *cx,
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
   SWFDEC_AS_VALUE_SET_NUMBER (ret,
-      SWFDEC_COLOR_R (text->background_color) << 16 |
-      SWFDEC_COLOR_G (text->background_color) << 8 |
-      SWFDEC_COLOR_B (text->background_color));
+      SWFDEC_COLOR_RED (text->background_color) << 16 |
+      SWFDEC_COLOR_GREEN (text->background_color) << 8 |
+      SWFDEC_COLOR_BLUE (text->background_color));
 }
 
 static void
@@ -519,7 +527,7 @@ swfdec_text_field_movie_get_border (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->text->border);
+  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->border);
 }
 
 static void
@@ -534,8 +542,8 @@ swfdec_text_field_movie_set_border (SwfdecAsContext *cx,
 
   swfdec_as_value_to_number (cx, &argv[0]);
 
-  if (text->text->border != value) {
-    text->text->border = value;
+  if (text->border != value) {
+    text->border = value;
     swfdec_movie_invalidate_last (SWFDEC_MOVIE (text));
   }
 }
@@ -551,9 +559,9 @@ swfdec_text_field_movie_get_borderColor (SwfdecAsContext *cx,
 
 
   SWFDEC_AS_VALUE_SET_NUMBER (ret,
-      SWFDEC_COLOR_R (text->border_color) << 16 |
-      SWFDEC_COLOR_G (text->border_color) << 8 |
-      SWFDEC_COLOR_B (text->border_color));
+      SWFDEC_COLOR_RED (text->border_color) << 16 |
+      SWFDEC_COLOR_GREEN (text->border_color) << 8 |
+      SWFDEC_COLOR_BLUE (text->border_color));
 }
 
 static void
@@ -586,7 +594,7 @@ swfdec_text_field_movie_get_bottomScroll (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_NUMBER (ret, text->scroll_bottom);
+  SWFDEC_AS_VALUE_SET_NUMBER (ret, text->scroll + text->lines_visible);
 }
 
 static void
@@ -611,11 +619,11 @@ swfdec_text_field_movie_do_set_hscroll (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "i", &value);
 
-  value = CLAMP (value, 0, text->hscroll_max);
-  if (value != text->hscroll) {
-    text->hscroll = value;
-    text->scroll_changed = TRUE;
+  value = CLAMP (value, 0, (int) swfdec_text_field_movie_get_hscroll_max (text));
+  if ((guint) value != text->hscroll) {
     swfdec_movie_invalidate_last (SWFDEC_MOVIE (text));
+    text->hscroll = value;
+    swfdec_text_field_movie_emit_onScroller (text);
   }
 }
 
@@ -628,11 +636,7 @@ swfdec_text_field_movie_get_maxhscroll (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  if (!text->text->word_wrap) {
-    SWFDEC_AS_VALUE_SET_NUMBER (ret, text->hscroll_max);
-  } else {
-    SWFDEC_AS_VALUE_SET_NUMBER (ret, 0);
-  }
+  SWFDEC_AS_VALUE_SET_INT (ret, swfdec_text_field_movie_get_hscroll_max (text));
 }
 
 static void
@@ -644,7 +648,7 @@ swfdec_text_field_movie_get_maxscroll (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_NUMBER (ret, text->scroll_max);
+  SWFDEC_AS_VALUE_SET_NUMBER (ret, text->scroll_max + 1);
 }
 
 static void
@@ -683,7 +687,7 @@ swfdec_text_field_movie_do_get_scroll (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_NUMBER (ret, text->scroll);
+  SWFDEC_AS_VALUE_SET_NUMBER (ret, text->scroll + 1);
 }
 
 static void
@@ -696,12 +700,12 @@ swfdec_text_field_movie_do_set_scroll (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "i", &value);
 
-  value = CLAMP (value, 1, text->scroll_max);
-  if (value != text->scroll) {
-    text->scroll_bottom += value - text->scroll;
+  if ((guint) value != text->scroll + 1) {
+    swfdec_text_field_movie_update_layout (text);
+    value = CLAMP (value - 1, 0, (int) text->scroll_max);
     text->scroll = value;
-    text->scroll_changed = TRUE;
     swfdec_movie_invalidate_last (SWFDEC_MOVIE (text));
+    swfdec_text_field_movie_emit_onScroller (text);
   }
 }
 
@@ -717,7 +721,7 @@ swfdec_text_field_movie_get_autoSize (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  switch (text->text->auto_size) {
+  switch (text->auto_size) {
     case SWFDEC_AUTO_SIZE_NONE:
       SWFDEC_AS_VALUE_SET_STRING (ret, SWFDEC_AS_STR_none);
       break;
@@ -741,7 +745,6 @@ swfdec_text_field_movie_set_autoSize (SwfdecAsContext *cx,
     SwfdecAsValue *ret)
 {
   SwfdecTextFieldMovie *text;
-  SwfdecAutoSize old;
   const char *s;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
@@ -751,30 +754,23 @@ swfdec_text_field_movie_set_autoSize (SwfdecAsContext *cx,
 
   if (SWFDEC_AS_VALUE_IS_BOOLEAN (&argv[0])) {
     if (SWFDEC_AS_VALUE_GET_BOOLEAN (&argv[0])) {
-      text->text->auto_size = SWFDEC_AUTO_SIZE_LEFT;
+      text->auto_size = SWFDEC_AUTO_SIZE_LEFT;
     } else {
-      text->text->auto_size = SWFDEC_AUTO_SIZE_NONE;
+      text->auto_size = SWFDEC_AUTO_SIZE_NONE;
     }
-    return;
-  }
+  } else {
+    swfdec_as_value_to_number (cx, &argv[0]);
+    s = swfdec_as_value_to_string (cx, &argv[0]);
 
-  swfdec_as_value_to_number (cx, &argv[0]);
-  s = swfdec_as_value_to_string (cx, &argv[0]);
-
-  old = text->text->auto_size;
-  if (!g_ascii_strcasecmp (s, "none")) {
-    text->text->auto_size = SWFDEC_AUTO_SIZE_NONE;
-  } else if (!g_ascii_strcasecmp (s, "left")) {
-    text->text->auto_size = SWFDEC_AUTO_SIZE_LEFT;
-  } else if (!g_ascii_strcasecmp (s, "right")) {
-    text->text->auto_size = SWFDEC_AUTO_SIZE_RIGHT;
-  } else if (!g_ascii_strcasecmp (s, "center")) {
-    text->text->auto_size = SWFDEC_AUTO_SIZE_CENTER;
-  }
-
-  if (text->text->auto_size != old) {
-    swfdec_text_field_movie_auto_size (text);
-    // FIXME: fix scrolling
+    if (!g_ascii_strcasecmp (s, "left")) {
+      text->auto_size = SWFDEC_AUTO_SIZE_LEFT;
+    } else if (!g_ascii_strcasecmp (s, "right")) {
+      text->auto_size = SWFDEC_AUTO_SIZE_RIGHT;
+    } else if (!g_ascii_strcasecmp (s, "center")) {
+      text->auto_size = SWFDEC_AUTO_SIZE_CENTER;
+    } else {
+      text->auto_size = SWFDEC_AUTO_SIZE_NONE;
+    }
   }
 }
 
@@ -787,7 +783,7 @@ swfdec_text_field_movie_get_password (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->text->password);
+  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, swfdec_text_layout_get_password (text->layout));
 }
 
 static void
@@ -802,15 +798,8 @@ swfdec_text_field_movie_set_password (SwfdecAsContext *cx,
 
   swfdec_as_value_to_number (cx, &argv[0]);
 
-  if (text->text->password != value) {
-    text->text->password = value;
-    if (!value && text->asterisks != NULL) {
-      g_free (text->asterisks);
-      text->asterisks = NULL;
-      text->asterisks_length = 0;
-    }
-    swfdec_movie_invalidate_last (SWFDEC_MOVIE (text));
-  }
+  swfdec_text_layout_set_password (text->layout, value);
+  swfdec_movie_invalidate_last (SWFDEC_MOVIE (text));
 }
 
 static void
@@ -822,7 +811,8 @@ swfdec_text_field_movie_get_wordWrap (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->text->word_wrap);
+  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, 
+      swfdec_text_layout_get_word_wrap (text->layout));
 }
 
 static void
@@ -837,12 +827,7 @@ swfdec_text_field_movie_set_wordWrap (SwfdecAsContext *cx,
 
   swfdec_as_value_to_number (cx, &argv[0]);
 
-  if (text->text->word_wrap != value) {
-    text->text->word_wrap = value;
-    swfdec_movie_invalidate_last (SWFDEC_MOVIE (text));
-    swfdec_text_field_movie_auto_size (text);
-    // special case: don't set scrolling
-  }
+  swfdec_text_layout_set_word_wrap (text->layout, value);
 }
 
 /*
@@ -857,7 +842,7 @@ swfdec_text_field_movie_get_embedFonts (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->text->embed_fonts);
+  SWFDEC_AS_VALUE_SET_BOOLEAN (ret, text->embed_fonts);
 }
 
 static void
@@ -872,10 +857,10 @@ swfdec_text_field_movie_set_embedFonts (SwfdecAsContext *cx,
 
   swfdec_as_value_to_number (cx, &argv[0]);
 
-  if (!text->text->embed_fonts && value)
+  if (!text->embed_fonts && value)
     SWFDEC_FIXME ("Using embed fonts in TextField not supported");
 
-  text->text->embed_fonts = value;
+  text->embed_fonts = value;
 
   // FIXME: resize
 }
@@ -957,7 +942,8 @@ swfdec_text_field_movie_get_textColor (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  SWFDEC_AS_VALUE_SET_NUMBER (ret, text->format_new->color);
+  SWFDEC_AS_VALUE_SET_NUMBER (ret, 
+      swfdec_text_buffer_get_default_attributes (text->text)->color);
 }
 
 // This doesn't work the same way as TextFormat's color setting
@@ -971,7 +957,8 @@ swfdec_text_field_movie_set_textColor (SwfdecAsContext *cx,
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "i", &value);
 
-  text->format_new->color = swfdec_text_field_movie_int_to_color (cx, value);
+  /* FIXME: don't access private variables, you lazy hacker! */
+  text->text->default_attributes.color = swfdec_text_field_movie_int_to_color (cx, value);
 }
 
 SWFDEC_AS_NATIVE (104, 300, swfdec_text_field_movie_get_gridFitType)
@@ -1074,13 +1061,20 @@ swfdec_text_field_movie_getNewTextFormat (SwfdecAsContext *cx,
     SwfdecAsValue *ret)
 {
   SwfdecTextFieldMovie *text;
+  SwfdecTextFormat *format;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
-  swfdec_text_format_init_properties (cx);
+  format = SWFDEC_TEXT_FORMAT (swfdec_text_format_new (cx));
+  if (format == NULL)
+    return;
 
-  SWFDEC_AS_VALUE_SET_OBJECT (ret,
-      SWFDEC_AS_OBJECT (swfdec_text_format_copy (text->format_new)));
+  swfdec_text_attributes_copy (&format->attr, 
+      swfdec_text_buffer_get_default_attributes (text->text),
+      SWFDEC_TEXT_ATTRIBUTES_MASK);
+  format->values_set = SWFDEC_TEXT_ATTRIBUTES_MASK;
+
+  SWFDEC_AS_VALUE_SET_OBJECT (ret, SWFDEC_AS_OBJECT (format));
 }
 
 SWFDEC_AS_NATIVE (104, 105, swfdec_text_field_movie_setNewTextFormat)
@@ -1090,14 +1084,15 @@ swfdec_text_field_movie_setNewTextFormat (SwfdecAsContext *cx,
     SwfdecAsValue *ret)
 {
   SwfdecTextFieldMovie *text;
-  SwfdecAsObject *obj;
+  SwfdecTextFormat *format;
 
-  SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "o", &obj);
+  SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "o", &format);
 
-  if (!SWFDEC_IS_TEXT_FORMAT (obj))
+  if (!SWFDEC_IS_TEXT_FORMAT (format))
     return;
 
-  swfdec_text_format_add (text->format_new, SWFDEC_TEXT_FORMAT (obj));
+  swfdec_text_buffer_set_default_attributes (text->text, 
+      &format->attr, format->values_set);
 }
 
 SWFDEC_AS_NATIVE (104, 102, swfdec_text_field_movie_setTextFormat)
@@ -1108,34 +1103,41 @@ swfdec_text_field_movie_setTextFormat (SwfdecAsContext *cx,
 {
   SwfdecTextFieldMovie *text;
   SwfdecTextFormat *format;
-  int val, start_index, end_index;
+  int val;
   guint i;
+  const char *string;
+  gsize start, end, length;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
   if (argc < 1)
     return;
 
-  i = 0;
-  if (argc <= i + 1) {
-    start_index = 0;
-    end_index = g_utf8_strlen (text->input->str, -1);
+  string = swfdec_text_buffer_get_text (text->text);
+  length = g_utf8_strlen (string, -1);
+  if (argc <= 1) {
+    start = 0;
+    end = length;
+    i = 0;
   } else {
-    start_index = val = swfdec_as_value_to_integer (cx, &argv[i++]);
-    start_index = CLAMP (start_index, 0, g_utf8_strlen (text->input->str, -1));
-    if (argc <= i + 1) {
+    val = swfdec_as_value_to_integer (cx, &argv[0]);
+    start = MAX (val, 0);
+    start = MIN (start, length);
+    if (argc <= 2) {
       if (val < 0) { // fail
-	start_index = end_index = 0;
-      } else{
-	end_index = start_index + 1;
+	return;
+      } else {
+	end = MIN (start + 1, length);
       }
+      i = 1;
     } else {
-      end_index = swfdec_as_value_to_integer (cx, &argv[i++]);
+      val = swfdec_as_value_to_integer (cx, &argv[1]);
+      end = MAX (val, 0);
+      end = CLAMP (end, start, length);
+      i = 2;
     }
-    end_index =
-      CLAMP (end_index, start_index, g_utf8_strlen (text->input->str, -1));
   }
-  if (start_index == end_index)
+  if (start == end)
     return;
 
   if (!SWFDEC_AS_VALUE_IS_OBJECT (&argv[i]))
@@ -1144,17 +1146,21 @@ swfdec_text_field_movie_setTextFormat (SwfdecAsContext *cx,
     return;
 
   format = SWFDEC_TEXT_FORMAT (SWFDEC_AS_VALUE_GET_OBJECT (&argv[i]));
+  start = g_utf8_offset_to_pointer (string, start) - string;
+  end = g_utf8_offset_to_pointer (string, end) - string;
 
-  swfdec_text_field_movie_set_text_format (text, format,
-      g_utf8_offset_to_pointer (text->input->str, start_index) -
-      text->input->str,
-      g_utf8_offset_to_pointer (text->input->str, end_index) -
-      text->input->str);
+  swfdec_text_buffer_set_attributes (text->text, start, end - start, 
+      &format->attr, format->values_set);
 
   swfdec_movie_invalidate_last (SWFDEC_MOVIE (text));
-  swfdec_text_field_movie_auto_size (text);
-  // special case: update the max values, not the current values
-  swfdec_text_field_movie_update_scroll (text, FALSE);
+  /* special case: update the max values, not the current values */
+  /* swfdec_text_field_movie_update_layout (text); */
+  /* FIXME: emit onScroller? */
+  text->layout_width = swfdec_text_layout_get_width (text->layout);
+  text->layout_height = swfdec_text_layout_get_height (text->layout);
+  text->scroll_max = swfdec_text_layout_get_n_rows (text->layout)
+    - swfdec_text_layout_get_visible_rows_end (text->layout, 
+	text->layout_area.height);
 }
 
 SWFDEC_AS_NATIVE (104, 101, swfdec_text_field_movie_getTextFormat)
@@ -1165,40 +1171,49 @@ swfdec_text_field_movie_getTextFormat (SwfdecAsContext *cx,
 {
   SwfdecTextFieldMovie *text;
   SwfdecTextFormat *format;
-  int val, start_index, end_index;
+  int val;
+  const char *string;
+  guint mask;
+  const SwfdecTextAttributes *attr;
+  gsize start, end, length;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "");
 
+  string = swfdec_text_buffer_get_text (text->text);
+  length = g_utf8_strlen (string, -1);
+
+  format = SWFDEC_TEXT_FORMAT (swfdec_text_format_new (cx));
+  SWFDEC_AS_VALUE_SET_OBJECT (ret, SWFDEC_AS_OBJECT (format));
+
   if (argc == 0) {
-    start_index = 0;
-    end_index = g_utf8_strlen (text->input->str, -1);
+    start = 0;
+    end = length;
   } else {
-    start_index = val = swfdec_as_value_to_integer (cx, &argv[0]);
-    start_index = CLAMP (start_index, 0, g_utf8_strlen (text->input->str, -1));
+    val = swfdec_as_value_to_integer (cx, &argv[0]);
+    start = MAX (val, 0);
+    start = MIN (start, length);
     if (argc == 1) {
       if (val < 0) { // fail
-	start_index = end_index = 0;
-      } else{
-	end_index = start_index + 1;
+	return;
+      } else {
+	end = MIN (start + 1, length);
       }
     } else {
-      end_index = swfdec_as_value_to_integer (cx, &argv[1]);
+      val = swfdec_as_value_to_integer (cx, &argv[1]);
+      end = MAX (val, 0);
+      end = CLAMP (end, start, length);
     }
-    end_index =
-      CLAMP (end_index, start_index, g_utf8_strlen (text->input->str, -1));
   }
 
-  if (start_index == end_index) {
-    format = SWFDEC_TEXT_FORMAT (swfdec_text_format_new (cx));
-  } else {
-    format = swfdec_text_field_movie_get_text_format (text,
-      g_utf8_offset_to_pointer (text->input->str, start_index) -
-      text->input->str,
-      g_utf8_offset_to_pointer (text->input->str, end_index) -
-      text->input->str);
-  }
+  if (start == end)
+    return;
 
-  SWFDEC_AS_VALUE_SET_OBJECT (ret, SWFDEC_AS_OBJECT (format));
+  start = g_utf8_offset_to_pointer (string, start) - string;
+  end = g_utf8_offset_to_pointer (string, end) - string;
+  mask = swfdec_text_buffer_get_unique (text->text, start, end - start);
+  attr = swfdec_text_buffer_get_attributes (text->text, start);
+  swfdec_text_attributes_copy (&format->attr, attr, mask);
+  format->values_set = mask;
 }
 
 SWFDEC_AS_NATIVE (104, 100, swfdec_text_field_movie_replaceSel)
@@ -1218,8 +1233,8 @@ swfdec_text_field_movie_replaceText (SwfdecAsContext *cx,
 {
   SwfdecTextFieldMovie *text;
   int start_index, end_index;
-  const char *str;
-
+  const char *string, *str;
+  gsize start, end;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_TEXT_FIELD_MOVIE, &text, "iis", &start_index,
       &end_index, &str);
@@ -1228,15 +1243,27 @@ swfdec_text_field_movie_replaceText (SwfdecAsContext *cx,
     return;
   if (end_index < start_index)
     return;
+  /* if there was a style sheet set when setting the text, modifications are
+   * not allowed */
+  if (text->style_sheet_input)
+    return;
 
-  start_index = MIN (start_index, g_utf8_strlen (text->input->str, -1));
-  end_index = MIN (end_index, g_utf8_strlen (text->input->str, -1));
+  string = swfdec_text_buffer_get_text (text->text);
+  start = end = g_utf8_strlen (string, -1);
+  start = MIN ((gsize) start_index, start);
+  end = MIN ((gsize) end_index, end);
+  start = g_utf8_offset_to_pointer (string, start) - string;
+  end = g_utf8_offset_to_pointer (string, end) - string;
 
-  swfdec_text_field_movie_replace_text (text,
-      g_utf8_offset_to_pointer (text->input->str, start_index) -
-      text->input->str,
-      g_utf8_offset_to_pointer (text->input->str, end_index) -
-      text->input->str, str);
+  /* anyone explain to me why Flash 7 has this special case plz */
+  if (end > start && (str != SWFDEC_AS_STR_EMPTY || cx->version != 7))
+    swfdec_text_buffer_delete_text (text->text, start, end - start);
+
+  swfdec_text_buffer_insert_text (text->text, start, str);
+  /* yes, modifying text changes the default format */
+  swfdec_text_buffer_set_default_attributes (text->text,
+      swfdec_text_buffer_get_attributes (text->text, 0),
+      SWFDEC_TEXT_ATTRIBUTES_MASK);
 }
 
 // static
@@ -1317,6 +1344,8 @@ swfdec_text_field_movie_createTextField (SwfdecAsContext *cx,
   SwfdecAsValue val;
 
   SWFDEC_AS_CHECK (SWFDEC_TYPE_MOVIE, &parent, "siiiii", &name, &depth, &x, &y, &width, &height);
+  width = ABS (width);
+  height = ABS (height);
 
   edittext = g_object_new (SWFDEC_TYPE_TEXT_FIELD, NULL);
   edittext->html = FALSE;
@@ -1339,12 +1368,10 @@ swfdec_text_field_movie_createTextField (SwfdecAsContext *cx,
   edittext->indent = 0;
   edittext->leading = 0;
 
-  SWFDEC_GRAPHIC (edittext)->extents.x0 = SWFDEC_DOUBLE_TO_TWIPS (x);
-  SWFDEC_GRAPHIC (edittext)->extents.x1 =
-    SWFDEC_GRAPHIC (edittext)->extents.x0 + SWFDEC_DOUBLE_TO_TWIPS (width);
-  SWFDEC_GRAPHIC (edittext)->extents.y0 = SWFDEC_DOUBLE_TO_TWIPS (y);
-  SWFDEC_GRAPHIC (edittext)->extents.y1 =
-    SWFDEC_GRAPHIC (edittext)->extents.y0 + SWFDEC_DOUBLE_TO_TWIPS (height);
+  SWFDEC_GRAPHIC (edittext)->extents.x0 = 0;
+  SWFDEC_GRAPHIC (edittext)->extents.x1 = SWFDEC_DOUBLE_TO_TWIPS (width);
+  SWFDEC_GRAPHIC (edittext)->extents.y0 = 0;
+  SWFDEC_GRAPHIC (edittext)->extents.y1 = SWFDEC_DOUBLE_TO_TWIPS (height);
 
   movie = swfdec_movie_find (parent, depth);
   if (movie)
@@ -1354,6 +1381,13 @@ swfdec_text_field_movie_createTextField (SwfdecAsContext *cx,
       SWFDEC_GRAPHIC (edittext), name);
   g_assert (SWFDEC_IS_TEXT_FIELD_MOVIE (movie));
   g_object_unref (edittext);
+
+  swfdec_movie_begin_update_matrix (movie);
+  movie->matrix.x0 = SWFDEC_DOUBLE_TO_TWIPS (x);
+  movie->matrix.y0 = SWFDEC_DOUBLE_TO_TWIPS (y);
+  movie->modified = TRUE;
+  swfdec_movie_end_update_matrix (movie);
+
   swfdec_movie_initialize (movie);
   swfdec_movie_update (movie);
 
@@ -1380,7 +1414,8 @@ swfdec_text_field_movie_createTextField (SwfdecAsContext *cx,
       SWFDEC_AS_STR___constructor__, &val,
       SWFDEC_AS_VARIABLE_HIDDEN | SWFDEC_AS_VARIABLE_VERSION_6_UP);
 
-  swfdec_as_function_call (fun, SWFDEC_AS_OBJECT (movie), 0, NULL, rval);
+  swfdec_as_function_call (fun, SWFDEC_AS_OBJECT (movie), 0, NULL, 
+      cx->version > 7 ? rval : NULL);
   cx->frame->construct = TRUE;
   swfdec_as_context_run (cx);
 }
@@ -1518,10 +1553,7 @@ swfdec_text_field_movie_construct (SwfdecAsContext *cx, SwfdecAsObject *object,
 {
   if (!cx->frame->construct) {
     SwfdecAsValue val;
-    if (!swfdec_as_context_use_mem (cx, sizeof (SwfdecAsObject)))
-      return;
-    object = g_object_new (SWFDEC_TYPE_AS_OBJECT, NULL);
-    swfdec_as_object_add (object, cx, sizeof (SwfdecAsObject));
+    object = swfdec_as_object_new_empty (cx);
     swfdec_as_object_get_variable (cx->global, SWFDEC_AS_STR_TextField, &val);
     if (SWFDEC_AS_VALUE_IS_OBJECT (&val)) {
       swfdec_as_object_set_constructor (object,
