@@ -38,6 +38,9 @@
  * will want to use swfdec_loader_get_url() to get its url and then use the 
  * functions in this section to access it.
  *
+ * Note that URLs inside Swfdec mirror the behavior of Flash and will therefore
+ * not miror standards behavior, but the behavior of Flash.
+ *
  * @see_also: #SwfdecLoader
  */
 
@@ -223,8 +226,12 @@ swfdec_url_new_parent (const SwfdecURL *url)
   SwfdecURL *ret;
   
   path = g_strdup (url->path);
-  if (path)
-    swfdec_url_path_to_parent_path (path);
+  if (path) {
+    if (!swfdec_url_path_to_parent_path (path)) {
+      g_free (path);
+      path = NULL;
+    }
+  }
   ret = swfdec_url_new_components (url->protocol, url->host, url->port,
       path, NULL);
   g_free (path);
@@ -269,9 +276,9 @@ swfdec_url_new_relative (const SwfdecURL *url, const char *string)
     /* relative URL */
     char *cur = g_strdup (url->path);
     while (g_str_has_prefix (string, "../")) {
-      if (!swfdec_url_path_to_parent_path (cur)) {
+      if (cur && !swfdec_url_path_to_parent_path (cur)) {
 	g_free (cur);
-	return NULL;
+	cur = NULL;
       }
       string += 3;
     }
@@ -279,9 +286,13 @@ swfdec_url_new_relative (const SwfdecURL *url, const char *string)
       g_free (cur);
       return NULL;
     }
-    path = g_strconcat (cur, "/", string, NULL);
-    g_free (cur);
-    cur = path;
+    if (cur) {
+      path = g_strconcat (cur, "/", string, NULL);
+      g_free (cur);
+      cur = path;
+    } else {
+      cur = g_strdup (string);
+    }
     query = strchr (cur, '?');
     if (query == NULL) {
       path = *string ? g_strdup (cur) : NULL;
@@ -525,19 +536,18 @@ swfdec_url_is_local (const SwfdecURL *url)
 }
 
 /**
- * swfdec_url_equal:
+ * swfdec_url_host_equal:
  * @a: a #SwfdecURL
  * @b: a #SwfdecURL
  *
- * Compares the 2 given URLs for equality. 2 URLs are considered equal, when
- * they point to the same resource. This function is intended to be 
- * used together with swfdec_url_hash() in a #GHashtable.
- *
- * Returns: %TRUE if the 2 given urls point to the same resource, %FALSE 
+ * Compares the 2 given URLs for equality, ignoring path. 2 URLs are
+ * considered to have equal hosts when they have the same protocol,
+ * host, and port.
+ * Returns: %TRUE if the 2 given urls point to the same host, %FALSE
  *          otherwise.
  **/
 gboolean
-swfdec_url_equal (gconstpointer a, gconstpointer b)
+swfdec_url_host_equal (gconstpointer a, gconstpointer b)
 {
   const SwfdecURL *urla = a;
   const SwfdecURL *urlb = b;
@@ -557,6 +567,30 @@ swfdec_url_equal (gconstpointer a, gconstpointer b)
   if (urla->port != urlb->port)
     return FALSE;
 
+  return TRUE;
+}
+
+/**
+ * swfdec_url_equal:
+ * @a: a #SwfdecURL
+ * @b: a #SwfdecURL
+ *
+ * Compares the 2 given URLs for equality. 2 URLs are considered equal, when
+ * they point to the same resource. This function is intended to be
+ * used together with swfdec_url_hash() in a #GHashtable.
+ *
+ * Returns: %TRUE if the 2 given urls point to the same resource, %FALSE
+ *          otherwise.
+ **/
+gboolean
+swfdec_url_equal (gconstpointer a, gconstpointer b)
+{
+  const SwfdecURL *urla = a;
+  const SwfdecURL *urlb = b;
+
+  if (!swfdec_url_host_equal (urla, urlb))
+    return FALSE;
+
   if (urla->path == NULL) {
     if (urlb->path != NULL)
       return FALSE;
